
package PHTDB;

require Exporter;
@ISA = qw(Exporter Foomatic::DB);
@EXPORT_OK = qw(normalizename $hcache $cachedir
		get_sku_link get_sku_vendor get_sku_price get_sku_stock); 

use lib "@LIBDIR@";
use Foomatic::DB;
use Foomatic::Defaults;
use Storable qw/nstore retrieve/;
use CGI qw/ :standard :html3 /;

sub new { 
    return Foomatic::DB::new(@_); 
}

sub normalizename { return Foomatic::DB::normalizename(@_); }
sub translate_printer_id { return Foomatic::DB::translate_printer_id(@_); }

$cachedir = "@CACHEDIR@";
$hcache = "$cachedir/hcache";
sub get_hcache { return "$cachedir/hcache"; }

sub get_skus {
    my ($this, $id) = (@_);

    if (!defined($this->{'skucache'})) {
	open REF, "refer.data\n" 
	    or do {
		#warn "Cannot open refer.data";
		return ();
	    };
	while(<REF>) {
	    next if m!^\s*#!;
	    next if m!^\s*$!;

	    if (m!^([\-\_\w]+)\s+?(.*)$!) {
		#print STDERR "got id:$1 sku:$2\n";
		my $sku = $2;
		chomp $sku;
		push(@{$this->{'skucache'}{$1}}, (split(',',$sku)));
	    }
	}
	close REF;

	# interpret ref: include key...
	my $id;
	for $id (keys(%{$this->{'skucache'}})) {
	    my $list = $this->{'skucache'}{$id};
	    my $item;
	    my @newlist = undef;
	    for $item (@{$list}) {
		if ($item =~ m!^ref\:([^\s]+)!) {
		    my $r = $1;
		    push(@newlist, @{$this->{'skucache'}{$r}});
		} else {
		    push(@newlist, $item);
		}
	    }
	    $this->{'skucache'}{$id} = \@newlist;
	}
    }

    return @{$this->{'skucache'}{$id}};
}

sub _load_pricecache {
    my ($this) = @_;

    if (!defined($this->{'pricecache'})) {
	if (-f "refer.prices") {
	    $this->{'pricecache'} = retrieve("refer.prices")
		or warn "Could not load refer.prices; starting empty";
	}
    }

}

sub get_sku_price {
    my ($this, $sku) = (@_);

    $this->_load_pricecache();
    return $this->{'pricecache'}{$sku}{'price'};
}
sub get_sku_stock {
    my ($this, $sku) = (@_);

    $this->_load_pricecache();
    return ($this->{'pricecache'}{$sku}{'nostock'} ? 0 : 1);
}
sub get_sku_timestamp {
    my ($this, $sku) = (@_);

    $this->_load_pricecache();
    return $this->{'pricecache'}{$sku}{'timestamp'};
}

sub set_sku_price {
    my ($this, $sku, $price, $nostock, $timestamp) = @_;
    
    if (!defined($timestamp)) {
	delete ($this->{'pricecache'}{$sku});
    } else {
	$this->{'pricecache'}{$sku} = {'price' => $price,
				       'timestamp' => $timestamp,
				       'nostock' => $nostock };
    }
}

sub write_pricedata {
    my ($this) = @_;
    nstore($this->{'pricecache'}, "refer.prices")
	or die "Could not write refer.prices";
}

sub get_sku_vendor {
    my ($this, $vend) = @_;

    %vendors = ('buy' => 
		'<a href="http://click.linksynergy.com/fs-bin/click?id=dXmY9R8pmFU&offerid=37803.10000077&type=3&subid=0" >Buy.com</a><IMG border=0 width=1 height=1 src="http://ad.linksynergy.com/fs-bin/show?id=dXmY9R8pmFU&bids=37803.10000077&type=3&subid=0">',

                'amz' => '<A HREF="http://www.amazon.com/exec/obidos/redirect?tag=linuxprintingorg&path=tg/browse/-/172635">Amazon.com</A>',

                'azd' => '<A HREF="http://www.amazon.de/exec/obidos/redirect?tag=linuxprinting-21&path=tg/browse/-/713008">Amazon.<b>DE</b></A>',

		'tig' => '<IMG SRC="http://service.bfast.com/bfast/serve?bfmid=26399296&siteid=34553049&bfpage=tiger_button" BORDER="0" WIDTH="1" HEIGHT="1" NOSAVE ><A HREF="http://service.bfast.com/bfast/click?bfmid=26399296&siteid=34553049&bfpage=tiger_button" TARGET="_top">TigerDirect</A>',

		);

    return $vendors{$vend};
}

sub get_sku_link {
    my ($this, $vend, $sku, $text) = @_;

    # also fix in get_sku_vendor above...
    my $imgtag = 1;

    if ($vend eq 'buy') {
	return a({href => 
		      "http://click.linksynergy.com/fs-bin/stat?id=dXmY9R8pmFU&offerid=27222.$sku&type=2&subid=0"}, 
		 $text) . ($imgtag ? 
		     '<IMG border=0 width=1 height=1 src="http://ad.linksynergy.com/fs-bin/show?id=dXmY9R8pmFU&bids=27222&type=2&subid=0" alt="">' : '');
    } elsif ($vend eq 'amz') {
	return a({href=>"http://www.amazon.com/exec/obidos/ASIN/$sku/ref=nosim/linuxprintingorg"},$text);
    } elsif ($vend eq 'azd') {
	return a({href=>"http://www.amazon.de/exec/obidos/ASIN/$sku/ref=nosim/linuxprinting-21"},$text);
    } elsif ($vend eq 'pcm') {
	$sku =~ m!(\d+)\/(\d+)!;
	my ($aid, $dpno) = ($1, $2);

	return a({href=>"http://www.commission-junction.com/track/track.dll?AID=$aid&PID=577662&TYPE=PN&URL=http%3A%2F%2Fwww%2Ecc%2Dinc%2Ecom%2Freferrals%2Fdefault%2Easp%3Fstore%3Dpcmall%26dpno%3D$dpno%26source%3Dzwb12165"},
		 $text)
	    . ($imgtag ? "<img src=\"http://www.commission-junction.com/banners/tracker.exe?AID=$aid&PID=577662&banner=0.gif\" height=\"1\" width=\"1\" border=\"0\">" : '');
	
    } elsif ($vend eq 'tig') {
	# tigerdirect via reporting.net
	return a({href=>"http://service.bfast.com/bfast/click?bfmid=26399296&siteid=34553049&bfpid=".$sku."&bfmtype=computer_equipm"},
		 $text)
	    . ($imgtag ? "<IMG SRC=\"http://service.bfast.com/bfast/serve?bfmid=26399296&siteid=34553049&bfpid=".$sku."&bfmtype=computer_equipm\" BORDER=\"0\" WIDTH=\"1\" HEIGHT=\"1\" NOSAVE >" : '');
    } else {
	return "[BAD SKU] $text";
    }

    die;

}

1;
