#! /usr/bin/perl

use CGI qw/ :standard :html3 /;
use CGI escape, unescape, font, center;

use lib "@LIBDIR@";
use Foomatic::Defaults;
use PHTDB qw/$hcache $cachedir/;
use driverpackages qw/numpackagesformask downloadlinkline drivermask/;

my %driver_types = ('F' => 'Filter',
                    'V' => 'OpenPrinting Vector',
                    'I' => 'IJS',
                    'C' => 'CUPS Raster',
		    'P' => 'PostScript',
		    'U' => 'Ghostscript Uniprint',
		    'G' => 'Ghostscript built-in');

my $query = new CGI;

# Check for special format request
my $type = $query->param('type');
die "illegal type" if ($type and $type !~ /^[FPUG]$/);

# Nothing above here should be computationally intensive!
#####################

# OK, is there a cached page?
my $cachepage = "$hcache/driver-list-type-$type.html";
my @output = ();
if (-f $cachepage) {
    for (my $i = 0; $i < 20; $i++) {
	open CFILE, $cachepage or die "error opening $cachepage";
	my $o = join('', <CFILE>);
	close CFILE;
	# Only output the cashed page if it was really completely written
	# Before introduction of this measure pages would not display due
	# to an incomplete cache file until the next page rebuild (or
	# until manually nuking the cache).
	if ($o =~ m!</html>\s*$!s) {
	    @output = ($o);
	    last;
	}
	# If we have non-empty output, the other process is already
        # writing the cache file or there is definitely written a broken
        # cache. Retry only once in this case.
        $i = 18 if $o;
        sleep 10;
    }
}
if (!@output) {
    # Open cache file already now to tell concurrent processes that we are
    # already computing this case
    umask 0002;
    if (! -d $hcache) {
	mkdir $hcache or warn "Could not mkdir $hcache";
	chmod 2775, $hcache;
    }
    die "Error writing $cachepage\n" if !(open CFILE, "> $cachepage");

    use Foomatic::DB;
    use Theme qw/navbar navbar2/;

    my $db = new Foomatic::DB;
    my $overfile = "$cachedir/pcache/overview.perl";
    $db->{'overviewfile'} = $overfile;
    
    # Build data structure of all data
    my ($p, %dat, %prncount, $pr, $dr);

    # Count how many printers each driver supports
    $db->get_overview();
    for $pr (@{$db->{'overview'}}) {
	for $dr (@{$pr->{'drivers'}}) {
	    $prncount{$dr} ++;
        }
    }

    # Fetch list of all drivers
    my @drivers = $db->get_driverlist();
    my $d;
    for $d (@drivers) {
	my $drvdata = $db->get_driver($d);
	
	my $id = $d;
	my $type = $drvdata->{'type'};
	my $count = $prncount{$d};
	my $descr = $drvdata->{'shortdescription'};
	my $packages = $drvdata->{'packages'};
	$descr =~ s/(<\s*p\s*\/?\s*>\s*|<\s*br\s*\/?\s*>\s*)+/ /g;
	$descr =~ s/[\s\n]+/ /gs;
	$descr =~ s/^ //;
	$descr =~ s/ $//;
	$descr =~ s/ /&nbsp;/g;


	$dat{$d}->{'type'} = $type;
	$dat{$d}->{'count'} = $count;
	$dat{$d}->{'id'} = $id;
	$dat{$d}->{'descr'} = $descr;
	$dat{$d}->{'packages'} = $packages;
    }
    
    my @dlist;
    my $driv;

  driver_loop:
    for $driv (sort {Foomatic::DB::normalizename($a) cmp
		     Foomatic::DB::normalizename($b) } keys(%dat)) {
	
	my $d = $dat{$driv};
	my $url = "show_driver.cgi?driver=$driv";
	my $dtype = $driver_types{$d->{'type'}};
	
	next driver_loop if ($type and $type ne $d->{'type'});
	
	# don't bother printing the type if we're listing one type only;
	# the user knows what type it is!
	my $typestr = $type ? "" : " ($dtype)";
	
	my $count = $d->{'count'};
	my $drivmask = drivermask($d);

	push(@dlist,
	     li(a({href=>$url}, $driv) .
		($d->{'descr'} ? ": " . $d->{'descr'} : ()),
		sprintf("$typestr, %d printers%s",
			$count,
			(numpackagesformask($drivmask) > 0 ?
			 downloadlinkline($drivmask,
					 ", <b>Driver packages available for:</b> ",
					 1) . " <font size=-3>(<a href=\"http://www.linux-foundation.org/en/OpenPrinting/Database/DriverPackages\">How to install</a>)</font>" : ""))));
    }
    
    my $dtype = "";
    $dtype = $driver_types{$type} if $type;

    push(@output,
	 start_html(-title => "OpenPrinting database - All Drivers",
		    -author => 'Till Kamppeter & Grant Taylor',
		    -BGCOLOR => 'white'),
	 navbar2('driver_list.cgi',
		 "All $dtype Drivers",
		 p(a({href=>"./"},"Home"),
		   ' <b>:</b> ',
		   a({href=>"database.html"}, 'Database'),
		   ' <b>:</b> ',
		   "Drivers"),
		p("This is our list of printer drivers for Linux and Unix. If you know of any driver not represented here, please let us know!"),
		ul(@dlist)),
	 end_html());

    # Write preparsed web page cache
    if ($cachepage) {
	print CFILE @output;
	close CFILE;
    } else {
	warn "unable to write $cachepage";
    }
}

print header(-expires=>'3m', -type=>'text/html; charset=UTF-8'), @output;

exit 0;

