
package driverpackages;

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(listpackagefiles listpackagefilesformask
		numpackagesformask packagesformask
		downloadlinkline drivermask driverinfobox %driver_types
		%archs);

use CGI qw/ :standard :html3 /;
use CGI dd, dt, dl, ul, li, escape, unescape, font, center;

use lib "@LIBDIR@";
use Foomatic::DB qw/normalizename/;

my $driverdir="@DRIVERDIR@";
my $driverdirurl="@DRIVERDIRURL@";

our %archs = ('noarch' => 'All architectures',
	      'any' => 'All architectures',
	      'all' => 'All architectures',
	      'i386' => 'x86 32 bit',
	      'i486' => 'x86 32 bit',
	      'i586' => 'x86 32 bit',
	      'i686' => 'x86 32 bit',
	      'x86_32' => 'x86 32 bit',
	      'ia32' => 'x86 32 bit',
	      'amd64' => 'x86 64 bit',
	      'x86_64' => 'x86 64 bit',
	      'ia64' => 'Itanium 64 bit',
	      'powerpc' => 'Power PC 32 bit',
	      'ppc' => 'Power PC 32 bit',
	      'ppc32' => 'Power PC 32 bit',
	      'ppc64' => 'Power PC 64 bit',
	      's390' => 'S390',
	      's390x' => 'S390x',
	      'src' => 'Source',
	      'source' => 'Source');

our @archorder = ('noarch', 'any', 'all', 'i386', 'i486', 'i586', 'i686', 'x86_32', 'ia32', 'amd64', 'x86_64', 'ia64', 'powerpc', 'ppc', 'ppc32', 'ppc64', 's390', 's390x', 'src', 'source');

our @pkgsysorder = ('rpm', 'deb');

our %driver_types = ('F' => 'Filter',
		     'V' => 'OpenPrinting Vector',
		     'I' => 'IJS',
		     'C' => 'CUPS Raster',
		     'P' => 'PostScript',
		     'U' => 'Ghostscript Uniprint',
		     'G' => 'Ghostscript built-in');

our $debsection="*";
our $lsexclude="/main/|/contrib/";

our @scopeorder = ('general', 'printer', 'scanner', 'fax', 'gui');

sub isarchitecture {
    # Check whether $file is for one of the architectures $archlist
    my ($archlist, $file) = @_;
    my $filearch;
    if ($file =~
	     m!^(|.*?/)([^/]+)-([^/-]+)-([^/-]+)lsb([^/-]+)\.([^\./-]+)\.rpm$!) {
	$filearch = $archs{lc($6)};
    } elsif ($file =~ m!RPMS/([^/]+)/!i) {
	$filearch = $archs{lc($1)};
    } elsif ($file =~ m!SRPMS/!i) {
        $filearch = $archs{'src'};
    } elsif ($file =~ m!_([^_]+)\.deb$!) {
	$filearch = $archs{lc($1)};
    } elsif ($file =~ m!debian/dists/[^/]+/[^/]+/binary-([^/]+)/!i) {
	$filearch = $archs{lc($1)};
    } elsif ($file =~ m!debian/dists/[^/]+/[^/]+/source[^/]*/!i) {
        $filearch = $archs{'src'};
    } else {
	return 0;
    }
    for my $arch (split(/\W+/, $archlist)) {
	return 1 if ($archs{lc($arch)} eq $filearch) ||
	    ($arch =~ /^(all)$/i) ||
	    (($arch =~ /^(bin)$/i) &&
	     ($filearch ne $archs{'src'})) ||
	    (($archs{lc($arch)} ne $archs{'src'}) &&
	     ($filearch eq $archs{'noarch'}));
    }
    return 0;
}

sub ispkgsys {
    # Check whether $file is for one of the package systems in $pkgsys
    my ($pkgsys, $file) = @_;
    my $filepkgsys;
    if ($file =~ m!\.rpm$!i) {
	$filepkgsys = "rpm";
    } elsif ($file =~ m!\.deb$!i) {
        $filepkgsys = "deb";
    } else {
	return 0;
    }
    for my $ps (split(/\W+/, $pkgsys)) {
	return 1 if ($ps eq $filepkgsys) ||
	    ($ps =~ /^(all)$/i);
    }
    return 0;
}

sub islsbversion {
    # Check whether $file is for one of the LSB versions in $lsbversion
    my ($lsbversion, $file) = @_;
    my $filelsbver;
    if ($file =~ m!lsb(\d+(\.\d+)*)!i) {
	$filelsbver = $1;
    } else {
	return 0;
    }
    for my $lsbver (split(/[^\d\.a-z]+/i, $lsbversion)) {
	return 1 if ($lsbver eq $filelsbver) ||
	    ($lsbver =~ /^(all)$/i);
    }
    return 0;
}

sub listpackagefiles {
    # Get info about downloadable driver packages (only locally hosted files)
    my ($pkgtype, $pkgsys, $lsbversion) = @a;
    $pkgtype = "bin" if !defined($pkgtype);
    $pkgsys = "all" if !defined($pkgsys);
    $lsbversion = "all" if !defined($lsbversion);
    return grep {
	islsbversion($lsbversion, $_);
    } grep {
	ispkgsys($pkgsys, $_);
    } grep {
        isarchitecture($pkgtype, $_);
    } split(/\s+/,
	    `cd $driverdir; ls -1 RPMS/*/*.rpm SRPMS/*.rpm debian/dists/*/$debsection/*/*.deb 2>/dev/null | grep -Ev \"$lsexclude\" 2>/dev/null`);
}

sub listpackagefilesformask {
    # Get info about downloadable driver packages for a given driver name/
    # URL/file mask
    my ($mask, $pkgtype, $pkgsys, $lsbversion) = @_; 
    $pkgtype = "bin" if !defined($pkgtype);
    $pkgsys = "all" if !defined($pkgsys);
    $lsbversion = "all" if !defined($lsbversion);
    my @res;
    foreach my $m (split(/;/, $mask)) {
	my ($p, $scope, $path, $files);
	$m =~ m!^(\(([^\)]*)\)|)(.*)$! || next;
	$scope = $2;
	$p = $3;
	if ($p =~ m!^(.*/)([^/]*)$!) {
	    $path = $1;
	    $files = $2;
	    $path = $driverdir . '/' . $path
		if $path !~ m!^((https?|ftp)\:/|)/!;
	} else {
            $path = $driverdir . '/';
            if ($p =~ /\.deb$/) {
		        $files = "debian/dists/*/$debsection/*/$p";
            } elsif ($p =~ /\.rpm$/) {
                $files = "RPMS/*/$p SRPMS/$p";
            } else {
                $files = "RPMS/*/${p}-[0-9]*.rpm SRPMS/${p}-[0-9]*.rpm " .
                    "debian/dists/*/$debsection/*/${p}[_-][0-9]*.deb";
            }
	}
	if ($path =~ m!^(https?|ftp)\://!) {
	    my $dirhtml = Foomatic::DB::getpage(undef, $path, 1);
	    if ($dirhtml) {
		$files =~ s/\?/[^\\\">]/sg;
		$files =~ s/\*/[^\\\">]*?/sg;
		$files =~ s/(\.rpm) +/$1\|/sg;
		$files =~ s/(\.deb) +/$1\|/sg;
		push(@res, map {
		    $_ = ($scope ? "($scope)" : "") . $path . $_;
		} grep {
		    islsbversion($lsbversion, $_);
		} grep {
		    ispkgsys($pkgsys, $_);
		} grep {
		    isarchitecture($pkgtype, $_);
		} ($dirhtml =~ /href=\"?(?![\?\&\/])($files)\"?/g));
	    }
	} else {
	    push(@res, map {
		my $e = $path . $_;
		$e =~ s/^$driverdir/$driverdirurl/;
		$e = "($scope)$e" if $scope;
		$e
	    } grep {
		islsbversion($lsbversion, $_);
	    } grep {
		ispkgsys($pkgsys, $_);
	    } grep {
		isarchitecture($pkgtype, $_);
	    } split(/\s+/,
		    `cd $driverdir; ls -1 $files 2>/dev/null | grep -Ev \"$lsexclude\" 2>/dev/null`));
	}
    }
    return @res;
}

sub numpackagesformask {
    # How many downloadable packages are available for this driver/file mask?
    my ($mask, $pkgtype, $pkgsys, $lsbversion) = @_; 
    $pkgtype = "bin" if !defined($pkgtype);
    $pkgsys = "all" if !defined($pkgsys);
    $lsbversion = "all" if !defined($lsbversion);
    return
	scalar(listpackagefilesformask($mask, $pkgtype, $pkgsys, $lsbversion));
}

sub packagesformask {
    # Make a list of available packages for a given driver/file mask
    my ($mask, $pkgtype, $pkgsys, $lsbversion) = @_; 
    $pkgtype = "bin" if !defined($pkgtype);
    $pkgsys = "all" if !defined($pkgsys);
    $lsbversion = "all" if !defined($lsbversion);
    my @packages =
	listpackagefilesformask($mask, $pkgtype, $pkgsys, $lsbversion);
    my @packageinfo = ();
    my %filesalreadythere;
    foreach my $package (@packages) {
	my ($scope, $url, $file, $basename, $version, $release, $lsbver,
	    $arch, $pkgsys);
	if ($package =~ m!^(\(([^\(\)]*)\)|)(([^\(\)]*?/|)(([^/]+?)-([^/-]+)-([^/-]+)lsb([^/-]+)\.([^\./-]+)\.(rpm)))$!) {
	    $scope = $2;
	    $url = $3;
	    $file = $5;
	    $basename = $6;
	    $version = $7;
	    $release = $8;
	    $lsbver = $9;
	    $arch = $10;
	    $pkgsys = $11;
	} elsif ($package =~ m!^(\(([^\(\)]*)\)|)(([^\(\)]*?/|)(([^/]+?)_([^_/-]+)-([^_/-]+)lsb([^_/-]+)_([^_/-]+)\.(deb)))$!) {
	    $scope = $2;
	    $url = $3;
	    $file = $5;
	    $basename = $6;
	    $version = $7;
	    $release = $8;
	    $lsbver = $9;
	    $arch = $10;
	    $pkgsys = $11;
	} else {
	    next;
	}
	next if $filesalreadythere{$file};
	$filesalreadythere{$file} = 1;
	my $packagedata = {};
	$scope ||= "general";
	my $realversion = $version;
	if ($release =~ /^0\.([^\.]+)\.\d+$/) {
	    my $extraversion = $1;
	    $extraversion =~ s/^(\d)/-$1/;
	    $realversion .= $extraversion;
	}
	$packagedata->{scope} = $scope;
	$packagedata->{basename} = $basename;
	$packagedata->{url} = $url;
	$packagedata->{file} = $file;
	$packagedata->{realversion} = $realversion;
	$packagedata->{version} = $version;
	$packagedata->{release} = $release;
	$packagedata->{lsbver} = $lsbver;
	$packagedata->{sortindex} = "${version}_${release}_${lsbver}";
	$packagedata->{arch} = $arch;
	$packagedata->{pkgsys} = lc($pkgsys);
	$packagedata->{text} = "$realversion (" . uc($pkgsys) .
	    " for LSB $lsbver)";
	push(@packageinfo, $packagedata);
    }
    return sort {
	foreach my $ar (@archorder) {
	    return -1 if (($archs{$a->{arch}} eq $archs{$ar}) &&
			  ($archs{$b->{arch}} ne $archs{$ar}));
	    return  1 if (($archs{$b->{arch}} eq $archs{$ar}) &&
			  ($archs{$a->{arch}} ne $archs{$ar}));
	}
	my $x = ($a->{basename} cmp $b->{basename});
	return $x if $x != 0;
	$x = ($a->{scope} cmp $b->{scope});
	return $x if $x != 0;
	$x = $b->{sortindex} cmp $a->{sortindex};
	return $x if $x != 0;
	foreach my $psys (@pkgsysorder) {
	    return -1 if ((lc($a->{pkgsys}) eq lc($psys)) &&
			  (lc($b->{pkgsys}) ne lc($psys)));
	    return  1 if ((lc($b->{pkgsys}) eq lc($psys)) &&
			  (lc($a->{pkgsys}) ne lc($psys)));
	}
	return 0;
    } @packageinfo;
}

sub downloadlinkline {
    # Generate a line of links to download the packages for a given driver
    # file mask
    # mode=0: All packages; mode=1: Newest package for each architecture
    my ($mask, $text, $mode, $pkgtype) = @_; 
    $pkgtype = "bin" if !defined($pkgtype);
    my @downloadpackages = packagesformask($mask, $pkgtype);
    my $driverdownload = "";
    if ($#downloadpackages >= 0) {
	$driverdownload = $text;
	my $oldarch = "";
	my $oldbasename = "";
	my $oldpkgsys = "";
	my $notfirst = 0;
	my %basenames;
	my %cmpbasenames;
	foreach $download (@downloadpackages) {
	    $basenames{$download->{basename}} = 1;
	    my $cmpbasename = $download->{basename};
	    $cmpbasename =~ s/^openprinting-(ppds-|)//;
	    $cmpbasenames{$cmpbasename} = 1;
	}
	my $numbasenames = scalar(keys %basenames);
	my $numcmpbasenames = scalar(keys %cmpbasenames);
	foreach $download (@downloadpackages) {
	    my $arch = $download->{arch};
	    my $basename = $download->{basename};
	    my $pkgsys = $download->{pkgsys};
	    my $lsbver = $download->{lsbver};
	    next if ($mode == 1) && ($archs{$arch} eq $archs{$oldarch}) &&
		($basename eq $oldbasename) && ($pkgsys eq $oldpkgsys); 
	    $driverdownload .= ", " if $notfirst;
	    $notfirst = 1;
	    if (($archs{$arch} ne $archs{$oldarch}) &&
		(($mode == 0) || ($numcmpbasenames > 1))) {
		if ($mode == 0) {
		    $driverdownload .= "<b>$archs{$arch}:</b> ";
		} else {
		    $driverdownload .= "$archs{$arch}: ";
		}
	    }
	    $driverdownload .= "$basename " if (($basename ne $oldbasename) &&
						($numcmpbasenames > 1) &&
						($mode == 0));
	    $driverdownload .= "<a href=\"$download->{url}\" type=\"" .
		(lc($pkgsys) eq "rpm" ? 
		 "application/x-redhat-package-manager" :
		 (lc($pkgsys) eq "deb" ?
		  "application/x-debian-package" : 
		  "application/octet-stream")) . "\">" .
		($mode == 0 ? $download->{text} :
		 ($numcmpbasenames > 1 ? $basename : $archs{$arch}) .
		 " (" . uc($pkgsys) . " for LSB $lsbver)") . "</a>";
	    $oldarch = $arch;
	    $oldbasename = $basename;
	    $oldpkgsys = $pkgsys;
	}
    }
    return $driverdownload;
}

sub graph {

    my ($value) = @_;
    my $color;
    if ($value <= 33) {
	$color = "red";
    } elsif ($value <= 66) {
	$color = "orange";
    } else {
	$color = "green";
    }
    my $units = int($value / 5.0);
    my $bar1;
    for (my $i = 0; $i < $units; $i ++) { $bar1 .= "|"; }
    my $bar2;
    for (my $i = $units; $i < 20; $i ++) { $bar2 .= "|"; }
    return font({-color=>$color}, $bar1) . font({-color=>"#ffffff"}, $bar2) .
	"&nbsp;&nbsp;" . $value;
}

sub drivermask {

    # Get file mask for downloadable packages

    my ($drvdata, $driver) = @_;
    $driver ||= $drvdata->{'name'};
    $driver ||= $drvdata->{'id'};
    $driver =~ s!^driver/!!;
    my $driverfilemask;
    if (defined($drvdata->{'packages'})) {
	for my $package (@{$drvdata->{'packages'}}) {
	    $driverfilemask .= ";" if $driverfilemask;
	    if ($package->{'scope'}) {
		my $u = $package->{'url'};
		$u =~ s/;/;($package->{'scope'})/g;
		$driverfilemask .= "($package->{'scope'})$u";
	    } else {
		$driverfilemask .= $package->{'url'};
	    }
	}
    }
    $driverfilemask =
	"$driver;openprinting-$driver;openprinting-ppds-$driver" if
	!$driverfilemask;
    return $driverfilemask;
}


sub driverinfobox {

    my ($drvdata, $printer, $prndrvdata, $readyppd, $foomaticppd, $pkgtype,
	$extracomment) = @_;
    $pkgtype = "bin" if !defined($pkgtype);

    my $driver = (defined($drvdata->{'name'}) ?
		  $drvdata->{'name'} : $prndrvdata->{'id'});
    $driver =~ s!^driver/!!;

    # Is the driver really supplied by the manufacturer of the given
    # printer (HP's HPIJS used on a Brother printer is not
    # manufacturer-supplied).
    $printer =~ /^([^\-]+)\-/;
    my $make = $1;
    $make =~ s/(\%20|_|\+)/ /g;
    my $manufacturersupplied = undef;
    if (defined($drvdata->{'manufacturersupplied'})) {
	if ($drvdata->{'manufacturersupplied'} =~ /^\s*[01]\s*$/) {
	    $manufacturersupplied = $drvdata->{'manufacturersupplied'};
	} elsif ($make) {
	    $manufacturersupplied = 
		($make =~ m,^($drvdata->{'manufacturersupplied'})$,i ?
		 "1" : "0");
	} else {
	    $manufacturersupplied = "1";
	}
    }   

    # Get file mask for downloadable packages
    my $driverfilemask = drivermask($drvdata, $driver);

    #Get info specific to the printer/driver combo
    my ($pd_data, $pd_comment) = (undef, undef);
    my ($excmaxresx, $excmaxresy, $exccolor, $exctext, $exclineart,
	$excgraphics, $excphoto, $excload, $excspeed) =
	    (undef, undef, undef, undef, undef, undef, undef, undef, undef);
    for (@{$drvdata->{'printers'}}) {
	next if (!$_->{'id'});
	if ($_->{'id'} eq $printer) {
	    $pd_data = $_;
	    $pd_comment = $_->{'comment'};
	    $excmaxresx = $_->{'excmaxresx'};
	    $excmaxresy = $_->{'excmaxresy'};
	    $exccolor = $_->{'exccolor'};
	    $exctext = $_->{'exctext'};
	    $exclineart = $_->{'exclineart'};
	    $excgraphics = $_->{'excgraphics'};
	    $excphoto = $_->{'excphoto'};
	    $excload = $_->{'excload'};
	    $excspeed = $_->{'excspeed'};
	    last;
	}
    }

    # Compute driver type string
    my $drvtype = $drvdata->{'type'};
    my $typestr = $driver_types{$drvtype};

    my $url = $drvdata->{'url'};

    my $nbspshortdescr = $drvdata->{'shortdescription'};
    $nbspshortdescr =~ s/(<\s*p\s*\/?\s*>\s*|<\s*br\s*\/?\s*>\s*)+/<br>/g;
    $nbspshortdescr =~ s/[\s\n]+/ /gs;
    $nbspshortdescr =~ s/^ //;
    $nbspshortdescr =~ s/ $//;
    $nbspshortdescr =~ s/ /&nbsp;/g;

    my $prndrvdatacomment = $prndrvdata->{'comment'};
    $prndrvdatacomment =~ s/(<\s*p\s*\/?\s*>\s*|<\s*br\s*\/?\s*>\s*)+/<br>/g;
    $prndrvdatacomment =~ s/[\s\n]+/ /gs;
    $prndrvdatacomment =~ s/^ //;
    $prndrvdatacomment =~ s/ $//;
    #$prndrvdatacomment =~ s/ /&nbsp;/g;

    $pd_comment =~ s/(<\s*p\s*\/?\s*>\s*|<\s*br\s*\/?\s*>\s*)+/<br>/g;
    $pd_comment =~ s/[\s\n]+/ /gs;
    $pd_comment =~ s/^ //;
    $pd_comment =~ s/ $//;
    #$pd_comment =~ s/ /&nbsp;/g;

    my $nbsplicense = $drvdata->{'license'};
    $nbsplicense =~ s/[\s\n]+/ /gs;
    $nbsplicense =~ s/^ //;
    $nbsplicense =~ s/ $//;
    $nbsplicense =~ s/ /&nbsp;/g;

    my $infobox =
	table({border => 0, bgcolor => "#d0d0d0",
	       cellpadding => 1, cellspacing => 0, width => 610},
	      Tr(td({colspan => 8},
		    table({border => 0, bgcolor => "#b0b0b0",
			   cellpadding => 5, cellspacing => 0,
			   width => "100%"},
			  Tr({valign=>"top", bgcolor => "#b0b0b0"},
			     td(font({-face=>"ms sans serif, helvetica"},
				     (!$printer ?
				      font({-size=>"+2"},
					   b(($url
					      ? a({href=>$url}, $driver) :
					      $driver))) :	
 				      font({-size=>"+2"},
					   b((defined($drvdata->{'name'}) ?
					      a({href=>"show_driver.cgi" .
						     "?driver=$driver" .
						     "&fromprinter=$printer"},
						$driver) : $driver))) .
				      ($url ?
				       font({-size=>"-2"},
					    "&nbsp;&nbsp(" .
					    a({href=>$url},
					      "driver home page") . ")") :
				       ())))))))),
	      #Tr({bgcolor => "#ffffff"},
	      #   td({colspan => 2},"&nbsp;")),
	      (defined($extracomment) &&
	       ($extracomment ne "") ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 600, colspan => 6},
		     font({-face=>"ms sans serif, helvetica"},
			  b($extracomment))),
		  td({width => 5}, "&nbsp;")) :
	       ()),
	      (defined($drvdata->{'obsolete'}) &&
	       ($drvdata->{'obsolete'} ne "0") ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 600, colspan => 6},
		     font({-face=>"ms sans serif, helvetica"},
			  b("This&nbsp;driver&nbsp;is&nbsp;obsolete.") .
			  ($drvdata->{'obsolete'} ne "1" ?
			   "&nbsp;Recommended&nbsp;replacement" .
			   "&nbsp;driver:&nbsp;" .
			   a({href => "show_driver.cgi?driver=" .
				  $drvdata->{'obsolete'} .
				  ($printer ? "&fromprinter=$printer" : ())},
			     $drvdata->{'obsolete'}) : ()))),
		  td({width => 5}, "&nbsp;")) :
	       ()),
	      ($drvdata->{'shortdescription'} ||
	       $prndrvdata->{'comment'} ||
	       $pd_comment ||
	       $drvdata->{'supplier'} ||
	       (defined($drvdata->{'manufacturersupplied'}) &&
		($drvdata->{'manufacturersupplied'} ne "0")) ||
	       $drvdata->{'license'} || $drvdata->{'licensetext'} || 
	       $drvdata->{'licenselink'} ||
	       defined($drvdata->{'free'}) ||
	       (defined($drvdata->{'patents'}) && $drvdata->{'patents'}) ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 600, colspan => 6},
		     font({-face=>"ms sans serif, helvetica"},
			  join("<br>",
			       ($drvdata->{'shortdescription'} ?
				$nbspshortdescr : ()),
			       ($prndrvdata->{'comment'} ?
				$prndrvdatacomment : ()),
			       ($pd_comment ?
				$pd_comment : ()),
			       ($drvdata->{'supplier'} ||
				(defined($manufacturersupplied) &&
				 ($manufacturersupplied ne "0")) ?
				font({-size => -2},
				     ($drvdata->{'supplier'} ?
				      ("Supplier:&nbsp;" .
				       $drvdata->{'supplier'} .
				       (defined($manufacturersupplied) &&
					($manufacturersupplied ne "0") ?
					($printer ?
					 b("&nbsp;(this&nbsp;printer's&nbsp;manufacturer)") :
					 b("&nbsp;(printer&nbsp;manufacturer)")) :
					())) :
				      (defined($manufacturersupplied) &&
				       ($manufacturersupplied ne "0") ?
				       ($printer ?
					b("Supplied&nbsp;by&nbsp;this&nbsp;" .
					  "printer's&nbsp;manufacturer") :
					b("Supplied&nbsp;by&nbsp;" .
					  "printer&nbsp;manufacturer")) : ()))) :
				()),
			       ($drvdata->{'license'} ||
				$drvdata->{'licensetext'} ||
				$drvdata->{'licenselink'} ||
				defined($drvdata->{'free'}) ?
				font({-size => -2},
				     ($drvdata->{'license'} ?
				      ("License:&nbsp;" .
				       "$nbsplicense&nbsp;" .
				       (!defined($drvdata->{'free'}) ||
					($drvdata->{'free'} ne "0") ?
					"(free&nbsp;software" :
					"(non-free&nbsp;software") .
				       ($drvdata->{'licensetext'} ||
					$drvdata->{'licenselink'} ?
					",&nbsp;" .
					a({href => "show_license.cgi?driver=" .
					       $driver . "&original=1"},
					  "show license text") .
					")" : ")")) :
				      (!defined($drvdata->{'free'}) ||
				       ($drvdata->{'free'} ne "0") ?
				       "This&nbsp;driver&nbsp;is&nbsp;" .
				       "free&nbsp;software" :
				       "This&nbsp;driver&nbsp;is&nbsp;" .
				       "non-free&nbsp;software") . 
				      ($drvdata->{'licensetext'} ||
				       $drvdata->{'licenselink'} ?
				       "&nbsp;(" .
				       a({href => "show_license.cgi?driver=" .
					      $driver . "&original=1"},
					 "show license text") .
				       ")." : "."))) :
				()),
			       ((defined($drvdata->{'patents'}) &&
				 $drvdata->{'patents'}) ?
				font({-size => -2},
				     "<b>&nbsp;&nbsp;" .
				     "This driver contains algorithms " .
				     "which are (possibly) patented",
				     ($drvdata->{'licensetext'} ||
				      $drvdata->{'licenselink'} ?
				      (" (See " .
				       a({href => "show_license.cgi?driver=" .
					      $driver . "&original=1"},
					 "license text") .
				       ").</b>") : ".</b>")) :
				())))),
		  td({width => 5}, "&nbsp;")) :
	       ()),
	      ($drvdata->{'supportcontacts'} > 0 ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 100, colspan => 1},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  "User&nbsp;support:&nbsp;")),
		  td({width => 500, colspan => 5},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  join("<br>",
			       map {
				   my $d = $_->{'description'};
				   $d =~ s/[\s\n]+/ /gs;
				   $d =~ s/^ //;
				   $d =~ s/ $//;
				   $d =~ s/ /&nbsp;/g;
				   (!$_->{'url'} &&
				    !$_->{'description'} ? () :
				    ($_->{'url'} ?
				     a({href=>$_->{'url'}}, $d) : $d) .
				    ($_->{'level'} ?
				     "&nbsp;($_->{'level'})" :
				     ""));
			       } @{$drvdata->{'supportcontacts'}}))),
		  td({width => 5}, "&nbsp;")) :
	       ()),
	      ($drvdata->{'drvmaxresx'} ||
	       $drvdata->{'drvmaxresy'} ||
	       defined($drvdata->{'drvcolor'}) ||
	       $excmaxresx || $excmaxresy || defined($exccolor) ||
	       $typestr ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 600, colspan => 6},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  join("&nbsp;&nbsp;",
			       ($drvdata->{'drvmaxresx'} ||
				$drvdata->{'drvmaxresy'} ||
				$excmaxresx || $excmaxresy ?
				("Max.&nbsp;rendering&nbsp;" .
				 "resolution:&nbsp;" .
				 ($excmaxresx ? $excmaxresx :
				  ($drvdata->{'drvmaxresx'} ?
				   $drvdata->{'drvmaxresx'} :
				   ($drvdata->{'drvmaxresy'} ?
				    $drvdata->{'drvmaxresy'} :
				    $excmaxresy))) . "x" .
				 ($excmaxresy ? $excmaxresy :
				  ($drvdata->{'drvmaxresy'} ?
				   $drvdata->{'drvmaxresy'} :
				   ($drvdata->{'drvmaxresx'} ?
				    $drvdata->{'drvmaxresx'} :
				    $excmaxresx))) . "dpi") : ()),
			       (defined($exccolor) ?
				($exccolor eq "0" ?
				 "Only&nbsp;monochrome&nbsp;output" :
				 "Color&nbsp;output") :
				(defined($drvdata->{'drvcolor'}) ?
				 ($drvdata->{'drvcolor'} eq "0" ?
				  "Only&nbsp;monochrome&nbsp;output" :
				  "Color&nbsp;output") : ())),
			       ($typestr ? 
				"Type:&nbsp;" . $typestr :
				())))),
		  td({width => 5}, "&nbsp;")) : ()),
	      (defined($drvdata->{'text'}) ||
	       defined($drvdata->{'lineart'}) ||
	       defined($drvdata->{'graphics'}) ||
	       defined($drvdata->{'photo'}) ||
	       defined($drvdata->{'load'}) ||
	       defined($drvdata->{'speed'}) ||
	       defined($exctext) || defined($exclineart) ||
	       defined($excgraphics) || defined($excphoto) ||
	       defined($excload) || defined($excspeed) ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 100},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  "Text:&nbsp;" .
			  "<br>Line&nbsp;art:&nbsp;")),
		  td({width => 100},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  (defined($exctext) ? graph($exctext) : 
			   (defined($drvdata->{'text'}) ?
			    graph($drvdata->{'text'}) : "Unknown")) .
			  "&nbsp;<br>" .
			  (defined($exclineart) ? graph($exclineart) : 
			   (defined($drvdata->{'lineart'}) ?
			    graph($drvdata->{'lineart'}) : "Unknown")) .
			  "&nbsp;")),
		  td({width => 100},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  "&nbsp;Graphics:&nbsp;" .
			  "<br>&nbsp;Photo:&nbsp;")),
		  td({width => 100},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  (defined($excgraphics) ? graph($excgraphics) : 
			   (defined($drvdata->{'graphics'}) ?
			    graph($drvdata->{'graphics'}) : "Unknown")) .
			  "&nbsp;<br>" .
			  (defined($excphoto) ? graph($excphoto) : 
			   (defined($drvdata->{'photo'}) ?
			    graph($drvdata->{'photo'}) : "Unknown")) .
			  "&nbsp;")),
		  td({width => 100},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  "&nbsp;System&nbsp;load:&nbsp;" .
			  "<br>&nbsp;Speed:&nbsp;")),
		  td({width => 100},
		     font({-face=>"ms sans serif, helvetica",
			   -size=>"-2"},
			  (defined($excload) ? graph($excload) : 
			   (defined($drvdata->{'load'}) ?
			    graph($drvdata->{'load'}) : "Unknown")) .
			  "<br>" .
			  (defined($excspeed) ? graph($excspeed) : 
			   (defined($drvdata->{'speed'}) ?
			    graph($drvdata->{'speed'}) : "Unknown")))),
		  td({width => 5}, "&nbsp;")) : ()),
	      ($foomaticppd || $readyppd ||
	       numpackagesformask($driverfilemask, $pkgtype) ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 100, colspan => 1},
		     font({-face=>"ms sans serif, helvetica"},
			  "<b>Download:&nbsp;</b>")),
		  td({width => 500, colspan => 5},
		     font({-face=>"ms sans serif, helvetica"},
			  join("<br>",
			       (numpackagesformask($driverfilemask, $pkgtype) ?
				downloadlinkline($driverfilemask,
						 "Driver&nbsp;packages: ",
						 ($printer ? 1 : 0),
						 $pkgtype) .
				" <font size=-3>(<a href=\"http://www.linux-foundation.org/en/OpenPrinting/Database/DriverPackages\">How to install</a>)</font>" :
				()),
			       ($foomaticppd || $readyppd ?
				"PPD&nbsp;file:&nbsp;" : ()) .
			       ($foomaticppd ?
				(a({href=>"$foomaticppd&show=1"},
				   "View&nbsp;PPD") .
				 ",&nbsp;" .
				 a({href=>"$foomaticppd&show=0"}, 
				   "directly&nbsp;download&nbsp;PPD")) : 
				($readyppd ? a({href=>"$readyppd"},
					       "Custom&nbsp;PPD") : 
				 ()))))),
		  td({width => 5}, "&nbsp;")) : ()) .
	      (defined($drvdata->{requires}) &&
	       (scalar(@{$drvdata->{requires}}) > 0) ?
	       Tr({valign=>"top"},
		  td({width => 5}, "&nbsp;"),
		  td({width => 100, colspan => 1},
		     font({-face=>"ms sans serif, helvetica"},
			  "<b>Dependencies:&nbsp;</b>")),
		  td({width => 500, colspan => 5},
		     font({-face=>"ms sans serif, helvetica"},
			  "To use this driver the following drivers " .
			  "need also to be installed: " .
			  join(", ",
			       map {
				   a({href=>"show_driver.cgi?driver=" .
					  $_->{'driver'}},
				     $_->{'driver'}) . 
				     ($_->{'version'} ?
				      " (version $_->{'version'})" : "");
			       } @{$drvdata->{requires}}))),
		  td({width => 5}, "&nbsp;")) : ()) .
	      Tr({valign=>"top"},
		 td({width => 610, colspan => 8}, 
		    font({-face=>"ms sans serif, helvetica",
			  -size=>"-5"}, "")))
	      );
    
    return $infobox;
}
