#! /usr/bin/perl

# TODO
# - Buttons to navigate in the file's history
# - History overview page?
# - Button to view histories of other printer names merged into this
#   entry
# - HTML checking of input fields, to avoid broken output, images from
#   external servers, active content, ...

my $debug = 0;

use Theme qw/navbar2/;

use lib "@LIBDIR@";
use Foomatic::Defaults;

use CGI qw/ :standard :html3 /;
use CGI escape, unescape;

use PHTDB qw/$hcache $cachedir/;

my $active = @PRINTEREDITACTIVE@;

my $pwd = `pwd`;
$pwd =~ s/[\s\r\n]+$//;

my $wdir = "@WIKISHDIR@";
my $foomaticpath="$pwd/@FOOUTILDIR@";
my $navloc = "show_printer.cgi";
#my $navbarcolor   = '#72D397';
#my $navbarcolorlt = '#8affb6';
my $navbarcolor   = '#dfdfdf';
my $navbarcolorlt = '#efefef';
my $query = new CGI;
my ($action) = $query->param('action');
my ($tag) = $query->url_param('tag');
my $status;
my $oidbeforemerge;

my $basev = $query->param('basev');
die if $basev =~ m![^\d\.]!; # not a filename, a cvs version number

# What are legal values for tag?
die unless $tag =~ m!^[\w\d\_\-]*$!;

# $action is only compared to constant strings (albeit via regex) so
# can get by without scrubbing.  But just in case.
$action =~ s![^a-zA-Z]!!g;

my $username = $ENV{'REMOTE_USER'};
my $hostname = $ENV{'REMOTE_ADDR'};

# Connect to the database
my $db = new PHTDB;

# record for printer data
my $p;
my $oid;
my $oidindb = 0;

# Make and model from command line
my ($make,
    $model) = ($query->param('make'),
	       $query->param('model'));
if ((!$make) && (my $mm = $query->param('mmake'))) {
    $make = $mm;
}

# Do not allow '--' in manufacturer and model name
$make =~ s/-+//g;
$model =~ s/-+/-/g;

# Clean manufacturer and model name somewhat
$make =~ s/^[\s_]+//;
$make =~ s/[\s_]$//;
$model =~ s/^[\s_]+//;
$model =~ s/[\s_]$//;

# Do we want to have a new entry?
my $newentry = $query->param('newentry');

# Has the user supplied his e-mail address?
my $email = $query->param('email');

if ($make && $model) {
    # Derive printer ID from make and model
    $make =~ s/[\.\-]//g;
    $oid = "$make-$model";
    $oid =~ s/\%20/_/g;
    $oid =~ s/[^\w\-\+]/_/g;
    $oid =~ s/_+/_/g;
    $oid =~ s/^_+//g;
    $oid =~ s/_+$//g;
    $oid =~ s/\+/plus/g;
    if (($oid eq '-') || ($oid =~ /^\-/) || ($oid =~ /\-$/) ||
	($oid !~ /\w/)) {
	# make/model has no valid characters, discard ID
	$oid = '';
    } else {
	# Make sure that no printer ID is an ID of the former PostGreSQL
	# database or a somehow else obsolete ID
	$oid = Foomatic::DB::translate_printer_id($oid);
    }
}

my $r = $query->param('recnum');
$oid = $r if $r;
if ($oid && !$make) {
    $oid =~ /^([^-]+)-/;
    $make = $1;
    $make =~ s/_/ /g;
}
if ($oid && !$model) {
    $oid =~ /^[^-]+-(.*)$/;
    $model = $1;
    $model =~ s/_/ /g;
}

my $urlmake = $make;
$urlmake =~ s/ /\%20/g;
my $idmake = $make;
$idmake =~ s/ /_/g;

my $fromwhere = $query->param('fromwhere');
if (!$fromwhere) {
    if ($oid) {
	$fromwhere = "show_printer.cgi?recnum=$oid";
    } elsif ($make) {
	$fromwhere = "printer_list.cgi?make=$urlmake";
    } else {
	$fromwhere = "printer_list.cgi";
    }
}

if (!$active) {
    # Tell the user that the printer input facility is currently deactivated
    print header(-expires=>'-15m', -type=>'text/html; charset=UTF-8');
    print start_html(-BGCOLOR=>"white",
		     -title=>($make && $model ?
			      "Editing printer $make $model" :
			      "Adding new printer"));
    print navbar2($navloc,
		  "Printer information input page",
		  p(a({href=>$fromwhere},
		      "Return"),
		    "to the Foomatic printer compatibility database."),
		  p(font({color=>"red"}, b("The printer input/editing facility is temporarily deactivated, please try again later."))));
    print end_html();
    exit 0;
}

my ($url, $resolution_x, $resolution_y, $color, $ascii);
my ($type, $refill, $func, $pjl, $notes);
my $postscript = 0;
my $postscript_level = '';
my $pdf = 0;
my $pdf_level = '';
my $pcl = 0;
my $pcl_level = '';
my $escp = 0;
my $escp_level = '';
my $escp2 = 0;
my $escp2_level = '';
my $hpgl2 = 0;
my $hpgl2_level = '';
my $tiff = 0;
my $tiff_level = '';
my $lips = 0;
my $lips_level = '';
my $proprietary = 0;
my ($ppdurl, $recdriver);
my $dnumber = 0;
my @dnames = ();
my @dlinks = ();
my @dcomments = ();
my ($general_ieee, $general_mdl, $general_mfg, $general_des, $general_cmd);
my ($par_ieee, $par_mdl, $par_mfg, $par_des, $par_cmd);
my ($usb_ieee, $usb_mdl, $usb_mfg, $usb_des, $usb_cmd);
my $snmp_des;
my $contrib_url;

if ($action !~ /(Add|Save|Exit|Delete)/i) {
    loaddata();
}

if ($action =~ /Undelete/i) {
    chdir $wdir;
    if (! $oid) {
	my $deletedfiles = `ls -1 $wdir/editprinter\:*--deleted,v`;
	my @deletedoids = map {
	    m!^$wdir/editprinter:(.*)--deleted,v$!
	} split('\n', $deletedfiles);
	my @links;
	for my $i (@deletedoids) {
	    my ($mk, $md);
	    $i =~ /^([^-]+)-(.*)$/;
	    $mk = $1;
	    $md = $2;
	    next if !$mk or !$md;
	    $mk =~ s/_/ /g;
	    $md =~ s/_/ /g;
	    push(@links, "<li><a href=\"edit_printer.cgi?action=Undelete&recnum=$i\">$mk&nbsp;$md</a></li>\n");
	}
	print header(-expires=>'-15m', -type=>'text/html; charset=UTF-8');
	print start_html(-BGCOLOR=>"white",
			 -title=>($make && $model ?
				  "Editing printer $make $model" :
				  "Adding new printer"));
	print navbar2($navloc,
		      "Printer information input page",
		      p(a({href=>"printer_list.cgi"},
			  "Return"),
			"to the Foomatic printer compatibility database."),
		      p("The following printer entries were deleted. Click the entry which you want to undelete."),
		      "<ul>", @links, "</ul>");
	print end_html();
	exit 0;
    } elsif (rename "editprinter:$oid--deleted,v", "editprinter:$oid,v") {
	# Rename Wikish RCS file so that printer gets undeleted
	loaddata();
	$status = "Entry for the $make $model successfully undeleted";
	# Do not exit, show message to user
	$action =~ s/Exit//ig;
	# Delete out-of-sync cache files
	if (-d $hcache) {
	    unlink("$hcache/show-printer-recnum-$oid.html");
	    unlink("$hcache/printer-list-make-$idmake.html");
	    unlink("$hcache/printer-list-make-Anyone.html");
	    unlink("$hcache/printer-list-make-.html");
	}
	if (-d "$cachedir/pcache") {
	    unlink("$cachedir/pcache/overview.perl");
	}
    } else {
	$status = "Could not undelete an entry for the $make $model, there was no such entry deleted before";
	# Do not exit, show error message to user
	$action =~ s/Exit//ig;
    }
}

if (($action =~ /Save/i) || ($action =~ /Add/i)) {

    # Derive printer ID from make and model
    $make = $query->param('backupmake') if !$make;
    $model = $query->param('backupmodel') if !$model;
    $make =~ s/[\.\-]//g;
    $oid = "$make-$model";
    $oid =~ s/\%20/_/g;
    $oid =~ s/[^\w\-\+]/_/g;
    $oid =~ s/_+/_/g;
    $oid =~ s/^_+//g;
    $oid =~ s/_+$//g;
    $oid =~ s/\+/plus/g;

    if (($oid eq '-') || ($oid =~ /^\-/) || ($oid =~ /\-$/) ||
	($oid !~ /\w/)) {
	# make/model has no valid characters, restore old values.
	$make = $query->param('backupmake');
	$model = $query->param('backupmodel');
	$oid = $query->param('backupid');
    } else {
	# Make sure that no printer ID is an ID of the former PostGreSQL
	# database or a somehow else obsolete ID
	$oid = Foomatic::DB::translate_printer_id($oid);
    }

    $urlmake = $make;
    $urlmake =~ s/ /\%20/g;
    $idmake = $make;
    $idmake =~ s/ /_/g;

    # Check whether the current printer ID exists in the Foomatic database
    if ($db->printer_exists($oid)) {
	$oidindb = 1;
    }

    # Fill variables from the command line
    $url = $query->param('url');
    $resolution_x = $query->param('resolution_x');
    $resolution_y = $query->param('resolution_y');
    $color = $query->param('color');
    $ascii = $query->param('ascii');
    $type = $query->param('type');
    $refill = $query->param('refill');
    $func = $query->param('func');
    $pjl = $query->param('pjl');
    $notes = $query->param('notes');

    $postscript = $query->param('postscript');
    $postscript_level = $query->param('postscript_level');
    $pdf = $query->param('pdf');
    $pdf_level = $query->param('pdf_level');
    $pcl = $query->param('pcl');
    $pcl_level = $query->param('pcl_level');
    $escp = $query->param('escp');
    $escp_level = $query->param('escp_level');
    $escp2 = $query->param('escp2');
    $escp2_level = $query->param('escp2_level');
    $hpgl2 = $query->param('hpgl2');
    $hpgl2_level = $query->param('hpgl2_level');
    $tiff = $query->param('tiff');
    $tiff_level = $query->param('tiff_level');
    $lips = $query->param('lips');
    $lips_level = $query->param('lips_level');
    $proprietary = $query->param('proprietary');

    $ppdurl = $query->param('ppdurl');

    $recdriver = $query->param('recdriver');

    # Assemble the list-o-driver information
    my $i;
    $dnumber = $query->param('dnumber');
    my $newdriver = $query->param("dname$dnumber");
    for ($i = 0; $i < $dnumber; $i ++) {
	$dname = $query->param("dname$i");
	$dname = $newdriver if ($newdriver && ($i == $dnumber - 1));
	# User didn't add a new driver, so the name of the new driver is empty
	next if !$dname;
	# User has unchecked this driver
	next if ($query->param("dactive$i") ne "on");
	push(@dnames, $dname);
	my $comment = $query->param("dcomment$i");
	push(@dcomments, $comment);
	my $drvdata = $db->get_driver($dname);
	my $lnk = $drvdata->{'url'};
	push(@dlinks, $lnk);
    }
    $dnumber = $#dnames + 1;

    $general_ieee = $query->param('general_ieee');
    $general_mdl = $query->param('general_mdl');
    $general_mfg = $query->param('general_mfg');
    $general_des = $query->param('general_des');
    $general_cmd = $query->param('general_cmd');

    $par_ieee = $query->param('par_ieee');
    $par_mdl = $query->param('par_mdl');
    $par_mfg = $query->param('par_mfg');
    $par_des = $query->param('par_des');
    $par_cmd = $query->param('par_cmd');

    $usb_ieee = $query->param('usb_ieee');
    $usb_mdl = $query->param('usb_mdl');
    $usb_mfg = $query->param('usb_mfg');
    $usb_des = $query->param('usb_des');
    $usb_cmd = $query->param('usb_cmd');

    $snmp_des = $query->param('snmp_des');

    $contrib_url = $query->param('contrib_url');

    $oidbeforemerge = $query->param('oidbeforemerge');

}

# Pre-fill the "Recommended driver" field with a reasonable value
$recdriver = $dnames[0] if $dnumber == 1; 
$recdriver = "" if !Foomatic::DB::member($recdriver, @dnames);

# Fill all variables for merging
my ($make_orig, $model_orig) = ($make, $model);
my ($url_orig, $resolution_x_orig, $resolution_y_orig, $color_orig,
    $ascii_orig) =
    ($url, $resolution_x, $resolution_y, $color, $ascii);
my ($type_orig, $refill_orig, $func_orig, $pjl_orig, $notes_orig) =
    ($type, $refill, $func, $pjl, $notes);
my $postscript_orig = $postscript;
my $postscript_level_orig = $postscript_level;
my $pdf_orig = $pdf;
my $pdf_level_orig = $pdf_level;
my $pcl_orig = $pcl;
my $pcl_level_orig = $pcl_level;
my $escp_orig = $escp;
my $escp_level_orig = $escp_level;
my $escp2_orig = $escp2;
my $escp2_level_orig = $escp2_level;
my $hpgl2_orig = $hpgl2;
my $hpgl2_level_orig = $hpgl2_level;
my $tiff_orig = $tiff;
my $tiff_level_orig = $tiff_level;
my $lips_orig = $lips;
my $lips_level_orig = $lips_level;
my $proprietary_orig = $proprietary;
my ($ppdurl_orig, $recdriver_orig) = ($ppdurl, $recdriver);
my $dnumber_orig = $dnumber;
my @dnames_orig = @dnames;
my @dlinks_orig = @dlinks;
my @dcomments_orig = @dcomments;
my ($general_ieee_orig, $general_mdl_orig, $general_mfg_orig,
    $general_des_orig, $general_cmd_orig) =
    ($general_ieee, $general_mdl, $general_mfg, $general_des, $general_cmd);
my ($par_ieee_orig, $par_mdl_orig, $par_mfg_orig, $par_des_orig,
    $par_cmd_orig) =
    ($par_ieee, $par_mdl, $par_mfg, $par_des, $par_cmd);
my ($usb_ieee_orig, $usb_mdl_orig, $usb_mfg_orig, $usb_des_orig,
    $usb_cmd_orig) =
    ($usb_ieee, $usb_mdl, $usb_mfg, $usb_des, $usb_cmd);
my $snmp_des_orig = $snmp_des;
my $contrib_url_orig = $contrib_url;

if ($action =~ /^Delete/i) {
    # Rename Wikish RCS file so that printer can get undeleted
    chdir $wdir;
    unlink("editprinter:$oid--deleted,v");
    if (rename "editprinter:$oid,v", "editprinter:$oid--deleted,v") {
	# Delete out-of-sync cache files
	if (-d $hcache) {
	    unlink("$hcache/show-printer-recnum-$oid.html");
	    unlink("$hcache/printer-list-make-$idmake.html");
	    unlink("$hcache/printer-list-make-Anyone.html");
	    unlink("$hcache/printer-list-make-.html");
	}
	if (-d "$cachedir/pcache") {
	    unlink("$cachedir/pcache/overview.perl");
	}
	# Tell user that deletion was successful
	print header(-expires=>'-15m', -type=>'text/html; charset=UTF-8');
	print start_html(-BGCOLOR=>"white",
			 -title=>($make && $model ?
				  "Editing printer $make $model" :
				  "Adding new printer"));
	print navbar2($navloc,
		      "Printer information input page",
		  p(a({href=>"printer_list.cgi"},
		      "Return"),
		    "to the Foomatic printer compatibility database."),
		  p(font({color=>"red"}, b("The printer entry for the $make $model was successfully deleted!"))),
		      $query->startform(-method=>"POST",
					-action=>"edit_printer.cgi"),
		      "<input type=\"hidden\" name=\"recnum\" value=\"$oid\">\n",
		      "<input type=submit name=action value=\"Undelete\">\n",
		      $query->endform());
	print end_html();
	exit 0;
    } else {
	$status = "Could not delete this printer entry";
	# Do not exit, show error message to user
	$action =~ s/Exit//ig;
    }
}

# Generate XML file
my $warnings = 0;
my $fatals = 0;
if ($action =~ /Save/i) {

    if ($db->printer_exists($oid)) {
	# Printer ID already exists
	$status = "This printer already exists in the Foomatic database. It cannot be overwritten by a user-contributed entry. If you want to contribute information to the existing printer entry, post into the \"User Notes\" section on the <a href=\"show_printer.cgi?recnum=$oid\">printer's page</a>. If you want to report bugs in the <a href=\"show_printer.cgi?recnum=$oid\">entry for this printer</a>, post your bug report on <a href=\"http://bugs.linux-foundation.org/\">our bug tracker</a>, product \"OpenPrinting\", component \"foomatic-db\". If you really want to add a new printer, check the spelling of manufacturer and model name.";
	# Do not exit, show error message to user
	$action =~ s/Exit//ig;
	$fatals = 1;
    } else {
	my $xml = "<printer id=\"printer/" . h($oid) . "\">\n";
	$xml .= "  <make>" . h($make) . "</make>\n";
	$xml .= "  <model>" . h($model) . "</model>\n";
	$xml .= "  <mechanism>\n";
	$xml .= ($type ? "    <$type />\n" : "    <!--no \"type\"-->\n");
	$xml .= ($color eq "on" ? 
		 "    <color />\n" : "    <!--not \"color\"-->\n");
	$xml .= "    <resolution>\n";
	$xml .= "      <dpi>\n";
	$xml .= ($resolution_x ? 
		 "        <x>" . h($resolution_x) . "</x>\n" :
		 "        <!--no \"x\"-->\n");
	$xml .= ($resolution_y ? 
		 "        <y>" . h($resolution_y) . "</y>\n" :
		 "        <!--no \"y\"-->\n");
	$xml .= "      </dpi>\n";
	$xml .= "    </resolution>\n";
	$xml .= "    <consumables>\n";
	$xml .= ($refill ? 
		 "      <comments>\n        <en>" .
		 h(fixhtml($refill)) . "</en>\n      </comments>\n" :
		 "      <!--no \"comments\"-->\n");
	$xml .= "      <!--one or more \"partno\" elements-->\n";
	$xml .= "    </consumables>\n";
	$xml .= "  </mechanism>\n";
	$xml .= ($url ? 
		 "  <url>" . h($url) . "</url>\n" :
		 "  <!--no \"url\"-->\n");
	$xml .= "  <lang>\n";
	$xml .= ($postscript eq "on" ? 
		 "    <postscript" .
		 ($postscript_level ? " level=\"" . h($postscript_level) . "\"" : "") . 
		 ">\n      " .
		 ($ppdurl ?
		  "<ppd>" . h($ppdurl) . "</ppd>\n" :
		  "<!--no URL for manufacturer's \"ppd\" file-->\n") .
		 "    </postscript>\n" : "");
	$xml .= ($pdf eq "on" ? 
		 "    <pdf" .
		 ($pdf_level ? " level=\"" . h($pdf_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($lips eq "on" ? 
		 "    <lips" .
		 ($lips_level ? " level=\"" . h($lips_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($pcl eq "on" ? 
		 "    <pcl" .
		 ($pcl_level ? " level=\"" . h($pcl_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($escp eq "on" ? 
		 "    <escp" .
		 ($escp_level ? " level=\"" . h($escp_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($escp2 eq "on" ? 
		 "    <escp2" .
		 ($escp2_level ? " level=\"" . h($escp2_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($hpgl2 eq "on" ? 
		 "    <hpgl2" .
		 ($hpgl2_level ? " level=\"" . h($hpgl2_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($tiff eq "on" ? 
		 "    <tiff" .
		 ($tiff_level ? " level=\"" . h($tiff_level) . "\"" : "") . 
		 " />\n" : "");
	$xml .= ($proprietary eq "on" ? 
		 "    <proprietary />\n" : "");
	$xml .= ($pjl eq "on" ?
		 "    <pjl />\n" : "");
	$xml .= ($ascii eq "on" ?
		 "    <text>\n      <charset>us-ascii</charset>\n    </text>\n" :
		 "");
	$xml .= "  </lang>\n";
	$xml .= "  <autodetect>\n";
	$xml .= ($general_ieee || $general_mdl || $general_mfg ||
		 $general_des || $general_cmd ?
		 "    <general>\n" : "");
	$xml .= ($general_ieee ?
		 "      <ieee1284>" . h($general_ieee) . "</ieee1284>\n" : "");
	$xml .= ($general_mfg ?
		 "      <manufacturer>" . h($general_mfg) .
		 "</manufacturer>\n" : "");
	$xml .= ($general_mdl ?
		 "      <model>" . h($general_mdl) . "</model>\n" : "");
	$xml .= ($general_des ?
		 "      <description>" . h($general_des) .
		 "</description>\n" : "");
	$xml .= ($general_cmd ?
		 "      <commandset>" . h($general_cmd) . "</commandset>\n" : "");
	$xml .= ($general_ieee || $general_mdl || $general_mfg ||
		 $general_des || $general_cmd ?
		 "    </general>\n" : "");
	$xml .= ($par_ieee || $par_mdl || $par_mfg || $par_des || $par_cmd ?
		 "    <parallel>\n" : "");
	$xml .= ($par_ieee ?
		 "      <ieee1284>" . h($par_ieee) . "</ieee1284>\n" : "");
	$xml .= ($par_mfg ?
		 "      <manufacturer>" . h($par_mfg) . "</manufacturer>\n" : "");
	$xml .= ($par_mdl ? "      <model>" . h($par_mdl) . "</model>\n" : "");
	$xml .= ($par_des ?
		 "      <description>" . h($par_des) . "</description>\n" : "");
	$xml .= ($par_cmd ?
		 "      <commandset>" . h($par_cmd) . "</commandset>\n" : "");
	$xml .= ($par_ieee || $par_mdl || $par_mfg || $par_des || $par_cmd ?
		 "    </parallel>\n" : "");
	$xml .= ($usb_ieee || $usb_mdl || $usb_mfg || $usb_des || $usb_cmd ?
		 "    <usb>\n" : "");
	$xml .= ($usb_ieee ?
		 "      <ieee1284>" . h($usb_ieee) . "</ieee1284>\n" : "");
	$xml .= ($usb_mfg ?
		 "      <manufacturer>" . h($usb_mfg) . "</manufacturer>\n" : "");
	$xml .= ($usb_mdl ? "      <model>" . h($usb_mdl) . "</model>\n" : "");
	$xml .= ($usb_des ?
		 "      <description>" . h($usb_des) . "</description>\n" : "");
	$xml .= ($usb_cmd ?
		 "      <commandset>" . h($usb_cmd) . "</commandset>\n" : "");
	$xml .= ($usb_ieee || $usb_mdl || $usb_mfg || $usb_des || $usb_cmd ?
		 "    </usb>\n" : "");
	$xml .= ($snmp_des ? "    <snmp>\n" : "");
	$xml .= ($snmp_des ?
		 "      <description>" . h($snmp_des) . "</description>\n" : "");
	$xml .= ($snmp_des ? "    </snmp>\n" : "");
	$xml .= "  </autodetect>\n";
	$xml .= "  <functionality>$func</functionality>\n";
	$xml .= ($recdriver ?
		 "  <driver>" . h($recdriver) . "</driver>\n" :
		 "  <!--no recommended \"driver\"-->\n");
	$xml .= ($dnumber > 0 ? "  <drivers>\n" : "");
	for (my $i = 0; $i < $dnumber; $i ++) {
	    $xml .= "    <driver>\n";
	    $xml .= "      <id>" . h($dnames[$i]) . "</id>\n";
	    $xml .= ($dcomments[$i] ? 
		     "      <comments>\n        <en>" .
		     h(fixhtml($dcomments[$i])) . "</en>\n      </comments>\n" :
		     "      <!--no \"comments\"-->\n");
	    $xml .= "    </driver>\n";
	}
	$xml .= ($dnumber > 0 ? "  </drivers>\n" : "");
	$xml .= ($contrib_url ? 
		 "  <contrib_url>" . h($contrib_url) . "</contrib_url>\n" :
		 "  <!--no \"contrib_url\"-->\n");
	$xml .= "  <unverified />\n";
	$xml .= ($notes ? 
		 "  <comments>\n    <en>" .
		 h(fixhtml($notes)) . "</en>\n  </comments>\n" :
		 "  <!--no \"comments\"-->\n");
	$xml .= "</printer>";
	
	#print header(-expires=>'-15m', -type=>'text/xml; charset=UTF-8'), $xml;
	#exit 0;
	
	my $now = time();
	my $comment = sprintf("ip=%s", $ENV{'REMOTE_ADDR'});
	$comment .= ", email=$email" if $email;
	
	# Make and model must be filled in at least
	if (($make !~ /\S/) || ($model !~ /\S/)) {
	    $status = "Data can only be saved when at least the \"Manufacturer\" and \"Model\" fields are filled in!";
	    # Do not exit, show error message to user
	    $action =~ s/Exit//ig;
	    $fatals = 1;
	} else {

	    # Content checking

	    # Collect status messages
	    my @smessages;

	    # Prevent abuse of the printer input interface for spam or for
	    # cheating search machines
	    foreach my $text ($make, $model, $resolution_x, $resolution_y,
			      $refill, @dnames, @dcomments,
			      $notes) {
		my ($spam, $spammsg) = text_is_spam ($text);
		if ($spam) {
		    push(@smessages,
			 "Spam-typical entry found: $spammsg in $text\n");
		    $fatals = 1;
		}
		#die "SPAM: $spammsg in $text\n" if $spam;
	    }

	    # No comment, nor a recommended driver, Paperweight, nothing
	    # except make/model filled in
	    if (($recdriver !~ /\S/) && ($func eq "F") &&
		($url !~ /\S/) && ($contrib_url !~ /\S/) &&
		($notes !~ /\S/) && ($type !~ /\S/) &&
		($resolution_x !~ /\S/) && ($resolution_y !~ /\S/) &&
		($email !~ /\S/) && ($dnumber == 0)) {
		push(@smessages,
		     "Please do not simply post only manufacturer and model name of the printer which you have. Make the entry useful by telling something about this printer, especially how well it works with Linux or Unix.");
		$fatals = 1;
	    }

	    # Manufacturer name not a real manufacturer
	    if ($make !~ /[A-Za-z]/) {
		push(@smessages,
		     "The manufacturer name does not contain any letter, are you sure that you have entered it correctly?");
		$fatals = 1;
	    }
	    if ($make !~ /^[A-Za-z]/) {
		push(@smessages,
		     "The manufacturer name does not start with a letter, are you sure that you have entered the manufacturer name here and not the model name?");
		$fatals = 1;
	    }
	    if ($make =~ /(deskjet|laserjet|designjet|officejet|photosmart|stylus|aculaser|stylewriter|imagewriter|deskwriter|laserwriter|pixus|pixma|selphy|imagerunner|\bbjc\b|\bbj\b|\blbp\b|\bhl\b|\bdcp\b|\bmfc\b|docuprint|docupage|phaser|workcentre|homecentre|optra|jetprinter|magicolor|pageworks|pagepro|aficio|varioprint|okipage|microline)/) {
		push(@smessages,
		     "The manufacturer name contains a common expression for model names, are you sure that you have entered the manufacturer name here and not the model name?");
		$fatals = 1;
	    }
	    if ($make =~ m!(\:\/\/|\bwww\b|href\s*\=)!i) {
		push(@smessages,
		     "The manufacturer name seems to be an URL. Did you use the correct field for entering the web link to the printer manufacturer?");
		$fatals = 1;
	    }
	    if ($make =~ /^(None)$/i) {
		push(@smessages,
		     "The manufacturer name seems not to be of a real manufacturer. Did you type it correctly?");
		$fatals = 1;
	    }

	    # Make and model are similar to an existing one?
	    my $smake = simplifystr($make);
	    my @mfgs = sort { Foomatic::DB::normalizename($a) cmp
			      Foomatic::DB::normalizename($b) } ($db->get_makes());
	    my @smfgs = grep { $smake eq simplifystr($_) } @mfgs;
	    my $tmpmake;
	    if (($#smfgs >= 0) &&
		(!Foomatic::DB::member($make, @smfgs))) {
		$tmpmake = $smfgs[0];
		push(@smessages,
		     "The manufacturer name you have entered is similar to \"" .
		     join("\", \"", @smfgs) .
		     "\". Are you sure you want to leave the manufacturer name this way? Note that manufacturer names are case-sensitive.");
		$fatals = 1;
	    } else {
		$tmpmake = $make;
	    }
		    
	    my $smodel = simplifystr($model);
	    my @mdls = sort { Foomatic::DB::normalizename($a) cmp
			      Foomatic::DB::normalizename($b) } ($db->get_models_by_make($tmpmake));
	    my @smdls = grep { $smodel eq simplifystr($_) } @mdls;
	    if (($#smdls >= 0) &&
		(!Foomatic::DB::member($model, @smdls))) {
		push(@smessages,
		     "The model name you have entered is similar to \"" .
		     join("\", \"", @smdls) .
		     "\". Are you sure you want to leave the model name this way? Note that model names are case-sensitive.");
		$fatals = 1;
	    }

	    # Look up models similar to the entered one in the database
	    # to avoid that the user enters a printer as a new model
	    # if it already exists
	    my @similarmodels = grep {
		/\S/
	    } map {
		my $entry = $db->get_printer($_);
		if ($entry->{'noxmlentry'}) {
		    $_ = '';
		} else {
		    $_ = $entry->{'make'} . ' ' . $entry->{'model'};
		}
	    } $db->find_printer("$make $model", 2, 1);
	    if (@similarmodels) {
		@similarmodels = @similarmodels[0..9] if
		    @similarmodels > 10;
		my $smlist = join("\", \"", @similarmodels);
		push(@smessages,
		     "Are you sure you entered manufacturer and model name correctly? The following similar models were found in the database: \"$smlist\"");
		$warnings = 1;
	    }

	    # Manufacturer name repeated in model name
	    my $qsmake = quotemeta($smake);
	    if ($smodel =~ /^$qsmake/i) {
		push(@smessages,
		     "Do not repeat the manufacturer's name in the field for the model name.");
		$fatals = 1;
	    }

	    # Manufacturer and model name do not correspond to each other
	    if ((($smodel =~ /(deskjet|laserjet|designjet|officejet|photosmart)/i) &&
		 ($smake ne "hp")) ||
		(($smodel =~ /(stylus|aculaser)/i) &&
		 ($smake ne "epson")) ||
		(($smodel =~ /(stylewriter|imagewriter|deskwriter|laserwriter)/i) &&
		 ($smake ne "apple")) ||
		(($smodel =~ /(pixus|pixma|selphy|imagerunner|\bbjc\b|\bbj\b|\blbp\b)/i) &&
		 ($smake ne "canon")) ||
		(($smodel =~ /(\bhl\b|\bdcp\b|\bmfc\b)/i) &&
		 ($smake ne "brother")) ||
		(($smodel =~ /(docuprint|docupage|phaser|workcentre|homecentre)/i) &&
		 ($smake !~ /^((fuji[\s-]*|)xerox|tektronix)$/i)) ||
		(($smodel =~ /(optra|jetprinter)/i) &&
		 ($smake !~ /^(lexmark)$/i)) ||
		(($smodel =~ /(magicolor|pageworks|pagepro)/i) &&
		 ($smake !~ /^(konica|(konica\s*|)minolta|qms)$/i)) ||
		(($smodel =~ /(aficio)/i) &&
		 ($smake !~ /^(ricoh)$/i)) ||
		(($smodel =~ /(varioprint)/i) &&
		 ($smake !~ /^(oce)$/i)) ||
		(($smodel =~ /(okipage|microline)/i) &&
		 ($smake !~ /^(oki)$/i)) ||
		($smodel =~ /^(color|gdi)$/i)) {
		push(@smessages,
		     "Are you really sure that there is a \"$model\" made by \"$make\"? Did you select the correct manufacturer?");
		$fatals = 1;
	    }

	    # Model name typed into manufacturer field
	    if ($smake =~ /(deskjet|laserjet|designjet|officejet|stylus|aculaser|stylewriter|imagewriter|deskwriter|laserwriter|pixus|pixma|selphy|imagerunner|\bbjc\b|\bbj\b|docuprint|docupage|phaser|workcentre|homecentre|magicolor|pageworks|pagepro|aficio|varioprint|okipage|microline)/i) {
		push(@smessages,
		     "Please enter the model name into the field for the model name and not into the field for entering a new manufacturer.");
		$fatals = 1;
	    }

	    # URL without "http://" in the beginning
	    if (($url =~ /\S/) && ($url !~ m!^(https?|ftp)://!)) {
		push(@smessages,
		     "The address of the manufacturer's web page for the printer does not start with \"http://\". Links to external pages must always start with \"http://\".");
		$fatals = 1;
	    }

	    # URL with illegal characters
	    if (($url =~ /\S/) && ($url =~ m![\s<>]!)) {
		push(@smessages,
		     "The address of the manufacturer's web page for the printer contains illegal characters.");
		$fatals = 1;
	    }

	    # PPD URL with illegal characters
	    if (($ppdurl =~ /\S/) && ($ppdurl =~ m![\s<>]!)) {
		push(@smessages,
		     "The address for downloading the PPD file for the printer contains illegal characters.");
		$fatals = 1;
	    }

	    # New driver, but no URL for it

	    # No comment, nor a recommended driver, how is this printer
	    # made working?
	    my $nodcomments = 1;
	    for my $d (@dcomments) {
		if ($d =~ /\S/) {
		    $nodecomments = 0;
		    last;
		}
	    }
	    if (($notes !~ /\S/) && ($recdriver !~ /\S/) &&
		($nodcomments)) {
		if ($func ne "F") {
		    push(@smessages,
			 "Please let us know how you got this printer to work. Either fill in the \"Notes\" field, fill in one of the driver's notes fields, or choose a \"Recommended driver\" (at least one driver must be added for that).");
		    $fatals = 1;
		} else {
		    push(@smessages,
			 "Are you sure that this printer does not work with Linux/Unix and a free software driver? Do you really want to rate it \"Paperweight\" without any further comments?");
		    $warnings = 1;
		}
	    }
	    
	    # Recommended a driver, but rated the printer as "Paperweight"
	    if (($recdriver =~ /\S/) && ($func eq "F")) {
		push(@smessages,
		     "You have chosen a recommended driver for your printer, so you have most probably made it somehow working, but you have rated it as a \"Paperweight\", which means that it does not work at all.");
		$fatals = 1;
	    }

	    # Contrib URL without "http://" in the beginning
	    if (($contrib_url =~ /\S/) && 
		($contrib_url !~ m!^(https?|ftp)://!)) {
		push(@smessages,
		     "The web address for additional info/contributions does not start with \"http://\". Links to external pages must always start with \"http://\".");
		$fatals = 1;
	    }

	    # Contrib URL with illegal characters
	    if (($contrib_url =~ /\S/) &&
		($contrib_url =~ m![\s<>]!)) {
		push(@smessages,
		     "The web address for additional info/contributions contains illegal characters.");
		$fatals = 1;
	    }

	    # Auto-detection: Only make and model filled in, probably typed
	    # and not taken from the actual printer
	    if (($usb_mfg && $usb_mdl && !$usb_des && !$usb_cmd &&
		 !$usb_ieee) ||
		($par_mfg && $par_mdl && !$par_des && !$par_cmd &&
		 !$par_ieee) ||
		($general_mfg && $general_mdl && !$general_des &&
		 !$general_cmd && !$general_ieee)) {
		push(@smessages,
		     "Have you really taken the auto-detection data from the actual printer, following the instructions in the input form? Do not type simply manufacturer and model name from above. If you do not have the connector, cable, or printer add-on for a certain connection type, leave the fields for this type blank.");
		$warnings = 1;
	    }

	    # If there is at least one fatal problem or a warning when the
            # the "Save Anyway" button is not used, do not save and display
	    # the warning/error messages
	    if ($fatals ||
		(($action !~ /Save\s*Anyway/i) && $warnings)) {
		if ($#smessages == 0) {
		    $status = $smessages[0];
		} else {
		    $status = "<ul><li>" . join("</li><li>", @smessages) .
			"</li></ul>";
		}
		# Do not exit, show warning/error message(s) to user
		$action =~ s/Exit//ig;
	    } else {
		
		chdir $wdir;
		my $id = "editprinter:$oid";
		my $oldoid = $query->param('backupid');
		$oldoid =~ /^([^-]+)-/;
		my $oldmake = $1;
		$oldmake =~ s/_/ /g;
		my $mergeneeded = 0;
		my $head = undef;
		
		# User has changed manufacturer or model, rename wikish
		# RCS file then to not loose the history
		if ($oid ne $oldoid) {
		    if ((-f "$id,v") || $db->printer_exists($oid)) {
			# New printer ID already exists
			$status = "This printer already exists! We have tried to merge the existing and your entry as far as possible, but at some places, if in the existing and your entry fields are filled differently, your version is shown in red near the input field. Please merge these fields manually.<p>If you have entered this printer's name in error, please click the \"Back\" button of your browser to undo the name change.";
			# Do not exit, show message to user
			$action =~ s/Exit//ig;
			$mergeneeded = 1;
			$oidbeforemerge = $oldoid if !$oidbeforemerge;
			$fatals = 1;
		    } else {
			if ($oldoid && !$newentry) {
			    rename "editprinter:$oldoid,v", "$id,v";
			    # Delete out-of-sync cache files
			    if (-d $hcache) {
				unlink("$hcache/show-printer-recnum-$oid.html");
				unlink("$hcache/show-printer-recnum-$oldoid.html");
				unlink("$hcache/printer-list-make-$idmake.html");
				unlink("$hcache/printer-list-make-$oldidmake.html");
				unlink("$hcache/printer-list-make-Anyone.html");
				unlink("$hcache/printer-list-make-.html");
			    }
			    if (-d "$cachedir/pcache") {
				unlink("$cachedir/pcache/overview.perl");
			    }
			}
		    }
		}

		if (!$mergeneeded) {
		    if ((-r "$id,v") && $newentry) {
			# New printer entry requested but printer already
			# exists
			$status = "This printer already exists! Please modify the manufacturer and/or model fields to create a new entry. If you want to rename the printer you are working on, uncheck the \"New entry\" box.";
			# Do not exit, show error message to user
			$action =~ s/Exit//ig;
			$fatals = 1;
		    } else {
			# If there is no file for the new printer, make one
			if (! -f "$id,v") {
			    # If we have merged our entry with an
			    # already existing printer, preserve
			    # the history of the original entry
			    if ($oidbeforemerge) {
				rename "editprinter:$oidbeforemerge,v", "$id,v"
				    if !$newentry;
				$oidbeforemerge = "";
			    }
			    if (! -f "$id,v") {
				sysrun("rcs -t/dev/null -i '$id'");
				chmod 0664, "$id,v";
				sysrun("touch '$id'");
				chmod 0664, "$id";
				sysrun("ci -r '$id'") or die;
				chmod 0664, "$id,v";
				# New record successfully created, unset $newentry
				if ($newentry) {
				    $newentry = undef;
				    $basev = undef;
				}
			    }
			}
			
			# submission lock, verify basev, check-in/unlock
			$head = undef;
			if (sysrun("co -l '$id'")) {
			    open PIPE, "rlog -b '$id'|" and do {
				while (<PIPE>) {
				    next unless m/^head: ([\d\.]+)/;
				    $head = $1;
				}
				close PIPE;
			    };
			    $basev = $head if !$basev;
			    
			    if ($head eq $basev) {
				# ok, check it in!
				
				open PRV, "co -q -p '$id'|";
				my $oldxml = join(//, <PRV>);
				close PRV;
				
				if ($oldxml ne $xml) {
				    open TMP, ">$id.$$" or die "tmpfile open error";
				    print TMP $xml;
				    close TMP;
				    rename "$id.$$", "$id" or die "rename error";
				    chmod 0664, "$id";
				    sysrun("ci -m'$comment' -r '$id'") or die "checkin error";
				    chmod 0664, "$id,v";
				    # Successfully saved, unset $newentry
				    $newentry = undef;
				    # update $basev
				    open PIPE, "rlog -b '$id'|" and do {
					while (<PIPE>) {
					    next unless m/^head: ([\d\.]+)/;
					    $head = $1;
					}
					close PIPE;
				    };
				    $basev = $head;
				    $status = "Printer data saved!";
				    
				    # If we have merged our entry with an
				    # already existing printer, preserve
				    # the history of the original entry
				    if ($oidbeforemerge) {
					rename "editprinter:$oidbeforemerge,v", "$id--$oidbeforemerge,v" if !$newentry;
					$oidbeforemerge = "";
				    }

				    # If the printer was renamed let "Exit"
				    # point to the correct entry
				    if ($fromwhere) {
					if ($fromwhere =~ 
					    /^show_printer\.cgi/) {
					    $fromwhere =
						"show_printer.cgi?recnum=$oid";
					} elsif ($fromwhere =~ 
					    /^printer_list\.cgi\?make/) {
					    $fromwhere =
						"printer_list.cgi?make=$make";
					}
				    }
				} else {
				    # same as current version!
				    sysrun("rcs -u '$id'");
				    unlink($id);
				    $status = "No data changed, so no new revision saved!";
				}
				
				if (-d $hcache) {
				    unlink("$hcache/show-printer-recnum-$oid.html");
				    unlink("$hcache/printer-list-make-$make.html");
				    unlink("$hcache/printer-list-make-Anyone.html");
				    unlink("$hcache/printer-list-make-.html");
				}
				if (-d "$cachedir/pcache") {
				    unlink("$cachedir/pcache/overview.perl");
				}
				#print "Location: show_printer.cgi?recnum=$subkey\r\n\r\n";
			    } else {
				# Doh! Merge needed...

				$status = "Someone else has edited this entry under you! We have tried to merge his and your entry as far as possible, but at some places, if he and you filled in the fields differently, your version is shown in red near the input field. Please merge these fields manually.";
				$mergeneeded = 1;
				$fatals = 1;
			    }
		    
			} else {
			    # Doh! Locked!
			    $status = "Entry is temporary locked (or broken?), please try again later!";
			    # Do not exit, show error message to user
			    $action =~ s/Exit//ig;
			    $fatals = 1;
			}
		    }
		}
		if ($mergeneeded) {
		    # unlock, unlink gfile, etc, etc
		    sysrun("rcs -u '$id'");
		    unlink($id);

		    # Load the new version
		    loaddata($head);

		    # Do not exit when "Save & Exit" was clicked but a merge
		    # is needed.
		    $action =~ s/Exit//ig;

		}
	    }
	}
    }
}

if ($action !~ /Exit/i) {
    # Show editing mask

    my (@body) = ();

    push(@body, Tr({bgcolor=>$navbarcolor},
		   td({align=>"right"},
		      "Your e-Mail:"),
		   td(textfield(-size=>32,
				-name=>'email',
				-value=>$email,
				-force=>1),
		   td(p("Your e-mail address. Not required, but if you supply it, we can ask you in the case that something is not clear in your new printer entry. We will not give your address to any third party, nor will we display it on web pages without any protection.")))));

    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     "New entry:"),
		  td(checkbox(-checked=>$newentry,
			      -name=>'newentry',
			      -label=>"",
			      -force=>1),
		  td(p("Check the box if you intend to create a new printer entry. If you have modified manufacturer and/or model name, the previous entry is conserved. If you want to rename an existing entry, uncheck this box.")))));
    
    my @mfgs = sort { Foomatic::DB::normalizename($a) cmp
		      Foomatic::DB::normalizename($b) } ($db->get_makes());
    my %mfgdescs = ();
    my $m;
    for $m (@mfgs) {
	$mfgdescs{$m} = $m;
    }
    unshift(@mfgs,'');
    $mfgdescs{''} = "Choose here";
    push(@body, Tr({bgcolor=>$navbarcolor},
		   td({align=>"right"},
		      "Manufacturer:"),
		   td(popup_menu(-name=>"mmake",
				 -values=>\@mfgs,
				 -labels=>\%mfgdescs,
				 -default=>($make ?
					    (Foomatic::DB::member($make,@mfgs)?
					     $make : '') :
					    ($make_orig &&
					     Foomatic::DB::member($make_orig,
								  @mfgs) ?
                                             $make_orig : '')),
				 -force=>1 ),
		      " OR ",
		      textfield(-size=>16,
				-name=>'make',
				-value=>($make ? 
					 (!Foomatic::DB::member($make,@mfgs)?
					  $make : '') :
					 ($make_orig &&
					  !Foomatic::DB::member($make_orig,
								@mfgs) ?
					  $make_orig : '')),
				-force=>1),
		      (!$make || ($make_orig eq $make) ? () :
		       (br(), font({color=>"red"}, h($make_orig))))),
		   td(p("Manufacturer name for the printer. If there
                         are already printers of this manufacturer,
                         choose the manufacturer name from the menu
                         and leave the input field blank.<br>
                         DO NOT write the model name into the input
                         field, the model name goes into the field below. 
                         Use the input field for manufacturers which are
                         not listed yet."))));
    push(@body, Tr({bgcolor=>$navbarcolorlt},
		   td({align=>"right"},
		      "Model:"),
		   td(textfield(-size=>32,
				-name=>'model',
				-value=>($model ? $model : $model_orig),
				-force=>1),
		      (!$model || ($model_orig eq $model) ? () :
		       (br(), font({color=>"red"}, h($model_orig))))),
		   td(p("Model name for the printer.  Please try
			 to follow the conventions used for
			 other printers in the same
			 family. DO NOT repeat the manufacturer's name
                         in this field."))));
    
    $url_str = "URL:";
    $url_str = a({href=>$url}, "URL:") if $url;
    push(@body, Tr({bgcolor=>$navbarcolor},
		   td({align=>"right"},
		      $url_str),
		   td(textfield(-name=>'url',
				-size=>32,
				-maxlength=>128,
				-value=>($url ? $url : $url_orig),
				-force=>1),
		      (!$url || ($url_orig eq $url) ? () :
		       (br(), font({color=>"red"}, h($url_orig))))),
		   td(p("Manufacturer's web page for this specific printer.  The maker's home page will already be linked to, so if you don't know where to find a page about this printer, leave this blank. And do not forget the \"http://\" in the beginning of the address."))));
    
    push(@body, Tr({bgcolor=>$navbarcolorlt},
		   td({align=>"right"},
		      "Resolution&nbsp;(X&nbsp;x&nbsp;Y):"),
		   td(textfield(-name=>'resolution_x',
				-size=>4,
				-value=>(($resolution_x || $resolution_y) ? 
					 $resolution_x :
					 $resolution_x_orig),
				-force=>1),
		      "x",
		      textfield(-name=>'resolution_y',
				-size=>4,
				-value=>(($resolution_x || $resolution_y) ? 
					 $resolution_y : 
					 $resolution_y_orig),
				-force=>1),
		      ((!$resolution_x && !$resolution_y) || 
		       (($resolution_x_orig eq $resolution_x) &&
			($resolution_y_orig eq $resolution_y)) ? () :
		       (br(), font({color=>"red"},
				   h($resolution_x_orig) . "x" .
				   h($resolution_y_orig))))),
		   td(p("Maximum X and Y resolution the printer can do.  Available Unix software may not support the finest modes; if so, please say so in the notes."))));
    push(@body,Tr({bgcolor=>$navbarcolor},
		  td({align=>"right"},
		     "Color:"),
		  td(checkbox(-checked=>$color,
			      -name=>'color',
			      -label=>"",
			      -force=>1),
		     (($color_orig xor $color) ?
		      (font({color=>"red"},
			    ($color_orig ? "[X]" : "[&nbsp;&nbsp;]"))) : 
		      ())),
		  td(p("Check the box if this printer can do color.  Some printers may not be able to do so without vendor drivers; say so in the notes if so."))));
    
    my %typedescs = ( '' => 'Unknown/Other',
		      'inkjet' => 'Inkjet',
		      'laser' => 'Laser',
		      'impact' => 'Impact',
		      'dotmatrix' => 'Dot Matrix',
		      'led' => 'LED',
		      'sublimation' => 'Dye Sublimation',
		      'transfer' => 'Thermal Transfer' );
    my @types = sort(keys(%typedescs));
    
    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     "Mechanism:"),
		  td(popup_menu(-name=>'type',
				-values=>\@types,
				-labels=>\%typedescs,
				-default=>$type,
				-force=>1),
		      ($type_orig eq $type ? () :
		       (font({color=>"red"}, $typedescs{$type_orig})))),
		  td(p("What sort of printing mechanism does this printer use?"))));
    
    push(@body,Tr({bgcolor=>$navbarcolor},
		  td({align=>"right"},
		     "Refill:"),
		  td(textfield(-name=>'refill',
			       -size=>32,
			       -value=>($refill ? $refill : $refill_orig),
				-force=>1),
		     (!$refill || ($refill_orig eq $refill) ? () :
		      (br(), font({color=>"red"}, h($refill_orig))))),
		  td(p("A short description of the non-paper consumable(s): cartridge, ribbon, toner, printhead, etc.  Ballpark refill pricing would be nice, too, if known."))));
    

    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     "Language:"),
		  td(checkbox(-name=>'postscript',
			      -checked=>$postscript,
			      -label=>'PostScript',
			      -force=>1),
		     (($postscript_orig xor $postscript) ?
		      (font({color=>"red"},
			    ($postscript_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) : 
		      ()),
		     "level",
		     textfield(-name=>'postscript_level',
			       -value=>($postscript_level ?
					$postscript_level :
					$postscript_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$postscript_level ||
		      ($postscript_level_orig eq $postscript_level) ? () :
		      font({color=>"red"}, h($postscript_level_orig))),
		     br(),
		     "&nbsp;&nbsp;&nbsp;URL for manufacturer's PPD", br(),
		     "&nbsp;&nbsp;&nbsp;",
		     textfield(-name=>'ppdurl',
			       -value=>($ppdurl ? $ppdurl : $ppdurl_orig),
			       -size=>30,
			       -force=>1),
		     (!$ppdurl || ($ppdurl_orig eq $ppdurl) ? () :
		      font({color=>"red"}, h($ppdurl_orig))),
		     br(),
		     checkbox(-name=>'pdf',
			      -checked=>$pdf,
			      -label=>'PDF',
			      -force=>1),
		     (($pdf_orig xor $pdf) ?
		      (font({color=>"red"},
			    ($pdf_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ()),
		     "level",
		     textfield(-name=>'pdf_level',
			       -value=>($pdf_level ?
					$pdf_level : $pdf_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$pdf_level ||
		      ($pdf_level_orig eq $pdf_level) ? () :
		      font({color=>"red"}, h($pdf_level_orig))),
		     br(),
		     checkbox(-name=>'lips',
			      -checked=>$lips,
			      -label=>'LIPS',
			      -force=>1),
		     (($lips_orig xor $lips) ?
		      (font({color=>"red"},
			    ($lips_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ()),
		     "level",
		     textfield(-name=>'lips_level',
			       -value=>($lips_level ?
					$lips_level : $lips_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$lips_level ||
		      ($lips_level_orig eq $lips_level) ? () :
		      font({color=>"red"}, h($lips_level_orig))),
		     br(),
		     checkbox(-name=>'pcl',
			      -checked=>$pcl,
			      -label=>'PCL',
			      -force=>1),
		     (($pcl_orig xor $pcl) ?
		      (font({color=>"red"},
			    ($pcl_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) : 
		      ()),
		     "level",
		     textfield(-name=>'pcl_level',
			       -value=>($pcl_level ?
					$pcl_level : $pcl_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$pcl_level ||
		      ($pcl_level_orig eq $pcl_level) ? () :
		      font({color=>"red"}, h($pcl_level_orig))),
		     br(),
		     checkbox(-name=>'escp',
			      -checked=>$escp,
			      -label=>'ESC/P',
			      -force=>1),
		     (($escp_orig xor $escp) ?
		      (font({color=>"red"},
			    ($escp_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ()),
		     "level",
		     textfield(-name=>'escp_level',
			       -value=>($escp_level ?
					$escp_level : $escp_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$escp_level ||
		      ($escp_level_orig eq $escp_level) ? () :
		      font({color=>"red"}, h($escp_level_orig))),
		     br(),
		     checkbox(-name=>'escp2',
			      -checked=>$escp2,
			      -label=>'ESC/P 2',
			      -force=>1),
		     (($escp2_orig xor $escp2) ?
		      (font({color=>"red"},
			    ($escp2_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ()),
		     "level",
		     textfield(-name=>'escp2_level',
			       -value=>($escp2_level ?
					$escp2_level : $escp2_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$escp2_level ||
		      ($escp2_level_orig eq $escp2_level) ? () :
		      font({color=>"red"}, h($escp2_level_orig))),
		     br(),
		     checkbox(-name=>'hpgl2',
			      -checked=>$hpgl2,
			      -label=>'HP-GL/2',
			      -force=>1),
		     (($hpgl2_orig xor $hpgl2) ?
		      (font({color=>"red"},
			    ($hpgl2_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ()),
		     "level",
		     textfield(-name=>'hpgl2_level',
			       -value=>($hpgl2_level ?
					$hpgl2_level : $hpgl2_level_orig),
			       -size=>4,
			       -force=>1),
		     (!$hpgl2_level ||
		      ($hpgl2_level_orig eq $hpgl2_level) ? () :
		      font({color=>"red"}, h($hpgl2_level_orig))),
		     br(),
		     checkbox(-name=>'tiff',
			      -checked=>$tiff,
			      -label=>'TIFF',
			      -force=>1),
		     (($tiff_orig xor $tiff) ?
		      (font({color=>"red"},
			    ($tiff_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ()),
		     "level",
		     textfield(-name=>'tiff_level',
				   -value=>($tiff_level ?
					    $tiff_level : $tiff_level_rig),
			       -size=>4,
			       -force=>1),
		     (!$tiff2_level ||
		      ($tiff2_level_orig eq $tiff_level) ? () :
		      font({color=>"red"}, h($tiff_level_orig))),
		     br(),
		     checkbox(-name=>'proprietary',
			      -checked=>$proprietary,
			      -label=>'Proprietary',
			      -force=>1),
		     (($proprietary_orig xor $proprietary) ?
		      (font({color=>"red"},
			    ($proprietary_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ())),
		  td(p("The printer control language spoken by this printer, and level or version if known.  Mail us and add a remark in the \"Notes:\" field if we've forgotten any languages."))));
    
    push(@body,Tr({bgcolor=>$navbarcolor},
		  td({align=>"right"},
		     "ASCII:"),
		  td(checkbox(-checked=>$ascii,
			      -name=>'ascii',
			      -label=>"",
			      -force=>1),
		     (($ascii_orig xor $ascii) ?
		      (font({color=>"red"},
			    ($ascii_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ())),
		  td(p("This printer will print text if you just send it plain ascii.  Uncheck for printers that <b>only</b> work with Ghostscript and a driver or the like."))));
    
    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     "PJL:"),
		  td(checkbox(-checked=>$pjl,
			      -name=>'pjl',
			      -label=>"",
			      -force=>1),
		     (($pjl_orig xor $pjl) ?
		      (font({color=>"red"},
			    ($pjl_orig ?
			     "[X]&nbsp;" : "[&nbsp;&nbsp;]&nbsp;"))) :
		      ())),
		  td(p("Check the box if this printer supports HP's Printer Job Language (PJL)."))));
    
    $func = 'F'if !$func;
    
    my %drivdescs = ( 'A' => 'Perfectly',
		      'B' => 'Mostly',
		      'D' => 'Partially',
		      'F' => 'Paperweight' );
    my @drivs = sort keys(%drivdescs);
    
    push(@body,Tr({bgcolor=>$navbarcolor},
		  td({align=>"right"},
		     "Functionality:"),
		  td(popup_menu(-name=>'func',
				-values=>\@drivs,
				-labels=>\%drivdescs,
				-default=>$func,
				-force=>1),
		     ($func_orig eq $func ? () :
		      (font({color=>"red"}, $drivdescs{$func_orig})))),
		  td(p("How well does this printer work using Un*x software (ie ghostscript).  Put details into the \"Notes:\" field.  Mostly means it prints, but minor things are missing. Partially means it prints, but major things are missing.<br>If you choose a non-Paperweight rating, choose/enter a driver and/or enter in the \"Notes:\" field how you made this printer working."))));
    
    
    # Assemble the list-o-driver information
    my ($i, $j);
    my @drivers = ();
    if ($#dnames >= 0) {
	for ($i = 0; $i <= $#dnames; $i++) {
	    my $dname = $dnames[$i];
	    my $lnk = $dlinks[$i];
	    my $comment = $dcomments[$i];
	    my $mergecomment;
	    for ($j = 0; $j <= $#dnames_orig; $j++) {
		if ($dnames_orig[$j] eq $dname) {
		    $mergecomment = $dcomments_orig[$j];
		}
	    }
	    push(@drivers,
		 dt("<input type=\"hidden\" name=\"dname$i\" value=\"$dname\">\n",
		    checkbox(-checked=>1,
			     -name=>"dactive$i",
			     -label=>'',
			     -force=>1),
		    ($lnk ? a({href=>$lnk}, $dname) : $dname)),
		 dd(textarea(-name=>"dcomment$i",
			     -rows=>4,
			     -columns=>75,
			     -value=>($comment ? $comment : $mergecomment),
			     -force=>1 ),
		    (!$comment ||
		     ($mergecomment eq $comment) ? () :
		     (br(), font({color=>"red"}, h($mergecomment)))))),
	}
    }
    if ($#dnames_orig >= 0) {
	my $k = $#dnames + 1;
	for ($i = 0; $i <= $#dnames_orig; $i++) {
	    my $dname = $dnames_orig[$i];
	    my $lnk = $dlinks_orig[$i];
	    my $comment = $dcomments_orig[$i];
	    my $alreadythere = 0;
	    for ($j = 0; $j <= $#dnames; $j++) {
		if ($dnames[$j] eq $dname) {
		    $alreadythere = 1;
		    last;
		}
	    }
	    next if $alreadythere;
	    push(@drivers,
		 dt("<input type=\"hidden\" name=\"dname$k\" value=\"$dname\">\n",
		    checkbox(-checked=>1,
			     -name=>"dactive$k",
			     -label=>'',
			     -force=>1),
		    ($lnk ? a({href=>$lnk}, $dname) : $dname)),
		 dd(textarea(-name=>"dcomment$k",
			     -rows=>4,
			     -columns=>75,
			     -value=>$comment,
			     -force=>1 ))),
	    push(@dnames, $dname);
	    $k++;
	}
    }
    $dnumber = $#dnames + 1;
    if ($dnumber > 0) {
	@dnames = sort { Foomatic::DB::normalizename($a) cmp
			 Foomatic::DB::normalizename($b) } (@dnames);
	unshift(@dnames, 'None');
	unshift(@drivers,
		"Recommended driver:",
		popup_menu(-name=>'recdriver',
			   -values=>\@dnames,
			   -default=>$recdriver,
			   -force=>1), br());
    }
    push(@body,
	 Tr({bgcolor=>$navbarcolorlt},
	    td({align=>'right',
		valign=>'center'},
	       "Drivers:"),
	    td({valign=>'top', colspan=>2},
	       p(font({size=>-1},
		      "Current drivers for this printer.",
		      "Uncheck a checkbox if you're
	               <em>certain</em> that this driver doesn't
                       work with this printer, and the driver
                       will go away.")))),
	 Tr({bgcolor=>$navbarcolorlt},
	    td('&nbsp;'),
	    td({colspan=>2},
	       dt(@drivers)))) if $dnumber > 0;

    my @drivs = sort { Foomatic::DB::normalizename($a) cmp
		       Foomatic::DB::normalizename($b) } ($db->get_driverlist());
    my %drivdescs = ();
    my $d;
    for $d (@drivs) {
	$drivdescs{$d} = $d;
    }
    unshift(@drivs,'');
    $drivdescs{''} = ($dnumber <= 0 ? 'No driver' : 'No new driver');
    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     "<input type=\"hidden\" name=\"dnumber\" value=\"" . ($dnumber + 1) .
		     "\">\n",
		     "<input type=\"hidden\" name=\"dactive$dnumber\" value=\"on\">\n",
		     ($dnumber <= 0 ? "Driver:" : "New driver:")),
		  td(popup_menu(-name=>"dname$dnumber",
				-values=>\@drivs,
				-labels=>\%drivdescs,
				-default=>'',
				-force=>1 ),
		     " OR ",
		     textfield(-name=>"dname" . ($dnumber + 1),
			       -value=>'',
			       -size=>16)),
		  td(p(($dnumber <= 0 ?
			"A driver known to work. " : 
			"Another driver known to work. ")))));
    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     ($dnumber <= 0 ? "Driver notes:" : "New driver notes:")),
		  td(textarea(-name=>"dcomment$dnumber",
			      -rows=>4,
			      -columns=>35,
			      -value=>'',
			      -force=>1 )),
		  td(p("Comments on using the above " . 
		       ($dnumber <= 0 ? "" : "new ") .
		       "driver with this printer."))));
    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right"},
		     "&nbsp;"),
		  td("<input type=submit name=action value=\"" .
		     ($dnumber <= 0 ? "Add driver" : "Add/remove driver(s)").
		     "\">\n"),
		  td(p("Click here to get " . 
		       ($dnumber <= 0 ? "the driver into a separate place, " : 
			"the new driver into the list, ") .
		       "so that you can add another driver." .
		       ($dnumber <= 0 ? "" :
			" Drivers unchecked in the list will be removed then.")
		       ))));
    $url_str = "Un*x URL";
    $url_str = a({href=>$contrib_url}, "Un*x URL")
	if $contrib_url;
    push(@body,Tr({bgcolor=>$navbarcolor},
		  td({align=>"right"},
		     "$url_str:"),
		  td(textfield(-name=>'contrib_url',
			       -value=>($contrib_url ?
					$contrib_url : $contrib_url_orig),
			       -size=>32,
			       -force=>1),
		     (!$contrib_url ||
		      ($contrib_url_orig eq $contrib_url) ? () :
		      (br(), font({color=>"red"}, h($contrib_url_orig))))),
		  td(p("Web address for important info about using this printer with Unix-like operating systems/free software; website with special tricks, mini-HOWTO, a user's experience, or whatever else helps to make it going. Do not forget the \"http://\" in the beginning of the address."))));
    
    push(@body,Tr({bgcolor=>$navbarcolorlt},
		  td({align=>"right", valign=>'top'},
		     "Notes:"),
		  td({colspan=>2},
		     p(font({size=>-1, color=>'#202020'},
			    "This is HTML that just gets pasted
				into the table; <font
				color='#000000'>watch those &lt;
				signs!</font> Anything big can be
				either linked to with the Un*x
				URL/More Info link, or you can mail us
				the comments to set up as a locally
				served page.")))),
	 Tr({bgcolor=>$navbarcolorlt},
	    td('&nbsp;'),
	    td({colspan=>2},
	       textarea(-name=>'notes',
			-rows=>10,
			-columns=>80,
			-value=>($notes ? $notes : $notes_orig),
			-force=>1 ),
	       (!$notes ||
		($notes_orig eq $notes) ? () :
		(br(), font({color=>"red"}, h($notes_orig)))))));
    
    push(@body, Tr({bgcolor=>$navbarcolor},
		   td({align=>"right", valign=>"top"},
		      "Auto-detect:"),
		   td({colspan=>2},
			  p(font({size=>-1},
				 'Auto-detection info for this printer (if you have the possibility to connect your printer in different ways, try all to gather as much auto-detection info as possible), do not fill this in if you are not able to get this data from the actual printer:<p>' .
				 '<b>Parallel port:</b> Parport probe information for this printer. These should be exactly the contents of the lines from /proc/parport/#/autoprobe (kernel 2.2.x) or /proc/sys/dev/parport/parport#/autoprobe* (kernel 2.4.x and 2.6.x). "#" is the parallel port number; ie typically 0, "*" can be nothing or a number.  Remove the leading MODEL:, MANUFACTURER:, DESCRIPTION:, and COMMAND SET:, and also remove the ending semicolon.  If you had <tt>MODEL: Stylus Color 670;</tt>, for example, you\'d put <tt>Stylus Color 670</tt> in the MODEL field here.<p>' .
				 '<b>USB:</b> Download the "<a href="/download/printing/getusbprinterid.pl">getusbprinterid.pl</a>" Perl script, make it executable ("chmod a+rx getusbprinterid.pl"), and then run (as "root") "./getusbprinterid.pl /dev/usb/lp0" (or "/dev/usblp0", "/dev/usb/lp1", or whatever the USB device file to access your printer is). If your printer is configured with <a href="http://hpoj.sf.net/">HPOJ</a> use the "ptal-devid" command. You will get the so-called device ID string as output. Cut and paste this into the "IEEE-1284 Device ID String" field. Take care that all is on one line in the field. Put also the elements of the IEEE string into the appropriate fields "MANUFACTURER/MFG", "MODEL/MDL", ...<p>' .
				 '<b>Network printer:</b> Auto-detection is done via SNMP (Simple Network Management Protocol). Download and install <a href="http://www.ibr.cs.tu-bs.de/projects/scli/">SCLI</a> and run "scli -c \'show printer info\' &lt;host name of the printer&gt;". Look for a "Description:" field in the output. Copy and paste its contents into the "Description" field below.<p>' .
				 'In most cases the IEEE-1284 auto-detection data is the same for USB and parallel port. So usually you should put this data into the "General" section below. If you see any deviations, enter them in the "Parallel Port" and "USB" sections. Leave fields blank if they are identical to the entry in the "General" section, if they are blank, or if they do not exist in your observed auto-detection data.')))), 
	 Tr({bgcolor=>$navbarcolor},
	    td('&nbsp;'),
	    td({colspan=>2},
	       dl(dt('<b>General (Parallel and USB)</b><br>' .
		     'IEEE-1284 Device ID String'),
		  dd(textfield(-size=>80,
			       -name=>'general_ieee',
			       -value=>($general_ieee ?
					$general_ieee :
					$general_ieee_orig),
			       -force=>1),
		     (!$general_ieee ||
		      ($general_ieee_orig eq $general_ieee) ? () :
		      (br(), font({color=>"red"}, h($general_ieee_orig))))),

		  dt('MANUFACTURER/MFG'),
		  dd(textfield(-size=>32,
			       -name=>'general_mfg',
			       -value=>($general_mfg ?
					$general_mfg : $general_mfg_orig),
			       -force=>1),
		     (!$general_mfg ||
		      ($general_mfg_orig eq $general_mfg) ? () :
		      (br(), font({color=>"red"}, h($general_mfg_orig))))),

		  dt('MODEL/MDL'),
		  dd(textfield(-size=>32,
			       -name=>'general_mdl',
			       -value=>($general_mdl ?
					$general_mdl : $general_mdl_orig),
			       -force=>1),
		     (!$general_mdl ||
		      ($general_mdl_orig eq $general_mdl) ? () :
		      (br(), font({color=>"red"}, h($general_mdl_orig))))),

		  dt('DESCRIPTION/DES'),
		  dd(textfield(-size=>32,
			       -name=>'general_des',
			       -value=>($general_des ?
					$general_des : $general_des_orig),
			       -force=>1),
		     (!$general_des ||
		      ($general_des_orig eq $general_des) ? () :
		      (br(), font({color=>"red"}, h($general_des_orig))))),

		  dt('COMMAND SET/CMD'),
		  dd(textfield(-size=>32,
			       -name=>'general_cmd',
			       -value=>($general_cmd ?
					$general_cmd : $general_cmd_orig),
			       -force=>1),
		     (!$general_cmd ||
		      ($general_cmd_orig eq $general_cmd) ? () :
		      (br(), font({color=>"red"}, h($general_cmd_orig))))),

		  dt('<br><b>Parallel Port</b><br>' .
		     'IEEE-1284 Device ID String'),
		  dd(textfield(-size=>80,
			       -name=>'par_ieee',
			       -value=>($par_ieee ?
					$par_ieee : $par_ieee_orig),
			       -force=>1),
		     (!$par_ieee ||
		      ($par_ieee_orig eq $par_ieee) ? () :
		      (br(), font({color=>"red"}, h($par_ieee_orig))))),

		  dt('MANUFACTURER'),
		  dd(textfield(-size=>32,
			       -name=>'par_mfg',
			       -value=>($par_mfg ?
					$par_mfg : $par_mfg_orig),
			       -force=>1),
		     (!$par_mfg ||
		      ($par_mfg_orig eq $par_mfg) ? () :
		      (br(), font({color=>"red"}, h($par_mfg_orig))))),

		  dt('MODEL'),
		  dd(textfield(-size=>32,
			       -name=>'par_mdl',
			       -value=>($par_mdl ?
					$par_mdl : $par_mdl_orig),
			       -force=>1),
		     (!$par_mdl ||
		      ($par_mdl_orig eq $par_mdl) ? () :
		      (br(), font({color=>"red"}, h($par_mdl_orig))))),

		  dt('DESCRIPTION'),
		  dd(textfield(-size=>32,
			       -name=>'par_des',
			       -value=>($par_des ?
					$par_des : $par_des_orig),
			       -force=>1),
		     (!$par_des ||
		      ($par_des_orig eq $par_des) ? () :
		      (br(), font({color=>"red"}, h($par_des_orig))))),

		  dt('COMMAND SET'),
		  dd(textfield(-size=>32,
			       -name=>'par_cmd',
			       -value=>($par_cmd ?
					$par_cmd : $par_cmd_orig),
			       -force=>1),
		     (!$par_cmd ||
		      ($par_cmd_orig eq $par_cmd) ? () :
		      (br(), font({color=>"red"}, h($par_cmd_orig))))),

		  dt('<br><b>USB</b><br>' .
		     'IEEE-1284 Device ID String'),
		  dd(textfield(-size=>80,
			       -name=>'usb_ieee',
				   -value=>($usb_ieee ?
					    $usb_ieee : $usb_ieee_orig),
			       -force=>1),
		     (!$usb_ieee ||
		      ($usb_ieee_orig eq $usb_ieee) ? () :
		      (br(), font({color=>"red"}, h($usb_ieee_orig))))),

		  dt('MANUFACTURER/MFG'),
		  dd(textfield(-size=>32,
			       -name=>'usb_mfg',
			       -value=>($usb_mfg ?
					$usb_mfg : $usb_mfg_orig),
			       -force=>1),
		     (!$usb_mfg ||
		      ($usb_mfg_orig eq $usb_mfg) ? () :
		      (br(), font({color=>"red"}, h($usb_mfg_orig))))),

		  dt('MODEL/MDL'),
		  dd(textfield(-size=>32,
			       -name=>'usb_mdl',
			       -value=>($usb_mdl ?
					$usb_mdl : $usb_mdl_orig),
			       -force=>1),
		     (!$usb_mdl ||
		      ($usb_mdl_orig eq $usb_mdl) ? () :
		      (br(), font({color=>"red"}, h($usb_mdl_orig))))),

		  dt('DESCRIPTION/DES'),
		  dd(textfield(-size=>32,
			       -name=>'usb_des',
			       -value=>($usb_des ?
					$usb_des : $usb_des_orig),
			       -force=>1),
		     (!$usb_des ||
		      ($usb_des_orig eq $usb_des) ? () :
		      (br(), font({color=>"red"}, h($usb_des_orig))))),

		  dt('COMMAND SET/CMD'),
		  dd(textfield(-size=>32,
			       -name=>'usb_cmd',
			       -value=>($usb_cmd ?
					$usb_cmd : $usb_cmd_orig),
			       -force=>1),
		     (!$usb_cmd ||
		      ($usb_cmd_orig eq $usb_cmd) ? () :
		      (br(), font({color=>"red"}, h($usb_cmd_orig))))),

		  dt('<br><b>Network Printer (SNMP)</b><br>' .
		     'Description'),
		  dd(textfield(-size=>32,
			       -name=>'snmp_des',
			       -value=>($snmp_des ?
					$snmp_des : $snmp_des_orig),
			       -force=>1),
		     (!$snmp_des ||
		      ($snmp_des_orig eq $snmp_des) ? () :
		      (br(), font({color=>"red"}, h($snmp_des_orig)))))))));

    print header(-expires=>'-15m', -type=>'text/html; charset=UTF-8');
    print start_html(-BGCOLOR=>"white",
		     -title=>($make && $model ?
			      ($oidindb ?
			       "Editing printer $make $model" :
			       "Editing new printer $make $model") :
			      "Adding new printer"));

    print navbar2($navloc,
		  "Printer information input page",
		  p(a({href=>"printer_list.cgi"},
		      "Return"),
		    "to the Foomatic printer compatibility database."),
		  ($oidindb ?
		   (p(a({href=>"show_printer.cgi?recnum=$oid"},
			"Go"),
		      " to this printer's information page.")) : ()),
		  p(a({href=>"edit_printer.cgi?action=Undelete"},
		      "Undelete"),
		    "a deleted printer entry."),
		  p("There are a fair number of things kept in this database
                    about each printer.  Unfortunately, this makes for a
                    somewhat long form.<p>
                    <b>Please</b> take a few minutes
                    and fill out anything you can. Note that your printer
                    entry is only useful if the reader can see how he gets 
                    the printer to work (driver, where to get the driver
                    if it is not yet listed at OpenPrinting, to which
                    printer is yours compatible, special tricks, ...) or
                    if he gets the message that this model definitely does
                    not work. If you get red warning/error messages, read
                    them and try to fix the problems. Use \"Save Anyway\"
                    only if you are absolutely sure that you have done it
                    the right way.<p>
                    Note that if you make a mistake you can edit everything
                    at any time after submitting. So do not create another
                    printer entry if you are not content. Simply correct
                    your entry where needed.<p>
                    Please correct also the printer entries made by others
                    if you see anything wrong in them."),
		  ($status ?
		   p(font({color=>"red"}, b($status))) : ()),
		  $query->startform(-method=>"POST",
				    -action=>"edit_printer.cgi"),
		  "<input type=\"hidden\" name=\"basev\" value=\"$basev\">\n",
		  #"<input type=\"hidden\" name=\"recnum\" value=\"$oid\">\n",
		  "<input type=\"hidden\" name=\"backupmake\" value=\"$make\">\n",
		  "<input type=\"hidden\" name=\"backupmodel\" value=\"$model\">\n",
		  "<input type=\"hidden\" name=\"backupid\" value=\"$oid\">\n",
		  ($oidbeforemerge ?
		   "<input type=\"hidden\" name=\"oidbeforemerge\" value=\"$oidbeforemerge\">\n" :
		   ()),
		  "<input type=\"hidden\" name=\"fromwhere\" value=\"$fromwhere\">\n",
		  #"<input type=submit name=action value=\"Previous\">\n",
		  "<input type=submit name=action value=\"Save\">\n",
		  ($warnings && !$fatals ?
		   "<input type=submit name=action value=\"Save Anyway\">\n" :
		   ()),
		  "<input type=submit name=action value=\"Save & Exit\">\n",
		  "<input type=submit name=action value=\"Exit\">\n",
		  ((-r "$wdir/editprinter:$oid,v") ?
		   "<input type=submit name=action value=\"Delete\">\n" : ()),
		  table(@body),
		  $query->endform());
    print end_html();
    undef $req;
} else {
    # Exit button clicked, jump to $fromwhere
    print redirect($fromwhere);
}

exit 0;

sub h {
    # Abbreviation for Foomatic::DB::htmlify()
    return Foomatic::DB::htmlify(@_);
}

sub sysrun {
    # Shell command call
    my $r = system @_;
    return (0 == ($r / 256));
}

sub loaddata {
    # Try to load existing data

    my ($rev) = @_; 
    my $date = localtime();

    # Load from Wikish
    my $head = '1.1';
    my $id = "editprinter:$oid";
    chdir $wdir;
    open PIPE, "rlog -b '$id'|" and do {
	while (<PIPE>) {
	    next unless m/^head: ([\d\.]+)/;
	    $head = $1;
	}
	close PIPE;
    	$basev = $head if !$basev;
	$basev = $rev if $rev;

	my $VAR1;
	eval (`co -q -r$basev -p '$id' | $foomaticpath/foomatic-perl-data -P`);
	$p = $VAR1 if $VAR1;
	$oidindb = $db->printer_exists($p->{'id'});
    };
    $basev = $head if !$basev;

    # Load from Foomatic database
    if (!$p->{'functionality'}) {
	if (!defined($db->printer_exists($oid))) {
	    # We don't have that printer, make it
	    $p->{id} = $oid;
	    $p->{make} = $make;
	    $p->{model} = $model;
	} else {
	    $p = $db->get_printer($oid);
	    $make = $p->{'make'};
	    $model = $p->{'model'};
	    $oidindb = 1;
	    $urlmake = $make;
	    $urlmake =~ s/ /\%20/g;
	    $idmake = $make;
	    $idmake =~ s/ /_/g;
	}
    }

    if ($p->{'functionality'}) {
	# Fill variables from loaded database entry
	$oid = $p->{'id'};
	$make = $p->{'make'};
	$model = $p->{'model'};
	$urlmake = $make;
	$urlmake =~ s/ /\%20/g;
	$idmake = $make;
	$idmake =~ s/ /_/g;
	$url = $p->{'url'};
	$resolution_x = $p->{'maxxres'};
	$resolution_y = $p->{'maxyres'};
	$color = $p->{'color'};
	$ascii = $p->{'ascii'};
	$type = $p->{'type'};
	$refill = $p->{'refill'};
	$func = $p->{'functionality'};
	$pjl = $p->{'pjl'};
	$notes = $p->{'comment'};
	$ppdurl = $p->{'ppdurl'};
	$recdriver = $p->{'driver'};

	for my $l (@{$p->{'languages'}}) {
	    if ($l->{'name'} =~ /postscript/i) {
		$postscript = 1;
		if (defined($l->{'level'})) {
		    $postscript_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /pdf/i) {
		$pdf = 1;
		if (defined($l->{'level'})) {
		    $pdf_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /lips/i) {
		$lips = 1;
		if (defined($l->{'level'})) {
		    $lips_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /pcl/i) {
		$pcl = 1;
		if (defined($l->{'level'})) {
		    $pcl_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /escp2/i) {
		$escp2 = 1;
		if (defined($l->{'level'})) {
		    $escp2_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /escp/i) {
		$escp = 1;
		if (defined($l->{'level'})) {
		    $escp_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /hpgl/i) {
		$hpgl2 = 1;
		if (defined($l->{'level'})) {
		    $hpgl2_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /tiff/i) {
		$tiff = 1;
		if (defined($l->{'level'})) {
		    $tiff_level = $l->{'level'};
		}
	    } elsif ($l->{'name'} =~ /proprietary/i) {
		$proprietary = 1;
	    }
	}

	# Assemble the list-o-driver information

	# Driver data from Foomatic database (driver XML files)
	@dnames = ();
	@dlinks = ();
	@dcomments = ();
	my @dlist = sort { Foomatic::DB::normalizename($a) cmp
			   Foomatic::DB::normalizename($b) } ($db->get_drivers_for_printer($oid));
	my $d;
	if (defined(@dlist)) {
	    for $d (@dlist) {
		my $drvdata = $db->get_driver($d);
		next if !$drvdata;
		$dnumber ++;
		my $dname = $d;
		push(@dnames, $dname);
		my $lnk = $drvdata->{'url'};
		push(@dlinks, $lnk);
		my ($pdelem, $pdnotes);
		for $pdelem (@{$drvdata->{'printers'}}) {
		    next if (!$pdelem->{'id'});

		    my $pid = $pdelem->{'id'};
		    next if ($pid ne "$oid");
		
		    my $n = $pdelem->{'comment'};
		    $pdnotes = $n if defined($n);
		}
		push(@dcomments, $pdnotes);
	    }
	}

	# Driver data from printer XML file (printer XML file from Wikish)
	@dlist = sort { Foomatic::DB::normalizename($a) cmp
			Foomatic::DB::normalizename($b) } (@{$p->{'drivers'}});
	$d;
	if (defined(@dlist)) {
	    for $d (@dlist) {
		next if !$d->{'id'};
		my $dname = $d->{'id'};
		my $drvdata = $db->get_driver($dname);
		my $dlink = $drvdata->{'url'};
		my $dcomment = $d->{'comment'};
		my $dfound = 0;
		for (my $j = 0; $j <= $#dnames; $j++) {
		    if ($dnames[$j] eq $dname) {
			# Replace existing driver entry
			$dlinks[$j] = $dlink;
			$dcomments[$j] = $dcomment;
			$dfound = 1;
		    }
		}
		if (!$dfound) {
		    # New driver entry
		    $dnumber ++;
		    push(@dnames, $dname);
		    push(@dlinks, $dlink);
		    push(@dcomments, $dcomment);
		}
	    }
	}

	$general_ieee = $p->{'general_ieee'};
	$general_mdl = $p->{'general_mdl'};
	$general_mfg = $p->{'general_mfg'};
	$general_des = $p->{'general_des'};
	$general_cmd = $p->{'general_cmd'};

	$par_ieee = $p->{'par_ieee'};
	$par_mdl = $p->{'par_mdl'};
	$par_mfg = $p->{'par_mfg'};
	$par_des = $p->{'par_des'};
	$par_cmd = $p->{'par_cmd'};

	$usb_ieee = $p->{'usb_ieee'};
	$usb_mdl = $p->{'usb_mdl'};
	$usb_mfg = $p->{'usb_mfg'};
	$usb_des = $p->{'usb_des'};
	$usb_cmd = $p->{'usb_cmd'};

	$snmp_des = $p->{'snmp_des'};

	$contrib_url = $p->{'contriburl'};

    }
}

sub simplifystr {
    my ($str) = @_;
    $str = lc($str);
    $str =~ s/\b\s*(limited|ltd\.?|inc\.?|int\.?|international)\b//g;
    $str =~ s/\bokidata\b/oki/g;
    #$str =~ s/[^a-z0-9+]//g;
    $str =~ s/\+/plus/g;
    $str =~ s/[^a-z0-9]+/ /g;
    $str =~ s/(?<=[a-z])(?=[0-9])/ /g;
    $str =~ s/(?<=[0-9])(?=[a-z])/ /g;
    $str =~ s/hewlett\s*packard/hp/g;
    $str =~ s/kyocera\s*mita/kyocera/g;
    $str =~ s/(minolta\s*qms|konica\s*minolta)/minolta/g;
    return $str;
}

sub text_is_spam {
    my ($text) = @_;

    # Return values
    # First value
    # 0: No spam-typical content
    # 1: Do not accept this input
    # Second value
    # Error message

    # Simple blacklist on spammer-typical URLs
    my $a;
    for $a (qw/tipcell\.com 367net\.com valu\.com herbs?\.com mobile4sale http:\/\/\d+\.[^\.\/]+\.[^\.\/]+ arbico/) {
        return (1, "Spam-typical text found: $a") if
		($text =~ m!$a!i);
    }

    # Auto-HTML the links
    $text = fixhtml($text);

    # Find out how many links are in the text
    my $linklimit = 3;
    my $linklimit2 = $linklimit * 2;
    my $linklimit1 = $linklimit2 + 1;
    if ($text =~ m!((https?|ftp:).*){$linklimit1,}!is) {
	return (1, "More than $linklimit links found!");
    }

    # Get the length of the text before removing tags
    my $lengthbefore = length($text);

    # Remove all HTML tags, even nested ones
    while ($text =~ s/<[^<>]*>//s) {};

    # Get the length of the text before removing tags
    my $lengthafter = length($text);

    # Too high visible text/HTML tag ratio? Consider input as spam
    $tagratiolimit = 0.4;
    if (($lengthbefore >= 10) &&
	(1 - $lengthafter/$lengthbefore > $tagratiolimit)) {
	my $tagpercentage = int((1 - $lengthafter / $lengthbefore) * 100);
	my $tagpercentagelimit = int($tagratiolimit * 100);
        return (1, "Too many HTML tags: $tagpercentage \%, allowed: $tagpercentagelimit \%");
    }

    # Now we can do word filtering, as no word is interrupted any more by
    # HTML tags intendedly inserted by the spammer (ex: via<!-- adflas -->gra)

    # Simple blacklist on spam-typical text
    my $a;
    for $a (qw/viagra cialis valtrex tramadol ativan vicodin temazepam xanax adipex online.*(casino|poker|roulette|pharmacy) home\s*owner skyrocket somaonline/) {
        return (1, "Spam-typical text found: $a") if
		($text =~ m!$a!i);
    }

    # Nothing found? Then return 0 as spam level
    return (0, 'OK');
}

sub fixhtml {
    my ($str) = @_;
    # Strange line breaks of non-Unix operating systems
    $str =~ s/(\r\n|\n\r)/\n/smg;  # Windows
    $str =~ s/\r/\n/smg;           # Mac
    # New paragraph on blank line
    $str =~ s/(?<![\>\n\r])\s*\n[\s\n\r]*\n/<p>\n\n/smg;
    # Make automatically links from URLs
    $str =~ s,(?<!\>)(?<!href=")(?<!href=)(https?://[^\s\n\r"\>\<]+),<a href="$1">$1</a>,smig;
    # Do not embed files from external servers
    $str =~ s,src[\s\n\r]*=[\s\n\r]*\"?https?://[^\s\n\r\"\>\<]+\"?,src="",smig;
    # Remove scripts
    $str =~ s,<[\s\n\r]*script.*<[\s\n\r]*/[\s\n\r]*script[\s\n\r]*>,,smig;
    return $str;
}
