#! /usr/bin/perl

use News::NNTPClient;
use Set::IntSpan;
use CGI;
use CGI qw/p/;
use Text::Wrap;
use HTTP::Date;
use Data::Dumper;

umask 0000;

$query = new CGI;
my $url = $query->self_url;

# Web server to use, fully-qualified URL
$THISPAGEURL = $url;
$THISPAGEURL =~ s,^(http://[^\?]+)(|\?.*)$,$1,;

# News server to use, fully-qualified domain name
$NEWSSERVER = '@NEWSSERVER@';

# Web forum server to use, fully-qualified domain name
$FORUMSERVER = '@FORUMSERVER@';

# Mailing list server to use, fully-qualified domain name
$LISTSERVER = '@LISTSERVER@';

# URL to general info pages of mailman
$MAILMANINFO = '@MAILMANINFO@';

# URL to archives of mailman
$MAILMANARCHIVES = '@MAILMANARCHIVES@';

# Default Organization: line for postings to your news server.
$DEFAULT_ORG = "@DEFAULT_ORG@";

# Maintainer e-mail address that appears at the bottom of every WebNews page:
$MAINTAINER='@MAINTAINER@';

# User-Agent header used when posting to newsgroups: 
$USER_AGENT = '@USER_AGENT@';

# Temporary "limbo" data directory
$LIMBODIR = '@LIMBODIR@';

my $redirect = undef;
my $title = '';
my @html = ();

# Fix up inputs
for $imageinput ('subscribe', 'Submit', 'chgsubscriptions') {
    if ($query->param("${imageinput}.x") || $query->param("${imageinput}.y")) {
	$query->delete("${imageinput}.x");
	$query->delete("${imageinput}.y");
	$query->param("${imageinput}", 'true');
    }
}
if ($query->param("Submit")) {
    $query->delete("Submit");
    $query->param("submit", "true");
}

# Bounce to proper url if newsportal-style link 
if ($query->param('name')) {
    print "Location: forums.cgi?group=" . $query->param('name') . "\r\n\r\n";
    exit(0);
}

# Short-circuit to mailing lists or forums if no news server is available
if (!$NEWSSERVER) {
    if ($FORUMSERVER) {
	print "Location: http://" . $FORUMSERVER . "/" .
	    "\r\n\r\n";
    } else {
	if ($query->param('group')) {
	    my $listname = get_listname($query->param('group'));
	    print "Location: http://" . $LISTSERVER . "/mailman/listinfo/" .
		$listname . "\r\n\r\n";	
	} else {
	    print "Location: http://" . $LISTSERVER . "/mailman/listinfo" .
		"\r\n\r\n";
	}
    }
    exit 0;
}

# Quicken startup time, by not loading nntp in certain cases,
$load_nntp = 1 unless ($query->param('group') && $query->param('post'));
$load_groups = 1 unless ($query->param('group'));

# Connect to news server
my @newsgroups = ();
if ($load_nntp) {
    $nntp = new News::NNTPClient($NEWSSERVER)
	or die "Couldn't connect to news server $NEWSSERVER ($!); " .
	"please try again later.";
    
    # We might have timed out waiting for all those stupid locks. Let's ping
    # the server to see whether things are okay. MODE READER is
    # basically a no-op. If we didn't load_active, this should just
    # shoot straight through.
    unless ($nntp->mode_reader) {
	$nntp->quit;
	$nntp->connect;
	if (! $nntp->mode_reader) {
	    $err = $nntp->code . " " . $nntp->message;
	    $nntp->quit;
	    push(@html,
		 "<P><B>Error in connecting to news server.</B>",
		 "<br>[$err]</P>\n");
	    goto out;
	}
    }
    
    if ($load_groups) {
	# Read in news active file
	my @active = $nntp->list('active', 'linuxprinting.*');
	@newsgroups = map { my $a = $_; $a =~ s/\s+.*$//; chomp $a; $a; } @active;
    }
}

# Main dispatch tree
if ($query->param('group')) {
    if ($query->param('post')) {
	do_post($query->param('group'));
    } elsif ($query->param('article')) {
	if ($query->param('submit')) {
	    do_submit($query->param('group'), $query->param('article'));
	} elsif ($query->param('reply')) {
	    do_reply($query->param('group'), $query->param('article'));
	} else {
	    do_article($query->param('group'), $query->param('article'));
	}
    } else {
	do_group($query->param('group'));
    }
} else {
    if ($query->param('delimbo')) {
	do_delimbo($query->param('delimbo'));
    } else {
	do_groupindex();
    }
}

out:

if ($load_nntp) {
	$nntp->quit;
}

if ($redirect) {
    print "Location: $redirect\r\n\r\n";
} else {
    use Theme;
    print ($query->header,
	   $query->start_html(-title=>"OpenPrinting Forums: $title",
			      -bgcolor=>'#FFFFFF'),
	   Theme::navbar2('forums.cgi', "Forums: $title", @html),
	   $query->end_html);
}

exit 0;



#
# Assorted functions:
#

sub get_nolist {
    my $nolist = 0;
    if ($_[0] eq 'linuxprinting.test') {
	$nolist = 1;
    } 
    return $nolist;
}

sub get_noweb {
    my $noweb = 0;
#    if ($_[0] eq 'linuxprinting.inkjet.architecture') {
#	$noweb=1;
#    }
    return $noweb;
}

sub get_listname {
    # Compute list name.  Ugly mix of explicitness and heuristic computation
    my $listname = $_[0];
    $listname =~ s!linuxprinting\.!!;
    $listname =~ s!\.general!!;
    if ($listname eq 'foomatic.devel') {
	$listname = 'printing-foomatic';
    } elsif ($listname eq '.general') {
	$listname = 'printing-user-general';
    } elsif ($listname eq 'inkjet.architecture') {
	$listname = 'printing-driver';
    } elsif ($listname eq 'minolta') {
	$listname = 'printing-user-konica-minolta';
    } else {
	$listname = "printing-user-$listname";
    }
    return $listname;
}

sub start_headers {
    return "<TABLE BORDER=0>\n";
}

sub end_headers {
    return "</TABLE><P>\n";
}

sub print_header {
    my ($field, $contents) = @_;
    $contents = entitize($contents);
    if ($field eq 'From:') {
	$contents =~ s/\@/ at /;
    }
    push (@html, 
	  "<TR><TD ALIGN=RIGHT><B>${field}</B></TD>",
	  "<TD ALIGN=LEFT>${contents}</TD></TR>\n");
}

sub maybe_print_header {
    my ($field, $contents) = @_;
    print_header($field, $contents) 
	if $field =~ /^(From|Subject|Date)/io;
}

sub fetch_article {
    my ($nntp, $group, $article) = @_;
    $nntp->group($group);
    return $nntp->article($article);
}

sub artindex {
    my ($group) = @_;
    $query->delete('article');
    $query->delete('delimbo');
    $query->delete('reply');
    $query->delete('realemail');
    $query->delete('article_body');
    $query->delete('submit');
    $query->delete('email');
    $query->delete('msgid');
    $query->delete('fullname');
    $query->delete('subject');
    $query->delete('reply');
    $query->param('group', $group);
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">[Message index]</A> \n";
}

sub grpindex {
    $query->delete('delimbo');
    $query->delete('reply');
    $query->delete('realemail');
    $query->delete('article_body');
    $query->delete('submit');
    $query->delete('email');
    $query->delete('msgid');
    $query->delete('fullname');
    $query->delete('subject');
    $query->delete('reply');
    $query->delete('article');
    $query->delete('delimbo');
    $query->delete('reply');
    $query->delete('group');
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">[Other forums]</A>\n";
}

sub reply {
    my ($group, $article) = @_;
    $query->param('group', $group);
    $query->param('article', $article);
    $query->param('reply', 'true');
    my $selfurl = $query->self_url;
    $query->delete('reply');
    return "<A HREF=\"$selfurl\">[Reply]</A>\n";
}

sub mailinglist {
    my ($group) = @_;
    my $listname = get_listname($group);
    my $nolist = get_nolist($group);
    if ($nolist) {
	return "";
    } else {
	return "<A HREF=\"$MAILMANINFO/$listname\">[Mailing List]</A>\n";
    }
}

sub archives {
    my ($group) = @_;
    my $listname = get_listname($group);
    my $nolist = get_nolist($group);
    if ($nolist) {
	return "";
    } else {
	return "<A HREF=\"$MAILMANARCHIVES/$listname/\">[Archives]</A>\n";
    }
}

sub returntoarticle {
    my ($group, $article) = @_;
    $query->param('group', $group);
    $query->param('article', $article);
    $query->delete('reply');
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">[Return to message]</A>\n";
}

sub post {
    my ($group) = @_;
    $query->param('group', $group);
    $query->param('post', 'true');
    $query->delete('reply');
    my $selfurl = $query->self_url;
    $query->delete('post');
    return "<A HREF=\"$selfurl\">[Post a message]</A>\n";
}

sub sortbynumber {
    my ($group) = @_;
    $query->param('group', $group);
    $query->delete('sortorder');
    $query->param('sortorder', 'number');
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">Number</A>\n";
}

sub sortbyfrom {
    my ($group) = @_;
    $query->param('group', $group);
    $query->delete('sortorder');
    $query->param('sortorder', 'from');
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">From</A>\n";
}

sub sortbysubject {
    my ($group) = @_;
    $query->param('group', $group);
    $query->delete('sortorder');
    $query->param('sortorder', 'subject');
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">Subject</A>\n";
}

sub sortbydate {
    my ($group) = @_;
    $query->param('group', $group);
    $query->delete('sortorder');
    $query->param('sortorder', 'date');
    my $selfurl = $query->self_url;
    return "<A HREF=\"$selfurl\">Date</A>\n";
}

sub wrap_line {
    $Text::Wrap::columns = 80;
    return split ("\n", wrap('','', $_[0]));
}

sub wrap_message {
    my @input = @_;
    my @output;
    foreach my $line (@input) {
	if (length($line) >= 80) {
	    push(@output, wrap_line($line));
	} else {
	    push(@output, $line);
	}
    }
    return @output;
}

sub do_submit {
    my ($group, $article) = @_;
    my ($fromaddr, $fullname, $from, $newsgroups, $subject, $cc);
    
    $title = "Submission error";
    
    # Get params, parse, srub...

    $fromaddr = $query->param('email');
    if (!scalar($fromaddr)) {
	$fromaddr = 'nobody@example.com';
    }
    $trueaddr = $query->param('realemail');
    $fullname = $query->param('fullname');
    $newsgroups = $query->param('group');
    $subject = $query->param('subject');
    $msgid = $query->param('msgid');
    $article_body = $query->param('article_body');
    
    if ($fromaddr !~ m!\@!) {
	$result = 'No \@-sign in public email.';
	goto err;
    }
    if ($trueaddr !~ m!\@!) {
	$result = 'No \@-sign in real email.';
	goto err;
    }
    if ($fromaddr =~ m![\"\'\<\>\(\)\!\#\,\;\:\n\r\`]!) {
	$result = 'Unexpected character in public email.';
	goto err;
    }
    if ($trueaddr =~ m![\"\'\<\>\(\)\!\#\,\;\:\n\r\`]!) {
	$result = 'Unexpected character in real email.';
	goto err;
    }
    if ($fullname =~ m![\"\'\<\>\(\)\!\#\,\;\:\n\r\`]!) {
	$result = 'Unexpected character in name.';
	goto err;
    }
    if ($msgid =~ m![\n\r]!) {
	$result = 'Unexpected character in i.';
	goto err;
    }
    if ($newsgroups =~ m![\n\r\,]!) {
	$result = 'Unexpected character in g.';
	goto err;
    }
    if ($subject =~ m![\n\r]!) {
	$result = 'Unexpected character in s.';
	goto err;
    }
    if (! $subject) {
	$result = 'No subject specified';
	goto err;
    }

    # Prepare article posting data
    $from = "$fromaddr ($fullname)";

    # Simple blacklist on confirmation address
    my $a;
    for $a (qw/gia\.com\.ua spam\.com tipcell\.com 367net\.com valu\.com herbs?\.com mobile4sale huangge\d+\@hotmail.com arbico/) {
	die "SPAM: confirmation address $trueaddr\n" if ($trueaddr =~ m!$a!i);
    }

    # Simple blacklist on text in body
    my $a;
    for $a (qw/tipcell\.com 367net\.com valu\.com Ericsson herbs?\.com mobile4sale http:\/\/\d+\.[^\.\/]+\.[^\.\/]+ arbico/) {
        die "SPAM: Body text $article_body\n" if ($article_body =~ m!$a!i);
    }

    $article_body =~ s/\r//g;
    my @article_body = split(/\n/, $article_body);

    # Prevent abuse of the newsgroup web interface for spam or for
    # cheating search machines
    my $inputlinect = scalar(@article_body) + 1;
    my $urllinesct = scalar(grep(m/tps?\:\/\//i, @article_body)) + 1;
    if (($urllinesct > 2) and ($inputlinect / $urllinesct < 3)) {
	die "SPAM: Suspicious url:lines ratio $urllinesct:$inputlinect in body!\n";
    }

    @article_body = wrap_message(@article_body);

    push(@headers,
	 "From: $from\n",
	 "Newsgroups: $newsgroups\n");
    push(@headers, "Subject: $subject\n") if $subject;
    push(@headers, "References: $msgid\n") if $msgid;

    my $remotehost = $query->remote_host();
    push(@headers, 
	 "Organization: $DEFAULT_ORG\n",
	 "User-Agent: $USER_AGENT\n",
	 "X-LinuxPrinting-Poster-Host: " . $remotehost . "\n");

    # No longer do we just up and post here.  Instead, we save state
    # in a file and reload it later when the user invokes us with the
    # magic number they learned via email.  Sigh.

    my $mothball = [ \@headers, \@article_body, $group, $trueaddr ];
    my $mothfile;
    do {
	$mothfile = sprintf("%u.%u", time(), rand(0xffffffff));
    } until ! -f $mothfile;
    open MOTH, ">$LIMBODIR/$mothfile" or die "Cannot open $LIMBODIR/$mothfile";
    print MOTH Dumper($mothball);
    close MOTH or die "Error closing $LIMBODIR/$mothfile";
    chmod 0664, "$LIMBODIR/$mothfile";

    my $mothurl = "$THISPAGEURL?delimbo=$mothfile";

    open MAIL, "|/usr/sbin/sendmail -bm -t" or die;
    print MAIL
	"To: $trueaddr
Subject: Please confirm this OpenPrinting forum post

Someone using your email address has recently attempted to post to a
OpenPrinting forum as detailed below.  If this was really you, 
please follow this web link to complete your post:

$mothurl

Please do not reply to this message.  There is no one listening.

Forum: $group
Public Name: $fullname
Public Email: $fromaddr
Posting IP: $remotehost
Subject: $subject

First part of the post:

",
join("\n", @article_body[0 .. 7]);

    if (close MAIL) {
	$title = "Thank You";
	push (@html, 
	      grpindex(),
	      artindex($group),
	      "<p>An email has been sent to your private address.<br>",
	      "Please follow the web link inside it to complete your post.</p>");
    } else {
	$result = "Mailing error";
err:
	$title = "Error";
	push (@html, 
	      grpindex(),
	      artindex($group),
	      "<p>Sorry, there was some sort of problem.<br>",
	      "[$result]",
	      "</p>");
    }
}

sub do_delimbo {
    my ($mothkey) = @_;

    $mothkey =~ s![^\d\.]!!;	# raw off the wire!

    if (-r "$LIMBODIR/$mothkey") {
	open MOTH, "<$LIMBODIR/$mothkey" or die;
	my $mothstr = join('', <MOTH>);
	my $VAR1;
	eval "$mothstr" or die;
	close MOTH;
	my $group = $VAR1->[2];
	my @headers = @{$VAR1->[0]};
	my @article_body = @{$VAR1->[1]};

	if (0) {
	    # test
	    print "Content-Type: text/plain\r\n\r\n";
	    print "Mothball data:\n";
	    print "\ngroup: $group\n\n";
	    print @headers, "\n", join("\n", @article_body);
	    exit(0);
	} else {
	    # Actually post...
	    $result = $nntp->post(@headers, "\n", @article_body);
	    if ($result) {
		# Bounce back if happy
		rename("$LIMBODIR/$mothkey", "$LIMBODIR/$mothkey.posted");
		chmod 0664, "$LIMBODIR/$mothkey.posted";
		$redirect = "forums.cgi?group=$group";
	    } else {
		# Error path - show an error message, either from the nntp
		# server or from our input processing code above.
		$result = $nntp->code . " " . $nntp->message;
		chomp $result;
		$title = "Submission error";
		push (@html, 
		      grpindex(),
		      artindex($group),
		      "<p>Your post was not completed.<br>",
		      "[$result]");
	    }
	}
    } else {
	$title = "Problem";
	push (@html, 
	      grpindex(),
	      "<p>Odd, no such post seems to be in limbo.",
	      "Your post was not completed.");
    }
}

sub entitize {
    my $what = $_[0];
    $what =~ s/\&/\&amp;/g;
    $what =~ s/\"/\&quot;/g;
    $what =~ s/\</\&lt;/g;
    $what =~ s/\>/\&gt;/g;
    return $what;
}

sub do_reply {
    my ($group, $article) = @_;
    my @article = fetch_article($nntp, $group, $article);
    if (!scalar(@article)) {
	return do_article($group, $article);
    }
    my $i = 0;
    for (; $i < @article; $i++) {
	chomp($line = $article[$i]);
	if ($line =~ /^([^:]*:) (.*)$/) {
	    ($field, $contents) = ($1, $2);
	    if ($field =~ /Message-ID/io) {
		$msgid = $contents;
	    } elsif ($field =~ /From/io) {
		$from = $contents;
	    } elsif ($field =~ /Newsgroups/io) {
		$newsgroups = $contents;
	    } elsif ($field =~ /Subject/io) {
		if ($contents =~ /^Re:/io) {
		    $subject = "$contents";
		} else {
		    $subject = "Re: $contents";
		}
	    }
	} elsif ($line =~ /^$/) {
	    last;
	}
    }
    my $attribution = "In article $msgid, $from writes:\n";
    my @article_body = @article[$i..$#article];
    @article_body = map { "> $_"; } @article_body;
    unshift(@article_body, $attribution);
    my $article_body = join "", @article_body;
    $query->delete('reply');
    $query->delete('user');
    $query->delete('group');
    my $selfurl = $query->self_url;
    
    $title="Reply to message $article in $group";
    push @html,  "<FORM METHOD=\"POST\" ACTION=\"$selfurl\">\n";
    return do_postingform("REPLY", $group, $article, $USER, $msgid,
			  $newsgroups, $subject, $DEFAULT_ORG, 
			  $article_body);
}

sub do_post {
    my ($group) = @_;
    $query->delete('post');
    $query->delete('user');
    $query->delete('group');
    my $selfurl = $query->self_url;
    
    $title = "Post new message to $group";
    push @html,  "<FORM METHOD=\"POST\" ACTION=\"$selfurl\">\n";
    return do_postingform("POST", $group, "0 but true", $USER, "",
			  $group, "", $DEFAULT_ORG, "");
}

sub do_postingform {
    my ($mode, $group, $article, $user, $msgid,
	$newsgroups, $subject, $org,
	$article_body) = @_;
    $query->param('group', $group);
    push (@html,
	  grpindex(),
	  artindex($group),
	  mailinglist($group),
	  archives($group),
	  "<INPUT TYPE=\"hidden\" NAME=\"article\" VALUE=\"$article\">");

    if ($mode eq 'REPLY') {
	push (@html, 
	      returntoarticle($group, $article),
	      "<INPUT TYPE=\"hidden\" NAME=\"msgid\" VALUE=\"$msgid\">");
    }
    
    $article_body = entitize($article_body);

    push (@html,
	  "<hr>",
	  "<INPUT TYPE=\"hidden\" NAME=\"group\" VALUE=\"$group\">",
	  "<TABLE>",
	  "<TR><TD ALIGN=RIGHT>Full Name:</TD><TD ALIGN=LEFT>",
	  "<INPUT TYPE=\"text\" SIZE=60 NAME=\"fullname\">",
	  "</TD></TR>\n",
	  "<TR><TD VALIGN=TOP ALIGN=RIGHT>Public Email:</TD><TD ALIGN=LEFT>", 
	  "<INPUT TYPE=\"text\" SIZE=60 NAME=\"email\">",
	  "<br><font size=-1>A contact email shown with your post, or may be empty.</font></TD></TR>\n",
	  "<TR><TD VALIGN=TOP ALIGN=RIGHT>Private Email:</TD><TD ALIGN=LEFT>", 
	  "<INPUT TYPE=\"text\" SIZE=60 NAME=\"realemail\">",
	  "<br><font size=-1>Used for confirmation; <font color=\"#ff0000\"><b>must be real</b></font></font>.</TD></TR>\n",
	  "<TR><TD ALIGN=RIGHT>Subject:</TD><TD ALIGN=LEFT>", 
	  "<INPUT TYPE=\"text\" SIZE=60 NAME=\"subject\" VALUE=\"$subject\">",
	  "</TD></TR>",
	  "</TABLE>",
	  "<BR>",
	  "<TEXTAREA COLS=80 ROWS=36 NAME=\"article_body\" ",
	  "   style=\"font-family: monospace\" wrap=on>",
	  "$article_body\n",
	  "</TEXTAREA>",
	  "<p><INPUT TYPE=\"submit\" NAME=\"Submit\" VALUE=\"Post\">",
	  "</FORM><P>");
}

sub do_article {
    my ($group, $article) = @_;
    my @article = fetch_article($nntp, $group, $article);
    my $i = 0;
    my $selfurl = $query->url(-absolute);
    
    # Compute list name.  Ugly mix of explicitness and heuristic computation
    my $listname = get_listname($group);
    my $noweb = get_noweb($group);
    my $nolist = get_nolist($group);

    $title = "$group";
    $full_headers = 0;
    $query->delete('headers');
    push @html,  grpindex();
    push @html,  artindex($group);
    if (scalar(@article)) {
	push @html,  reply($group, $article);
    }
    push (@html, mailinglist($group), archives($group));
    
    push @html,  "<HR>";
    if (!scalar(@article)) {
	if ($nolist) {
	    push (@html, 
		  p("No such article."));
	} else {
	    push (@html, 
		  p("No such article.  It can probably be found in the long-term <a href=\"/pipermail/$listname/\">archives</a>."));
	}
    } else {
	push @html,  start_headers();
	for (; $i < @article; $i++) {
	    chomp($line = $article[$i]);
	    if ($line =~ /^([^:]*:)( .*)$/) {
		if ($full_headers) {
		    print_header($1, $2);
		} else {
		    maybe_print_header($1, $2);
		}
	    } elsif ($line =~ /^$/) {
		last;
	    } else { 
		# continued header line, skip push @html,  "$line\n";
	    }
	}
	
	push @html,  end_headers();
	push @html,  "<PRE>";
	for (; $i < @article; $i++) {
	    push @html,  entitize($article[$i]);
	}
	push @html,  "</PRE><P>\n";
    }
}

sub xover_sort_by_subject {
    local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
	= split(/\t/, $a);
    local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
	= split(/\t/, $b);
    1 while $asubj =~ s/^Re(:|-)?\s*//g;
    1 while $bsubj =~ s/^Re(:|-)?\s*//g;
    local ($sortorder) = (uc($asubj) cmp uc($bsubj));
    return $sortorder ? $sortorder : ($anumb <=> $bnumb);
}

sub xover_sort_by_from {
    local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
	= split(/\t/, $a);
    local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
	= split(/\t/, $b);
    $afrom =~ s/\"//g;
    $bfrom =~ s/\"//g;
    local ($sortorder) = (uc($afrom) cmp uc($bfrom));
    return $sortorder ? $sortorder : ($anumb <=> $bnumb);
}

sub xover_sort_by_date {
    local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
	= split(/\t/, $a);
    local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
	= split(/\t/, $b);
    local ($sortorder) = (str2time($bdate) <=> str2time($adate));
    return $sortorder ? $sortorder : ($bnumb <=> $anumb);
}

sub xover_sort_by_number {
    local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
	= split(/\t/, $a);
    local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
	= split(/\t/, $b);
    return ($anumb <=> $bnumb);
}

sub xover_sort {
    my ($sort_order, @xover) = @_;
    $sort_order = "xover_sort_by_$sort_order";
    return sort $sort_order @xover;
}

sub do_group {
    my ($group) = @_;
    my ($sortorder) = 'date';
    my (@rows);
    my ($readarticles, $serverhas);
    
    $title = "Messages for $group";
    push (@html, 
	  grpindex(),
	  post($group),
	  mailinglist($group),
	  archives($group),
	  '<hr>');
    
    if ($query->param('sortorder')) {
	$sortorder = $query->param('sortorder');
    }
    
    $serverhas = Set::IntSpan->new(scalar $nntp->group($group));
    
    # Fields in xover: numb subj from date mesg refr char line xref
    @xover = $nntp->xover($serverhas->run_list);
    push @html,  "<TABLE>";
    push (@html, 
	  "<TR BGCOLOR=#999999><TH>", 
	  sortbydate($group),
	  "</TH><TH>",
	  sortbysubject($group), 
	  "</TH><TH>", 
	  sortbyfrom($group), 
	  "</TH></TR>\n");
    foreach my $hdr (xover_sort($sortorder, @xover)) {
	my (@hdr) = split(/\t/, $hdr);
	my ($numb,$subj,$from,$date,$mesg,$refr,$char,$line,$xref) = @hdr;
	$query->param('group',$group);
	$query->param('article',$numb);
	$query->delete('sortorder');
	my $selfurl = $query->self_url;
	$subj = substr($subj,0,70);
	$subj = "<A HREF=\"$selfurl\">"
	    . $subj
	    . "</A>";
	$from =~ s/\"//g;
	if ($from =~ m!\((.+)\)!) {
	    $from = $1;
	} elsif ($from =~ m!(.+) \<.+\>!) {
	    $from = $1;
	}
	$from = substr($from,0,25);
	$date = time2str(str2time($date));
	$date =~ s!^...\, !!;
	$date =~ s! \d\d\d\d!!;
	$date =~ s!:\d\d GMT.*!!;
	push @html,  $query->Tr($query->td([$date,$subj,$from]));
    }
    push @html,  "</TABLE>";
    push @html,  "<P>";
}

sub do_groupindex {
    $title = "Index";
    if ($#newsgroups >= 0) {
	push (@html,
	      "Welcome to the OpenPrinting forums.", 
	      "<hr>",
	      "<TABLE>",
	      "<TR BGCOLOR=#999999><TH>",
	      "Recent<br>Posts",
	      "</TH><TH>",
	      "Web Forum",
	      "</TH><TH>", 
	      "Mailing List",
	      "</TH><TH>", 
	      "Archives",
	      "</TH></TR>\n");
	
	my $group;
	my %seengrp = ();
	foreach $group ('linuxprinting.general',
			'linuxprinting.macosx.general',
			'linuxprinting.bsd.general',
			'linuxprinting.solaris.general',
			'linuxprinting.foomatic.devel',
			sort @newsgroups) {
	    next if $seengrp{$group}++;

	    my ($f, $l) = $nntp->group($group);
	    $rsp = $nntp->message;
	    if ($rsp =~ m!^(\d+)\s!) {
		$howmany = $1;
	    }
	    $query->param('group',$group);
	    $selfurl = $query->self_url;

	    # Compute list name.  Ugly mix of explicitness and heuristic computation
	    my $listname = get_listname($group);
	    my $noweb = get_noweb($group);
	    my $nolist = get_nolist($group);

	    push (@html,
		  "<tr>",
		  ($noweb ? "<td>&nbsp;</td>"
		   : "<td align=\"right\">$howmany&nbsp;</td>"),
		  ($noweb ? "<td>&nbsp;</td>"
		   : "<td><A HREF=\"$selfurl\">$group</A></td>"),
		  ($nolist ? "<td>&nbsp;</td>"
		   : "<td><a href=\"$MAILMANINFO/$listname\">$listname</a></td>"),
		  ($nolist ? "<td>&nbsp;</td>"
		   : "<td><a href=\"$MAILMANARCHIVES/$listname/\">Archives</a></td>"),
		  "</tr>");
	}
    }
    push (@html,  
	  "</TABLE>\n");
}

die "Shouldn't get here";
