#! /usr/bin/perl

use CGI qw/ :standard :html3 /;
use CGI escape, unescape, font, center;

use lib "@LIBDIR@";
use Foomatic::DB;
use PHTDB qw/$hcache $cachedir/;

my $ua = $ENV{'HTTP_USER_AGENT'};
my $query = new CGI;
my $dbh;

# Check for special format request

my $driver   = $query->param('driver');
my $poid     = $query->param('printer');
my $shortgui = $query->param('shortgui');
my $show     = $query->param('show');

die "bad oid $poid" unless $poid =~ m!^[\d\w\_\-\.]+$!;
die "bad driver $driver" unless $driver =~ m!^[\d\w\-\.\_]+$!;

$shortgui = ($shortgui ? 1 : 0);
$show = ($show ? 1 : 0);

if (defined($driver) and defined($poid)) {

    # Do we have an old numerical printer ID or an otherwise renamed or
    # aliased printer? If so, bounce to the correct printer ID.

    my $newid = Foomatic::DB::translate_printer_id($poid);
    if ($newid ne $poid) {
	print redirect("ppd-o-matic.cgi?driver=$driver&printer=$newid&show=$show");
	exit(0);
    }

    # Generate a filename
    
    $filename = join('-',($poid, $driver)) . ".ppd";
    $filename =~ s![ /\(\)]!_!g;
    $filename =~ s![\+]!plus!g;
    $filename =~ s!__+!_!g;
    $filename =~ s!_\.ppd$!\.ppd!;
    $filename =~ s!_-!-!;
    $filename =~ s!^_!!;

    # OK, is there a cached page?

    my $cachepage = "$hcache/ppd-o-matic-recnum-$poid-driver-$driver" .
	($shortgui ? "-shortgui" : "") . ".ppd";
    my @ppd = ();
    if ( -f $cachepage ) {
	open CFILE, $cachepage or die "error opening $cachepage";
	@ppd = <CFILE>;
	close CFILE;
    } else {

	# Initialize the database

	my $db = new Foomatic::DB;
	my $overfile = "$cachedir/pcache/overview.perl";
	$db->{'overviewfile'} = $overfile;

	# Do we have a link to a custom PPD file for this driver in the
	# printer XML file? Then show the custom PPD

	my $p = $db->get_printer($poid);
	if (defined($p->{'drivers'})) {
	    for my $d (@{$p->{'drivers'}}) {
		next if ($d->{'id'} ne $driver);
		if ($d->{'ppd'}) {
		    my $ppdpath = $d->{'ppd'};
		    if ($ppdpath =~ m!^PPD/!) {
			$ppdpath = "foomatic-db/db/source/" . $ppdpath;
		    }
		    print redirect($ppdpath);
		    exit(0);
		}
		last;
	    }
	}

	# Do we have a PostScript printer and a link to a manufacturer-
	# supplied PPD file? Then show the manufacturer-supplied PPD

	if ($driver =~ /^Postscript$/i) {
	    if ($p->{'ppdurl'}) {
		my $ppdpath = $p->{'ppdurl'};
		if ($ppdpath =~ m!^PPD/!) {
		    $ppdpath = "foomatic-db/db/source/" . $ppdpath;
		}
		print redirect($ppdpath);
		exit(0);
	    }
	}

	# Get all the data about this driver/printer pair

	$db->getdat($driver, $poid);

	# Do we have a PostScript printer and a link to a manufacturer-
	# supplied PPD file? Then show the manufacturer-supplied PPD

	if (($driver =~ /^Postscript$/i) && ($db->{'dat'}{'ppdurl'})) {
	    print redirect($db->{'dat'}{'ppdurl'});
	    exit(0);
	}

	# No PS printer with manufacturer-supplied PPD? Generate the PPD 
	# from Foomatic database
	# If $shortgui is set, the GUI names ("translations") in the PPD
	# will be limited to max. 39 characters (the PostScript driver
	# for Windows of CUPS has a bug and therefore chokes on longer
	# GUI strings).

	@ppd = $db->getppd($shortgui);

	# Write preparsed web page cache

	umask 0002;
	if (! -d $hcache) {
	    mkdir $hcache or warn "Could not mkdir $hcache";
	    chmod 2775, $hcache;
	}
	if (open CFILE, "> $cachepage.$$") {
	    print CFILE @ppd;
	    close CFILE;
	    rename "$cachepage.$$", "$cachepage" 
		or warn "unable to rename $cachepage.$$";
	} else {
	    warn "unable to write $cachepage";
	}
    }

    my $mimetype;
    my $disposition;
    my $charset;
    if ($show) {
	$mimetype = "text/plain";
	$disposition = "inline";
	$charset = "; charset=UTF-8";
    } else {
	$mimetype = "application/octet-stream";
	$disposition = "attachment";
	$charset = "";
    }
    print header(-expires=>'-15m',
		 -type=>"$mimetype; name=$filename$charset",
		 -Content_Disposition=>"$disposition; filename=$filename");
    print @ppd;
}

exit 0;

