#! /usr/bin/perl

use CGI qw/ :standard :html3 /;
use CGI escape, unescape, font, center, dd, dt, dl, ul, li;

use lib "@LIBDIR@";
use Foomatic::Defaults;
use PHTDB qw/normalizename $hcache $cachedir/;
use Theme qw/navbar navbar2/;

my $wikishdir = "@WIKISHDIR@";

my $query = new CGI;
$query->autoEscape(0);

# Check for printer maker specification.
my $in_make = $query->param('make');
my $in_makeid = $in_make;
$in_makeid =~ s/\+/_/g;
$in_makeid =~ s/\%20/_/g;
$in_makeid =~ s/ /_/g;
$in_make =~ s/_/ /g;
$in_make =~ s/\+/ /g;
$in_make =~ s/\%20/ /g;
if (defined($in_make) and ($in_make ne "Anyone") and 
    ($in_make !~ m!^[\w\s\-\ ]+$! or $in_make !~ m![a-z]!i or
     ((`ls -1 $libdir/db/source/printer/ | grep -c ^$in_makeid-` =~ /^0$/) and
      (`ls -1 $wikishdir/ | grep -c ^editprinter.$in_makeid-` =~ /^0$/)))) {
    die "Illegal make='$in_make'\n";
}

# Connect to the database
my $db = new PHTDB;

my @output = ();

# Check for page in cache, or recompute?
my $cachekey = "printer-list-make-$in_makeid.html";
my $cachepage = "$hcache/$cachekey";
my $overfile = "$cachedir/pcache/overview.perl";
$db->{'overviewfile'} = $overfile;

if (-f $cachepage) {
    for (my $i = 0; $i < 20; $i++) {
	open CFILE, $cachepage or die "Error reading $cachepage\n";
	$output = join('', <CFILE>);
	close CFILE;
	# Only output the cashed page if it was really completely written
	# Before introduction of this measure pages would not display due
	# to an incomplete cache file until the next page rebuild (or
	# until manually nuking the cache).
	if ($output =~ m!</html>\s*$!s) {
	    print header(-expires=>'3m', -type=>'text/html; charset=UTF-8'), $output;
	    exit (0);
	}
	# If we have non-empty output, the other process is already
	# writing the cache file or there is definitely written a broken
	# cache. Retry only once in this case.
	$i = 18 if $output;
	sleep 10;
    }
} 

# Open cache file already now to tell concurrent processes that we are already
# computing this case
umask 0002;
die "Error writing $cachepage\n" if !(open CFILE, "> $cachepage");

my @makes = sort {normalizename($a) cmp normalizename($b) } $db->get_makes();
my %models;
for (@makes) {
    $models{$_} = [$db->get_models_by_make($_)];
}

my %oids;

# Get everything from the user-contributed printers in the Wikish directory
# so that these files get read only once
my @userprinters;
my %isuserprinter;
opendir WDIR, $wikishdir and do {
    while (my $file = readdir(WDIR)) {
	next if $file !~ /^editprinter:([^-]+-.*),v$/;
	my $oid = $1;
	next if $infoomatic{$oid};
	next if (! -r "$wikishdir/$file");
	$file =~ s/,v$//;
	next if $file =~ /--/;
	$xml = `cd $wikishdir; co -q -p '$file'`;
	next if !$xml;

	$xml =~ m!<make>([^<]+?)</make>!sm;
	my $make = $1;
	$xml =~ m!<model>([^<]+?)</model>!sm;
	my $model = $1;
	$xml =~ m!<functionality>([^<]+?)</functionality>!sm;
	my $functionality = $1;
	$unverified = ($xml =~ m!<unverified\s*/>!sm);

	push(@makes, $make)
	    if !Foomatic::DB::member($make, @makes);
	push(@{$models{$make}}, $model)
	    if !Foomatic::DB::member($model, @{$models{$make}});
	$oids{$make}{$model} = $oid;
	push(@userprinters,
	     {
		 'oid' => $oid,
		 'make' => $make,
		 'model' => $model,
		 'functionality' => $functionality,
		 'verified' => ($unverified ? 0 : 1)
	     });
	$isuserprinter{$oid} = 1;
    }
    closedir WDIR or die;
};

my $javascript = $db->get_javascript2(\%models, \%oids);

# Figure out listing title
my $title = "";

# Inputs!
# $in_make
# $in_predef

# Figure out if we have a query at all
my $have_query = 0;
if ($in_make) {
    $have_query = 1;
}

my @makers = sort {normalizename($a) cmp normalizename($b) } (@makes);
my %maker_listed = ();
my $m;
for $m (@makes) {
    $maker_listed{$m->{'make'}}++;
}

if ($have_query) {
    my $over = $db->get_overview();

    my %funcv = ( 'A' => 'Perfectly',
		  'B' => 'Mostly',
		  'D' => 'Partially',
		  'X' => 'Unknown',
		  'F' => 'Paperweight' );
    
    my $p;
    my %infoomatic;

    for $p (@$over) {
	next if ($p->{'make'} ne $in_make
		 and $in_make ne 'Anyone');
	next if ($p->{'noxmlentry'} && $isuserprinter{$p->{'id'}});

	$p->{'functionality'} = 'X' if (!defined($p->{'functionality'}) ||
					$p->{'noxmlentry'});
	push (@{$printers{$p->{'make'}}{$p->{'functionality'}}}, 
	      { 'model' => $p->{'model'},
		'oid' => $p->{'id'},
		'verified' => ($p->{'unverified'} ? 0 : 1) });

	$infoomatic{$p->{'id'}} = 1 if !$p->{'noxmlentry'};
    }

    my $upr;
    for $upr (@userprinters) {
	next if ($upr->{'make'} ne $in_make
		 and $in_make ne 'Anyone');
	next if $infoomatic{$upr->{'oid'}};
	push (@{$printers{$upr->{'make'}}{$upr->{'functionality'}}}, 
	      { 'model' => $upr->{'model'},
		'oid' => $upr->{'oid'},
		'verified' => $upr->{'verified'} });
    }

    my @mlist;
    my $make;
    my $userprinter = 0;
    for $make (sort {normalizename($a) cmp normalizename($b) } (keys(%printers))) {
	my @flist;
	my $func;
      func_lab:
	for $func (qw/A B D X F/) {

	    if (! scalar($printers{$make}{$func})) {
		# No such printers; empty cell and short-circuit to next loop
		push (@flist, td('&nbsp;'));
		next func_lab;
	    }

	    my @plist;
	    my $printer;
	    for $printer (sort { normalizename($a->{'model'}) cmp 
				     normalizename($b->{'model'}) 
				     } @{$printers{$make}{$func}}) {
		my $model = $printer->{'model'};
		$model =~ s/ /\&nbsp;/g;
		my $oid = $printer->{'oid'};
		push (@plist,
		      font({size=>-1},
			   a({href=>"show_printer.cgi?recnum=$oid"},
			     $model . ($printer->{'verified'} 
				       ? "" 
				       : "*"))));
		$userprinter = 1 if !$printer->{'verified'};
	    }

	    push (@flist, 
		  td({valign=>'top'},
		     join('<br>',@plist)));
	}
	push (@mlist,
	      Tr(td({valign=>'top'}, $make),
		 @flist));
    }

    push (@output,
	  start_html(-title => "OpenPrinting database - " .
		     ($in_make && $in_make ne 'Anyone' ?
		      "All $in_make Printers" :
		      "All Printers"),
		     -author => "Grant Taylor & Till Kamppeter",
		     -BGCOLOR => "white",
		     -script=>$javascript),
    
	  navbar2('printer_list.cgi',
		  ($in_make && $in_make ne 'Anyone' ?
		   "All $in_make Printers" :
		   "All Printers"),
		  p(a({href=>"./"},"Home"),
		    ' <b>:</b> ',
		    a({href=>"database.html"}, 'Database'),
		    ' <b>:</b> ',
		    a({href=>"printer_list.cgi"}, "Printers"),
		    ' <b>:</b> ',
		    ($in_make 
		     ? ($in_make eq 'Anyone' 
			? 'All'
			: "$in_make")
		     : 'Printers')),

		  list_selectors(),
		  '<a href="edit_printer.cgi?newentry=1' .
		  ($in_make && $in_make ne "Anyone" ? "&make=$in_make" : "") .
		  '">Add a new printer</a>',
		  hr(),
		  $tip ? (p($tip) . hr()) : "",
		  ($userprinter ?
		   p("*",
		     "User-contributed printer which is not in the <a href=\"foomatic.html\">Foomatic</a> packages yet. The data ia not yet verified and/or proofread. The printer entry can be edited by everyone.") : ()),
		  table({border=>'1'},
			Tr(th({align=>'left'},'Make'),
			   th({align=>'left'},
			      font({color=>'green'}, 'Perfectly') . 
			      '&nbsp;' .
			      '<img src="linuxyes.png">' .
			      '<img src="linuxyes.png">' .
			      '<img src="linuxyes.png">'),
			   th({align=>'left'},
			      font({color=>'green'}, 'Mostly') .
			      '&nbsp;' .
			      '<img src="linuxyes.png">' .
			      '<img src="linuxyes.png">'),
			   th({align=>'left'},
			      font({color=>'orange'}, 'Partially') .
			      '&nbsp;' .
			      '<img src="linuxyes.png">'),
			   th({align=>'left'},
			      font({color=>'black'}, 'Unknown') .
			      '&nbsp;' .
			      '<img src="linuxyes.png">' . "<sup>???</sup>"),
			   th({align=>'left'},
			      font({color=>'red'}, 'Paperweight') .
			      '&nbsp;' .
			      '<img src="linuxno.png">')),
			@mlist)),
	  
	  end_html());

    print (header(-expires=>'3m', -type=>'text/html; charset=UTF-8'),
	   @output);

    cache_output(\@output);

} else {
    # no query inputs; just show intro text

    push (@output, 
	  start_html(-title => "OpenPrinting database - Printer Listings",
		     -author => "Grant Taylor & Till Kamppeter",
		     -BGCOLOR => "white",
		     -script=>$javascript),
	  intro_text(),
	  end_html());

    print (header(-expires=>'3m', -type=>'text/html; charset=UTF-8'),
	   @output);

    # Write preparsed perlfile cache
    cache_output(\@output);
}


exit 0;


sub cache_output {
    my ($output) = @_;

    return 0 if !(-d $cachedir);
    my @path = qw/hcache/;
    my $pelem;
    my $dir = "$cachedir";
    for $pelem (@path) {
	$dir = "$dir/$pelem";
	if (! -d $dir) {
	    umask 0002;
	    mkdir $dir or warn "Could not mkdir $dir";
	    chmod 2775, $dir;
	}
    }
    
    if ($cachepage) {
	print CFILE @$output;
	close CFILE;
    } else {
	warn "unable to write $cachepage";
    }
    
}

sub intro_text {

    my @body;

    push(@body,

	 p('For more information about this database, including a
            "key" to the ranking system, see <a
            href="database.html">the intro page</a>'),

         p('Please bookmark and link to the database\'s <a
            href="database.html">front page</a> instead of this one.'),

	 p('There are several ways to search.  You can just go
            straight to a particular printer, or you can list all
            printers from a given manufacturer.  Each method has a
            set of controls below:'),

	 # get by make, type, etc
	 list_selectors(),

	 p('You have experience with a printer not listed here? <a
	    href="edit_printer.cgi?newentry=1">Add it</a>.'),

	 p('You can also view a list of all <a
            href="driver_list.cgi">known drivers</a> showing what
            printers work with each.'),
	 
	 p('If you have any questions, please ask them in a <a 
            href="forums.cgi">forum</a>.'),

	 );
    
    return navbar2('printer_list.cgi',
		   'Printer Listings',
		   p(a({href=>"./"},"Home"),
		     ' <b>:</b> ',
		     a({href=>"database.html"}, 'Database'),
		     ' <b>:</b> ',
		     'Printers'),
		  @body);
}


sub list_selectors {
    
    my @alpsvals;
    my %alpslabs;
    for (@{$models{'Alps'}}) {
	my $id = $db->get_printer_from_make_model('Alps', $_);
	push (@alpsvals, $id);
	$alpslabs{$id} = $_;
    }
    @alpsvals = sort { normalizename($alpslabs{$a}) cmp 
		       normalizename($alpslabs{$b}) } @alpsvals;
    # Make the first choice in the model list longer, so that the
    # initially created model menu is wider, because some browsers
    # (as Konqueror) do not dynamically resize the menu when the
    # model names of the newly chosen make are longer.
    #$alpslabs{$alpsvals[0]} .= "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;";

    return 
	# spit out the printer selector form
	($query->startform(-method=>'GET',
			   -action=>'show_printer.cgi'),
	 table(Tr(td('<input type=submit value="Show">'),
		  td('the'),
		  td(popup_menu(-name=>'get_make',
				-onChange=>
				'reflectMake(this.form.get_make,this.form.recnum)',
				-default=>'Alps',
				-values=>\@makers,
		                -force=>1)),
		  td(popup_menu(-name=>'recnum',
				-values=>\@alpsvals,
				-default=>$alpsvals[0],
				-labels=>\%alpslabs)),
		  td('<noscript>
   This printer selector requires javascript to work.  Nothing
   else on this site does.  You can still find your printer with the
   per-manufacturer listing below.
  </noscript>
'))),
	 $query->endform()),
      
	# Spit out the per-mfg request form
	($query->startform(-method=>'GET',
			   -action=>'printer_list.cgi'),
	 table(Tr(td('<input type=submit value="Show">'),
		  td('all printers made by'),
		  td(popup_menu(-name=>'make',
				-values=>[Anyone, @makers],
				-default=>$in_make)))),
	 $query->endform());

}


__END__

      # The old perfectly working request form
      
      if (0) {
	  ($query->startform(-method=>'GET',
			     -action=>'printer_list.cgi'),
	   table(Tr(td('<input type=submit value="Show">'),
		    td('all perfectly working'),
		    td(popup_menu(-name=>'predef',
				  -values=>[qw/CJ BJ CL BL/],
				  -labels=>{'CJ'=>'Color inkjets',
					    'BJ'=>'Black and white inkjets',
					    'CL'=>'Color lasers and LEDs',
					    'BL'=>'BW lasers and LEDs'},
				  -default=>'CJ')))),
	   $query->endform()),
      }
