#! /usr/bin/perl

use Theme qw/navbar2/;
use CGI qw/ :standard :html3 /;
use CGI escape, unescape, font, center;
use lib "@LIBDIR@";
use Foomatic::Defaults;
use Foomatic::DB qw/normalizename/;
use PHTDB qw/$hcache $cachedir/;
use driverpackages qw/numpackagesformask downloadlinkline
                      drivermask driverinfobox/;

my %driver_types = ('F' => 'Filter',
		    'V' => 'OpenPrinting Vector',
		    'I' => 'IJS',
		    'C' => 'CUPS Raster',
		    'P' => 'PostScript',
		    'U' => 'Ghostscript Uniprint',
		    'G' => 'Ghostscript built-in');

my $ua = $ENV{'HTTP_USER_AGENT'};
my $query = new CGI;
my $dbh;

# Check for special format request
my $driver = $query->param('driver');
my $fromprinter = $query->param('fromprinter');
if ($driver =~ m!^[\d\w\-\.\_]+$!
    and ($fromprinter =~ m!^[\d\w\_\-\.]+$! or !$fromprinter)
    and (`ls -1 $libdir/db/source/driver/ | grep -c ^${driver}.xml` !~ /^0$/))
{

    my @output = ();

    my $cachepage = "$hcache/show-driver-driver-$driver.html";
    if (-f $cachepage) {
	for (my $i = 0; $i < 20; $i++) {
	    open CFILE, $cachepage or die "error opening $cachepage";
	    my $o = join('', <CFILE>);
	    close CFILE;
	    # Only output the cashed page if it was really completely written
	    # Before introduction of this measure pages would not display due
	    # to an incomplete cache file until the next page rebuild (or
	    # until manually nuking the cache).
	    if ($o =~ m!</html>\s*$!s) {
		@output = ($o);
		last;
	    }
	    # If we have non-empty output, the other process is already
	    # writing the cache file or there is definitely written a broken
	    # cache. Retry only once in this case.
	    $i = 18 if $o;
	    sleep 10;
	}
    }

    if (!@output) {
	# Open cache file already now to tell concurrent processes that we are
	# already computing this case
	umask 0002;
	if (! -d $hcache) {
	    mkdir $hcache or warn "Could not mkdir $hcache";
	    chmod 2775, $hcache;
	}
	die "Error writing $cachepage\n" if !(open CFILE, "> $cachepage");

	# Get stuff from database
	my $db = new Foomatic::DB;
	my $overfile = "$cachedir/pcache/overview.perl";
	$db->{'overviewfile'} = $overfile;

	# Construct structure with driver information
	my $drvdata = $db->get_driver($driver);

	# Get list of printers for driver
	my @printers = $db->get_printers_for_driver($driver);

	my $pn;
	for $pn (@printers) {
	    my $p = $db->get_printer($pn);
	    my $prn;
	    $prn->{'make'} = $p->{'make'};
	    $prn->{'model'} = $p->{'model'};
	    $prn->{'recnum'} = ($p->{'id'} ? $p->{'id'} : $pn);
	    $prn->{'noxmlentry'} = $p->{'noxmlentry'};
	    $prn->{'noentry'} = 0;
	    next if !$prn->{'recnum'};

	    # Fix make and model if no entry for the printer exists
	    if (!$prn->{'make'}) {
		$prn->{'recnum'} =~ m/^\s*([^\-]+?)\-/;
		$prn->{'make'} = $1;
		$prn->{'make'} =~ s/_/ /g;
		next if !$prn->{'make'};
		$prn->{'noentry'} = 1;
	    }

	    if (!$prn->{'model'}) {
		$prn->{'recnum'} =~ m/^\s*[^\-]+?\-(\S+)\s*$/;
		$prn->{'model'} = $1;
		$prn->{'model'} =~ s/_/ /g;
		$prn->{'model'} =~ s/plus/+/ig;
		next if !$prn->{'model'};
		$prn->{'noentry'} = 1;
	    }

	    # pd_notes are on the driver!
	    for (@{$drvdata->{'printers'}}) {
		next if (!$_->{id});
		if ($_->{id} eq $pn) {
		    $prn->{'pd_notes'} = $_->{'comment'};
		    last;
		}
	    }

	    push @plist, $prn;
	}

	## Now, construct HTML for all this stuff

	# Format up the list of printers
	my @printerlist;
	my $prn;
	for $prn (sort { normalizename($a->{'make'} . " " . $a->{'model'}) cmp
			 normalizename($b->{'make'} . " " . $b->{'model'}) }
		  @plist) {

	    my ($make, $model, $recnum, $notes) 
		= ($prn->{'make'},
		   $prn->{'model'},
		   $prn->{'recnum'},
		   $prn->{'pd_notes'});

	    push (@printerlist,
		  ($prn->{'noentry'} ?
		   li($make . " " . $model) :
		   li(a({'href'=>"show_printer.cgi?recnum=$recnum"},
			$make . " " . $model))));
	}

	my $drivmask = drivermask($drvdata); 
	my $driverdownload = (numpackagesformask($drivmask) ?
			      downloadlinkline($drivmask, 
					       "<b>Download driver packages: </b>",
					       0) . " <font size=-3>(<a href=\"http://www.linux-foundation.org/en/OpenPrinting/Database/DriverPackages\">How to install</a>)</font><br>" : "");
	my @systems = ();
	my @execdocs = ();
	if ($drvdata->{'cmd'}) {

	    # Build lists for PPD download and driver execution info
	    my %mkmodels;

	    for $prn (@plist) {
		next if $prn->{'noxmlentry'};
		my ($make, $model, $recnum, $notes) 
		    = ($prn->{'make'},
		       $prn->{'model'},
		       $prn->{'recnum'},
		       $prn->{'pd_notes'});
		$mkmodels{$recnum} = "$make $model";
		push(@recnums, $recnum);
	    }
	    my @recnums = sort { normalizename($mkmodels{$a}) cmp
				 normalizename($mkmodels{$b}) } keys(%mkmodels);

	    push(@systems,
		 dl(dt(b('PPD-O-Matic')),
		    dd("Automagic PPD file generator for",
		       ul(li(b("CUPS"),
			     "(<a href=\"cups-doc.html\">documentation</a>),"),
			  li(b("PPR"),
			     "(<a href=\"ppr-doc.html\">documentation</a>),"),
			  li(b("LPD/LPRng/GNUlpr"),
			     "(<a href=\"lpd-doc.html\">documentation</a>),"),
			  li(b("PDQ"),
			     "(<a href=\"pdq-doc.html\">documentation</a>),"),
			  li(b("Spooler-less"), " printing",
			     "(<a href=\"direct-doc.html\">documentation</a>),"),
			  li(b("PPD aware applications/clients"),
			     "(<a href=\"ppd-doc.html\">documentation</a>),")),
		       "<p>",
		       (($driver =~ /^(Postscript|pxlmono)$/i) ?
			"<em>Note:</em> If available for the particular " .
			"printer, a PPD file " .
			"from the manufacturer will be supplied. It can " .
			"be used as " .
			"well with Foomatic and gives access to all " .
			"options of the printer. " .
			"If the file is displayed on the screen and not " .
			"directly downloaded, " .
			"please use 'File'/'Save as ...' in your " .
			"browser's menus, do not " .
			"copy and paste the file with your mouse.<p>": ""),
		       "<em>Important for Windows clients</em>: " .
		       "The CUPS PostScript driver for Windows has a bug " .
		       "which makes it choking on PPD files which contain " .
		       "GUI texts longer than 39 characters. Therefore it " .
		       "is recommended to use Adobe's PostScript driver. If " .
		       "you still want to use the CUPS driver, please mark " .
		       "\"GUI texts limited to 39 characters\" to get an " .
		       "appropriate PPD file.",
		       startform(-method=>'GET',
				 -action=>'ppd-o-matic.cgi'),
		       hidden(-name=>'driver', -value=>$driver),
		       "Select printer:",
		       popup_menu(-name=>'printer',
				  -values=>\@recnums,
				  -labels=>\%mkmodels,
				  -default=>$fromprinter),
		       submit(-value=>'Generate PPD file'), br(),
		       radio_group(-name=>'show',
				   -values=>['0','1'],
				   -default=>'0',
				   -labels=>{'0'=>'download',
					     '1'=>'display'}),
		       "the PPD file", br(),
		       checkbox(-checked=>0,
				-name=>'shortgui',
				-label=>"GUI texts limited to 39 characters"),
		       endform(),

		       )
		    )
		 );
	    push(@execdocs,
		 dl(dd("Information on how this driver is used, independent ",
		       "of a particular printing system<p>"),
		    dd(startform(-method=>'GET',
				 -action=>'execution.cgi'),
		       hidden(-name=>'driver', -value=>$driver),
		       "Select printer:",
		       popup_menu(-name=>'printer',
				  -values=>\@recnums,
				  -labels=>\%mkmodels,
				  -default=>$fromprinter),
		       submit(-value=>'Show execution details'),
		       endform())
		    )
		 );
	}

	# Compute driver type string
	my $type = $drvdata->{'type'};
	my $typestr = $driver_types{$type};
	my $comment = $drvdata->{'comment'};
	my $url = $drvdata->{'url'};

	# Construct the whole page
	push(@output,
	     start_html(-title => "OpenPrinting database - Driver: $driver",
			-author => 'Grant Taylor & Till Kamppeter',
			-BGCOLOR => 'white'),
	     navbar2('driver_list.cgi',
		     "Driver: $driver",
		     p(a({href=>"./"},"Home"),
		       ' <b>:</b> ',
		       a({href=>"database.html"}, 'Database'),
		       ' <b>:</b> ',
		       a({href=>"driver_list.cgi"}, "Drivers"),
		       ' <b>:</b> ',
		       "$driver"),
		     p(driverinfobox($drvdata)),
		     p($comment),
		     scalar(@systems) ? h3("Printing system interfaces") : '',
		     @systems,
		     scalar(@execdocs) ? h3("Execution details") : '',
		     @execdocs,
		     h3("Printers"),
		     ul(scalar(@printerlist) ? @printerlist : " ")),
	     end_html());
	
	
	# Write preparsed web page cache
	if ($cachepage) {
	    print CFILE @output;
	    close CFILE;
	} else {
	    warn "unable to write $cachepage";
	}
    }
    
    print header(-expires=>'3m', -type=>'text/html; charset=UTF-8'), @output;
}

exit 0;
