#! /usr/bin/perl

use Theme qw/navbar2/;
use CGI qw/ :standard :html3 /;
use CGI escape, unescape, font, center;
use lib "@LIBDIR@";
use Foomatic::DB qw/normalizename/;
use PHTDB qw/$hcache $cachedir/;

my $ua = $ENV{'HTTP_USER_AGENT'};
my $query = new CGI;
my $dbh;

my $siteurl = "@SITEURL@";

# Check for special format request
my $driver = $query->param('driver');
my $original = ($query->param('original') ? 1 : 0);

if ($driver =~ m!^[\d\w\-\.\_]+$!) {

    my @output = ();

    my $cachepage = "$hcache/show-license-driver-$driver-$original.html";
    if (-f $cachepage) {
	for (my $i = 0; $i < 20; $i++) {
	    open CFILE, $cachepage or die "error opening $cachepage";
	    my $o = join('', <CFILE>);
	    close CFILE;
	    # Only output the cashed page if it was really completely written
	    # Before introduction of this measure pages would not display due
	    # to an incomplete cache file until the next page rebuild (or
	    # until manually nuking the cache).
	    if ($o =~ m!</html>\s*$!s) {
		@output = ($o);
		last;
	    }
	    # If we have non-empty output, the other process is already
	    # writing the cache file or there is definitely written a broken
	    # cache. Retry only once in this case.
	    $i = 18 if $o;
	    sleep 10;
	}
    }

    if (!@output) {
	# Open cache file already now to tell concurrent processes that we are
	# already computing this case
	umask 0002;
	if (! -d $hcache) {
	    mkdir $hcache or warn "Could not mkdir $hcache";
	    chmod 2775, $hcache;
	}
	die "Error writing $cachepage\n" if !(open CFILE, "> $cachepage");

	# Get stuff from database
	my $db = new Foomatic::DB;
	my $overfile = "$cachedir/pcache/overview.perl";
	$db->{'overviewfile'} = $overfile;

	# Construct structure with driver information
	my $drvdata = $db->get_driver($driver);
	my $licensetext;
	my $prefix = '';
	$prefix = 'orig' if $original;
	my $link;
	if ($link = $drvdata->{"${orig}licenselink"}) {
	    if ($link !~ m!^((https?|ftp)://|/)!) {
		$link = "foomatic-db/db/source/" . $link;
	    }
	    if ($link !~ m!^(https?|ftp)://!) {
		$link = $siteurl . "/" . $link;
	    }
	    $licensetext = $db->getpage($link, 1);
	} else {
	    $licensetext = $drvdata->{"${orig}licensetext"};
	}
	# Construct the whole page
	push(@output,
	     start_html(-title => "OpenPrinting database - " .
			"License for driver: $driver",
			-author => 'Till Kamppeter',
			-BGCOLOR => 'white'),
	     navbar2('show_license.cgi',
		     "License for driver $driver",
		     p(a({href=>"./"},"Home"),
		       ' <b>:</b> ',
		       a({href=>"database.html"}, 'Database'),
		       ' <b>:</b> ',
		       a({href=>"driver_list.cgi"}, "Drivers"),
		       ' <b>:</b> ',
		       a({href=>"show_driver.cgi?driver=$driver"}, "$driver"),
		       ' <b>:</b> ',
		       "License"),
		     (!$original || $link ?
		      p(join(", ",
			     (($link ?
			       a({href=>"$link"},
				 "View/Download license text file") : ()),
			      (!$original ?
			       a({href=>"show_license.cgi?driver=$driver&original=1"},
				 "View original license text") : ())))) : ()),
		     p("<pre>$licensetext</pre>")),
		     end_html());
	
	# Write preparsed web page cache
	if ($cachepage) {
	    print CFILE @output;
	    close CFILE;
	} else {
	    warn "unable to write $cachepage";
	}
    }
    
    print header(-expires=>'-15m', -type=>'text/html; charset=UTF-8'), @output;
}

exit 0;
