#! /usr/bin/perl

use Theme qw/navbar2/;
use CGI qw/ :standard :html3 /;
use CGI escape, unescape, font, center;

use lib "@LIBDIR@";
use Foomatic::Defaults;
use Foomatic::DB;
use PHTDB qw/$hcache $cachedir/;
use driverpackages qw/numpackagesformask downloadlinkline drivermask
    driverinfobox/;

my $pwd = `pwd`;
$pwd =~ s/[\s\r\n]+$//;

my $wikishdir = "@WIKISHDIR@";
my $foomaticpath="$pwd/@FOOUTILDIR@";
my $query = new CGI;

# Check for special format request
my $oid = $query->param('recnum');
if ($oid =~ m!^[\d\w\_\-\.]+$!) {

    # If they had a get_make=; then bounce them to the non-getmake url
    # so that they won't bookmark the goofy one.
    if (defined($query->param('get_make'))) {
	print redirect("show_printer.cgi?recnum=$oid");
	exit(0);
    }

    # Check if supplied printer ID exists as user contribution in the
    # Wikish directory
    my $wikishfile = "$wikishdir/editprinter:$oid,v";
    my $wikishid;
    if (-r $wikishfile) {
	$wikishid = "editprinter:$oid";
    } else {
	# Do we have an old numerical printer ID or an otherwise renamed or
	# aliased printer? If so, bounce to the correct printer ID.
	my $newid = PHTDB::translate_printer_id($oid);
	if ($newid ne $oid) {
	    print redirect("show_printer.cgi?recnum=$newid");
	    exit(0);
	}
    }

    # Nothing above here should be computationally intensive!
    #####################

    # OK, is there a cached page?
    my $cachepage = "$hcache/show-printer-recnum-$oid.html";

    my @output = ();
    if (-f $cachepage) {
	for (my $i = 0; $i < 20; $i++) {
	    open CFILE, $cachepage or die "error opening $cachepage";
	    my $o = join('', <CFILE>);
	    close CFILE;
	    # Only output the cashed page if it was really completely written
	    # Before introduction of this measure pages would not display due
	    # to an incomplete cache file until the next page rebuild (or
	    # until manually nuking the cache).
	    if ($o =~ m!</html>\s*$!s) {
		@output = ($o);
		last;
	    }
	    # If we have non-empty output, the other process is already
	    # writing the cache file or there is definitely written a broken
	    # cache. Retry only once in this case.
	    $i = 18 if $o;
	    sleep 10;
	}
    }
    if (!@output) {
	# Open cache file already now to tell concurrent processes that we are
	# already computing this case
	umask 0002;
	if (! -d $hcache) {
	    mkdir $hcache or warn "Could not mkdir $hcache";
	    chmod 2775, $hcache;
	}
	die "Error writing $cachepage\n" if !(open CFILE, "> $cachepage");

	# Get driver information out of the overview
	my $db = new PHTDB;
	my $overfile = "$cachedir/pcache/overview.perl";
	$db->{'overviewfile'} = $overfile;

	my @support = ();
	my @notes;
	my $headtable;
	my ($p, $make, $model, $nbspmake, $urlmake, $nbspmodel);
	# Printer entry from Foomatic database
	$p = $db->get_printer($oid);
	# Printer entry from Wikish directory
	my $wikishprinter = 0;
	if ((!defined($p) || ($p->{noxmlentry})) && $wikishid) {
	    my $VAR1;
	    eval (`cd $wikishdir; co -q -p '$wikishid' | $foomaticpath/foomatic-perl-data -P`);
	    $p = $VAR1 if $VAR1;
	    $wikishprinter = 1;
	}
	if(defined($p)) {
	    # Name
	    $model = $p->{'model'};
	    $make = $p->{'make'};
	    my $url = $p->{'url'};
	    $nbspmodel = $model;
	    $nbspmodel =~ s/ /&nbsp;/g;
	    $nbspmake = $make;
	    $nbspmake =~ s/ /&nbsp;/g;
	    $urlmake = $make;
	    $urlmake =~ s/ /\%20/g;

	    my $noxmlentry = $p->{noxmlentry};
	    
	    # Driver section
	    my $recommended;
	    my $recommended_str;
	    my $havedriver = 0;
	    my @drivers;
	    my ($d, $driver_str);
	    my @dlist;
	    my $foomaticdrv;
	    @dlist = $db->get_drivers_for_printer($oid);
	    for my $d (@dlist) {
		$foomaticdrv->{$d} = 1;
	    }
	    @dlist = () if (!$dlist[0]);
	    if (defined($p->{'drivers'})) {
		for my $d (@{$p->{'drivers'}}) {
		    my $drv = $d->{'id'};
		    push(@dlist, $drv) if !member($drv,@dlist);
		}
	    }
	    
	    if(defined(@dlist) && ($#dlist >= 0) && $dlist[0]) {
		# Recommended driver
		$recommended = $p->{'driver'};
		# sort: recommended first; then drivers from Foomatic driver
		#       entries, after that user-suggested, otherwise 
		#       alphabetical
		for $d (sort { ($recommended eq $a 
				? -1
				: ($recommended eq $b 
				   ? 1
				   : (($foomaticdrv->{$a} &&
				      !$foomaticdrv->{$b})
				      ? -1
				      : (($foomaticdrv->{$b} &&
					  !$foomaticdrv->{$a})
					 ? 1
					 : Foomatic::DB::normalizename($a) cmp
					   Foomatic::DB::normalizename($b)))))}
			@dlist) {
		    $havedriver = 1;
		    my $drvdata = $db->get_driver($d);
		    my $newdriver = (defined($drvdata) ? 0 : 1);
		    my $dname = $d;
		    my $desc = '';
		    $desc = $drvdata->{'comment'} if !$newdriver;
		    my $lnk = '';
		    $lnk = "show_driver.cgi?driver=$dname&fromprinter=$oid"
			if !$newdriver;

		    my ($pdelem, $pdnotes);
		    my $printerindrvxml = 0;
		    if (!$newdriver) {
			for $pdelem (@{$drvdata->{'printers'}}) {
			    next if (!$pdelem->{'id'});

			    my $pid = $pdelem->{'id'};
			    next if ($pid ne "$oid");

			    $printerindrvxml = 1;

			    my $n = $pdelem->{'comment'};
			    $pdnotes = "$n<p>" if defined($n);
			}
		    }

		    my $ppdlnk;
		    my $ppdurl;
		    my $prndriverlistentry;
		    if (defined($p->{'drivers'})) {
			for my $d (@{$p->{'drivers'}}) {
			    next if ($d->{'id'} ne $dname);
			    $prndriverlistentry = $d;
			    if ($d->{'ppd'}) {
				$ppdurl = $d->{'ppd'};
			    }
			    if ($d->{'comment'}) {
				if ($newdriver) {
				    $desc = $d->{'comment'};
				} else {
				    $pdnotes .= "$d->{'comment'}<p>";
				}
			    }
			    last;
			}
		    }

		    if (!$ppdurl) {
			if (($dname =~ /^Postscript$/i) &&
			    ($p->{'ppdurl'})) {
			    $ppdurl = $p->{'ppdurl'};
			} elsif (!$newdriver && !$p->{'unverified'} &&
				 $drvdata->{'cmd'} && !$noxmlentry) {
			    $ppdlnk =
				"ppd-o-matic.cgi?driver=$dname&printer=$oid";
			}
		    }
		    if ($ppdurl && ($ppdurl =~ m!^PPD/!)) {
			$ppdurl = "foomatic-db/db/source/" . $ppdurl;
		    }
		    if (($ppdurl !~ m!:/!) &&
			(! -r $ppdurl)) {
			$ppdurl = '';
		    }
		    my $homelnk = '';
		    $homelnk = $drvdata->{'url'} if !$newdriver;
		    my $downloadlinks = '';
		    my $dmask = drivermask($drvdata);
		    $downloadlinks = downloadlinkline($dmask,
				     "Driver packages for: ", 1) .
					 ", <font size=-3><a href=\"http://www.linux-foundation.org/en/OpenPrinting/Database/DriverPackages\">How to install</a></font>"
			if numpackagesformask($dmask) > 0;
		    my $driverstr = 
			b(($lnk ? 
			   a({href=>$lnk}, $dname) : $dname)) .
			   (($homelnk || $ppdlnk || $ppdurl || $downloadlinks) ?
			    "&nbsp;(" : "") .
			    ($homelnk ?
			     a({href=>$homelnk}, "Home page") : "") .
			     (($homelnk && ($ppdlnk || $ppdurl || $downloadlinks)) ?
			      ",&nbsp;" : "") .
			      ($ppdlnk ?
			       (a({href=>"$ppdlnk&show=1"},
				  "view&nbsp;PPD") .
				",&nbsp;" .
				a({href=>"$ppdlnk&show=0"}, 
				  "download&nbsp;PPD")) : 
			       ($ppdurl ? a({href=>"$ppdurl"},
				      "custom&nbsp;PPD") : 
				"")) . 
				((($ppdlnk || $ppdurl) && $downloadlinks) ?
				 ",&nbsp;" : "") .
				 $downloadlinks .
				 (($homelnk || $ppdlnk || $ppdurl || $downloadlinks) ?
				  ")" : "");
		    if ($d eq $recommended) {
			$recommended_str = $driverstr;
		    }
		    push(@drivers,
			 ($d eq $recommended ?
			  b("Recommended driver:") . "<br>" : ()) .
			 p(driverinfobox($drvdata, $oid,
					 $prndriverlistentry,
					 $ppdurl, $ppdlnk, "bin",
					 ($p->{'unverified'} && 
					  $foomaticdrv->{$d} ?
					  "The support of the printer by " .
					  "this driver is already confirmed " .
					  "in the OpenPrinting database." :
					  ""))) .
			 (($d eq $recommended) &&
			  (scalar(@dlist) > 1) ?
			  "<p>" . b("Other drivers:") . "<br>" : ()));
		}
		$driver_str = "The following driver(s) are known to drive this printer:" . ul(@drivers);
		$driver_str = "The following driver(s) are known to drive this printer:" . p(@drivers);

	    } else {
		$driver_str = font({color=>'red'}, "No driver info!?");
	    }
	    #
	    # End of driver info, now for comments
	    
	    # Resolution
	    my $res = " ";
	    my $res_x = $p->{'maxxres'};
	    my $res_y = $p->{'maxyres'};
	    $res = "${res_x}x$res_y" if (defined($res_x));
	    
	    # Functionality score
	    my %func_colormap = ( 'A' => 'green',
				  'B' => 'green',
				  'D' => 'orange',
				  'F' => 'red' );
	    my %func_verbose = ( 'A' => 'Perfectly',
				 'B' => 'Mostly',
				 'D' => 'Partially',
				 'F' => 'Paperweight' );
	    
	    my $func_letter = $p->{'functionality'};
	    my $func_color = $func_colormap{$func_letter};
	    my $func_verb = $func_verbose{$func_letter};
	    
	    # Mechanism type
	    my %typemap = ( 'inkjet' => 'inkjet',
			    'laser' => 'laser',
			    'impact' => 'impact',
			    'braille' => 'braille',
			    'dotmatrix' => 'dot&nbsp;matrix',
			    'led' => 'LED',
			    'sublimation' => 'dye&nbsp;sublimation',
			    'transfer' => 'thermal&nbsp;transfer' );
	    my $type = $typemap{$p->{'type'}};

	    my $color = ($p->{'color'} ? 1 : 0);
	    
	    my (@comments, @comments_br);
	    
	    if (!$noxmlentry) {
		# Comments
		my $coms = $p->{'comment'};
		push (@comments, $coms)
		    if (defined($coms));

		# Wikish (not in "comments" or from xml files, totally
		# separate)
		my @wikishtext = ('<h3>User Notes</h3>');
		if (-f "$wikishdir/printer:$oid,v") {
		    open PIPE, "co -q -p '$wikishdir/printer:$oid' |" or die;
		    my @wtxt = <PIPE>;
		    close PIPE or die;
		    grep { s/\&/\&amp;/g; s/\</&lt;/g; } @wtxt;
		    if (scalar(@wtxt)) {
			push(@wikishtext, '<pre style="font-family: monospace"><tt>' . join(//,@wtxt) . '</tt></pre>');
		    }
		}

		if (!$wikishprinter) {
		    push(@wikishtext, 
			 '<p><font size="-1"><a href="edit.cgi?id=printer:' . $oid . '">[Add/Edit User Notes]</a> <a href="wikish-history.cgi?id=printer:' . $oid . '&scale=31">[Wikish History]</a></font></p>');
		    push(@comments, join('',@wikishtext));
		}
	    }
		
	    # Discussion forum link (links to list/archive etc shown by forums.cgi)
	    my $forumurl = "http://forums.openprinting.org/list.php?33";
	    my $forummake = "various manufacturers";
	    if ($make =~ /^(brother)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?24";
		$forummake = "Brother";
	    } elsif ($make =~ /^(canon)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?25";
		$forummake = "Canon";
	    } elsif ($make =~ /^(epson)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?26";
		$forummake = "Epson";
	    } elsif ($make =~ /^(apollo|hp)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?20";
		$forummake = "HP and Apollo";
	    } elsif ($make =~ /(konica|minolta|qms)/i) {
		$forumurl = "http://forums.openprinting.org/list.php?27";
		$forummake = "Konica Minolta, Minolta, and QMS";
	    } elsif ($make =~ /^(kyocera)/i) {
		$forumurl = "http://forums.openprinting.org/list.php?28";
		$forummake = "Kyocera and Kyocera-Mita";
	    } elsif ($make =~ /^(lexmark|ibm)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?29";
		$forummake = "Lexmark and IBM";
	    } elsif ($make =~ /^(ricoh|gestetner|infotec|lanier|nrg|savin)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?30";
		$forummake = "Ricoh family and OEMs (Gestetner, Infotec, Lanier, NRG, Ricoh, Savin)";
	    } elsif ($make =~ /^(samsung)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?31";
		$forummake = "Samsung";
	    } elsif ($make =~ /^(xerox|tektronix)$/i) {
		$forumurl = "http://forums.openprinting.org/list.php?32";
		$forummake = "Xerox and Tektronix";
	    }
	    push (@comments,
		  "<h3>Discussion forum</h3>",
		  p('Look for help in our ' .
		    "<a href=\"$forumurl\">" .
		    "forum for printers from $forummake</a>."));

	    ## FAQ

	    if (open FAQLINKS, "< recnum-db") {
		my @faqlinks = <FAQLINKS>;
		close FAQLINKS;
		my @linkstoadd = grep { chomp } 
		                 grep { /$oid/ } @faqlinks;
		if ($#linkstoadd >= 0) {
		    push(@comments, "<h3>FAQ</h3>" .
		                    join("<br>", @linkstoadd) . "<p>");
		}
	    }

	    ## Miscellaneous

	    if (!$noxmlentry) {
		# lang/pjl
		my $pjl = $p->{'pjl'};
		push (@comments_br, '<h3>Miscellaneous</h3>Printer ' . ($pjl ? 'supports' : 'does not support') . ' PJL.');

		# lang/text
		my $text = $p->{'ascii'};
		if (defined($text)) {
		    push (@comments_br,
			  "Printer supports direct text printing with the `$text' charset.");
		}
	    
		# Consumables
		# TODO: refill pricing/links
		my $refil = $p->{'refill'};
		push (@comments_br, "Refill: $refil")
		    if (defined($refil));

		# Contributed URL
		my $curl = $p->{'contriburl'};
		push (@comments_br, a({href=>$curl}, 'Related website'))
		    if (defined($curl));
	    
		# Autodetect
		do {
		    my @methods = ();
		    if ((($p->{'general_mfg'}) && ($p->{'general_mdl'})) ||
			($p->{'general_des'}) || ($p->{'general_ieee'})) {
			push (@methods, "Parallel Port or USB");
		    } else {
			if ((($p->{'par_mfg'}) && ($p->{'par_mdl'})) ||
			    ($p->{'par_des'}) || ($p->{'par_ieee'})) {
			    push (@methods, "Parallel Port");
			}
			if ((($p->{'usb_mfg'}) && ($p->{'usb_mdl'})) ||
			    ($p->{'usb_des'}) || ($p->{'usb_ieee'})) {
			    push (@methods, "USB");
			}
		    }
		    if ((($p->{'snmp_mfg'}) && ($p->{'snmp_mdl'})) ||
			($p->{'snmp_des'}) || ($p->{'snmp_ieee'})) {
			push (@methods, "SNMP");
		    }
		    if (scalar(@methods)) {
			push (@comments_br,
			      'Known autodetection signatures: ' 
			      . join(', ', @methods)
			      . '.');
		    } else {
			push (@comments_br, 'No known autodetection signatures.');
		    }
		};

		# Look for referral links
		my @refers;
		my $skukey;
		if (0) {
		    for $skukey ($db->get_skus($oid)) {
			my $n = '';
			if ($make =~ m/^[aeiou]/i
			    or $make eq 'HP') {
			    $n = 'n';
			}

			if ($skukey =~ m!^([^:]+):([\-\%\w\d\/]+)!) {
			    my ($vend, $sku) = ($1, $2);

			    my $flavor = undef;
			    my $literal = 0;
			    if ($vend =~ s!([\(\{])(.+)([\}\)])!!) {
				$flavor = $2;
				$literal = ($1 eq '{' ? 1 : 0);
			    }
			    my $text = "$make $model$flavor";
			    $text = $flavor if $literal;
			    my $vendname = $db->get_sku_vendor($vend);
			    my $instock = $db->get_sku_stock($skukey);
			    my $stockstr = $instock ? '' : em(' but out of stock');
			    my $price = $db->get_sku_price($skukey);
			    my $pricenum = $price;
			    $pricenum =~ s![\$\,]!!g;
			    my $pricestr = sprintf("\$%.2f", $pricenum);
			    if ($vend =~ m!^azd!) {
				$pricestr = sprintf("EUR %.2f", $pricenum);
			    }
			    if ($pricenum == 1.00 and $vend eq 'amz') {
				$pricestr = 'unknown price';
			    }
			    if ($pricenum == 1.00 and $vend eq 'azd') {
				$pricestr = 'unknown price';
			    }
			    if (defined($price)) {
				push (@refers,
				      {'text'=>$text,
				       'price'=>$pricenum,
				       'refer'=>join (' ', 
						      ($db->get_sku_link($vend, $sku, $text),
						       "<font size=\"-1\">(recently $pricestr$stockstr from $vendname)</font>"))});
			    }
			}
		    }
		}
	    }

	    my $comment_str = join("<br>", @comments_br);
	    
	    $headtable =
		table({border => 0, bgcolor => "#d0d0d0",
		       cellpadding => 6, cellspacing => 0},
		      Tr({valign=>"top", bgcolor => "#b0b0b0"},
			 td({colspan => 2},
			    font({-face=>"ms sans serif, helvetica"},
				 font({-size=>"+1"},
				      $nbspmake . "&nbsp;" .
				      ($url
				       ? a({href=>$url}, $nbspmodel) :
				       $nbspmodel))))),
		      #Tr({bgcolor => "#ffffff"},
		      #   td({colspan => 2},"&nbsp;")),
		      ($noxmlentry ? () :
		       Tr({valign=>"top"},
			  td(font({-face=>"ms sans serif, helvetica"},
				  ($color ? "Color" : "BW") .
				  ($type ? "&nbsp;$type" : "") .
				  "&nbsp;printer" .
				  (($res =~ /\d+/) ?
				   ",&nbsp;max.&nbsp;$res&nbsp;dpi" : "") .
				  ",&nbsp;" .
				  ($func_letter eq "F" ?
				   "this&nbsp;is&nbsp;a&nbsp;" :
				   "works&nbsp;") .
				  font({color=>$func_color}, $func_verb))),
			  td({-align=>"right"},
			     font({-face=>"ms sans serif, helvetica"},
				  ($func_letter eq "F" ?
				   '<img src="linuxno.png">' :
				   ($func_letter eq "D" ?
				    '<img src="linuxyes.png">' :
				    ($func_letter eq "B" ?
				     '<img src="linuxyes.png">' .
				     '<img src="linuxyes.png">' :
				     ($func_letter eq "A" ?
				      '<img src="linuxyes.png">' .
				      '<img src="linuxyes.png">' .
				      '<img src="linuxyes.png">' : '')))))))),
		      (($recommended || $havedriver) ?
		       (Tr({valign=>"top"},
			   td({colspan => 2},
			      font({-face=>"ms sans serif, helvetica"},
				   ($recommended ?
				    ("Recommended&nbsp;driver:&nbsp;" .
				     ($recommended_str ?
				      $recommended_str :
				      b(a({href=>"show_driver.cgi?driver=$recommended"},
					  $recommended)))) : "") .
				   ((!$recommended && $havedriver) ?
				    "See&nbsp;drivers&nbsp;at&nbsp;" . 
				    "the&nbsp;bottom&nbsp;of&nbsp;" . 
				    "this&nbsp;page." . br() : "") .
				   (($recommended && $havedriver) ?
				    br() : "") .
				   ($havedriver ?
				    ("Generic&nbsp;instructions&nbsp;" .
				     "for:&nbsp;" .
				     a({href=>"cups-doc.html"},
				       "CUPS") . ",&nbsp;" .
				     a({href=>"lpd-doc.html"},
				       "LPD") . ",&nbsp;" .
				     a({href=>"lpd-doc.html"},
				       "LPRng") . ",&nbsp;" .
				     a({href=>"ppr-doc.html"},
				       "PPR") . ",&nbsp;" .
				     a({href=>"pdq-doc.html"},
				       "PDQ") . ",&nbsp;" .
				     a({href=>"direct-doc.html"},
				       "no&nbsp;spooler")) :
				    ""))))) : 
		       ""));

	    # User-contributed printer from Wikish directory?
	    my $unverified = ($p->{'unverified'} ? 1 : 0);
	    if ($unverified) {
		push (@notes,
		      font({-face=>"ms sans serif, helvetica"},
			   h3("User-contributed Printer Entry"),
			   "This printer entry is not yet included in the <a href=\"foomatic.html\">Foomatic</a> packages and the data on this page is not verified or proofread.",
			   br(),
			   "<b>NOTE:</b> For user-contributed printer entries no PPD files can be generated or downloaded.",
			   br(),
			   "This entry can be edited by everyone.",
			   p(),
			   b(a({href=>"edit_printer.cgi?recnum=$oid"},
			       "Edit this printer")),
			   "<p>"));
	    } elsif ($noxmlentry) {
		if (@dlist) { 
		    push (@notes,
			  font({-face=>"ms sans serif, helvetica"},
			       h3("The properties of this printer are not yet entered into the database"),
			       "This printer is only listed here because it is in the list of supported printers of the entries for the drivers shown below.<p>",
			       "This entry can be edited by everyone.",
			       br(),
			       b(a({href=>"edit_printer.cgi?" .
					"make=$urlmake&model=$model&newentry=1"},
				   "Edit this printer")),
			       "<p>"));
		} else {
		    push (@notes,
			  font({-face=>"ms sans serif, helvetica"},
			       h3("This printer is not known to the database"),
			       "Are you sure that you have typed the address of this page correctly? If you arrived here by a link from our web site, please " .
			       a({href=>"http://bugs.linux-foundation.org/"},
				 "report a bug") . ".<p>",
			       "This entry can be edited by everyone.",
			       br(),
			       b(a({href=>"edit_printer.cgi?" .
					"make=$urlmake&model=$model&newentry=1"},
				   "Edit this printer")),
			       "<p>"));
		}
	    }

	    if (!$noxmlentry) {
		# Link to add a similar printer
		push (@notes,
		      font({-face=>"ms sans serif, helvetica"},
			   b(a({href=>"edit_printer.cgi?" .
				    "make=$urlmake&model=$model&newentry=1"},
			       "Add a similar printer")),
			   "<p>"));
	    }
	    push(@notes,
		 font({-face=>"ms sans serif, helvetica"},
		      p(@support),
		      ($noxmlentry ? "" :
		       h3(($unverified ? "" : "Edited "), "Notes")),
		      p(@comments,
			$comment_str)));

	    push(@notes,
		 font({-face=>"ms sans serif, helvetica"},
		      h3("Drivers"),
		      p($driver_str)))
		if scalar(@dlist);
	
	    push (@output,
		  start_html(-title => "OpenPrinting database - Printer: $make $model",
			     -author => 'Grant Taylor & Till Kamppeter',
			     -BGCOLOR => 'white'),
		  navbar2('printer_list.cgi',
			  "Printer:&nbsp;$nbspmake&nbsp;$nbspmodel",
			  a({href=>"./"},"Home"),
			  ' <b>:</b> ',
			  a({href=>"database.html"}, 'Database'),
			  ' <b>:</b> ',
			  a({href=>"printer_list.cgi"}, "Printers"),
			  ' <b>:</b> ',
			  a({href=>"printer_list.cgi?make=$urlmake"}, "$make"),
			  ' <b>:</b> ',
			  "$model", "<p>",
			  $headtable,
			  br(),
			  @notes),
		  end_html());
	    
	    # Write preparsed web page cache
	    if ($cachepage) {
		print CFILE @output;
		close CFILE;
	    } else {
		warn "unable to write $cachepage";
	    }
	    
	} else {

	    push (@output,
		  start_html(-title => "OpenPrinting database",
			     -author => 'Grant Taylor & Till Kamppeter',
			     -BGCOLOR => 'white'),
		  navbar2('printer_list.cgi',
			  "Printer Not Found",
			  p(a({href=>"./"},"Home"),
			    ' <b>:</b> ',
			    a({href=>"database.html"}, 'Database'),
			    ' <b>:</b> ',
			    a({href=>"printer_list.cgi"}, "Printers"),
			    ' <b>:</b> ',
			    "??????"),
			  p('Sorry, that printer seems to be missing!'),
			  p(b(a({href=>"edit_printer.cgi?recnum=$oid"},
				"Add this printer")))),
		  end_html());

	}
	
    }

    print header(-expires=>'-15m', -type=>'text/html; charset=UTF-8'), @output;
}

exit 0;

# member( $a, @b ) returns 1 if $a is in @b, 0 otherwise.
sub member { my $e = shift; foreach (@_) { $e eq $_ and return 1 } 0 };
