#! /usr/bin/perl

my $wdir = "@WIKISHDIR@";

use Time::Local;
use CGI qw/ :standard :html3/;
use CGI escape, unescape, font, center;
use Theme qw/navbar2/;

my $query = new CGI;

my $starttime = $query->param('toff');
if ($starttime) {
    die 'bogus input' unless $starttime =~ m!^\d+$!;
} else {
    $starttime = 0;		# ie now!
}
my $scale = $query->param('scale');
if ($scale) {
    die 'bogus input' unless $scale =~ m!^\d+$!;
} else {
    $scale = 7;
}


my $gmstop = scalar(gmtime(time() - (24*3600*(($starttime+0)*$scale)) ));
my $gmstart  = scalar(gmtime(time() - (24*3600*(($starttime+1)*$scale)) ));
my $logrange = sprintf("'%s>%s'", $gmstop, $gmstart);

my @files = ();
my $id = $query->param('id');
if ($id and $id =~ m!^(\w+)\:([\d\w\-\.\_]+)$!) {
    @files = ($id);
} else {
    # for every ,v file
    opendir DIR, $wdir or die;
    while ($dent=readdir(DIR)) {
	next unless $dent =~ s!\,v$!!;
	push (@files, $dent);
    }
    closedir DIR;
}

my $dent;
for $dent (@files) {
    open LOG, "rlog -d$logrange '$wdir/$dent' |" or die;
    my $s=0;
    my $revline=undef;
    my $dateline=undef;
    my @comments;
    while (<LOG>) {
	# read [rev, ip, date] tuples from rlog -d 'localtime(time()-days)<localtime(time()-days-7)'
	if ($s == 0) {
	    if (m!\-\-\-\-\-\-\-!) {
		$s=1;
	    }
	} elsif ($s==1) {
	    $revline=$_;
	    $dateline=<LOG>;
	    $s=2;
	} elsif ($s==2) {
	    if (m!^\-\-\-\-\-\-\-\-! or m!^\=\=\=\=\=\=\=\=\=\=\=\=\=\=\=!) {
		my $rev;
		if ($revline=~m!revision ([\d\.]+)!) {
		    $rev = $1;
		    my $ip;
		    if ($comments[0] =~ m!ip=([\d\.]+)!) {
			$ip = $1;
		    }
		    my $who;
		    if ($comments[0] =~ m!who=([^\s]+)!) {
			$who = $1;
		    }
		    my $datestamp;
		    if ($dateline =~ m!(\d\d\d\d)/(\d\d)/(\d\d) (\d\d):(\d\d):(\d\d);!) {
			$datestamp = timegm($6,$5,$4,$3,$2-1,$1-1900);
			push(@updates, { name=> $dent, revision=>$rev, ip=>$ip, who=>$who, datestamp=>$datestamp });
		    }
		}
		
		$s=1;
		$revline=undef;
		$dateline=undef;
		@comments=();
	    } else {
		push(@comments, $_);
	    }
	}
    }
    close LOG;
}

# sort on date
my @sorted = sort { $b->{datestamp} <=> $a->{datestamp} } @updates;

# show as (colorized) unidiffs
chdir $wdir or die;
my @body;
for (@sorted) {
    my $v = $_;
    my $rev = $v->{revision};
    if ($rev =~ m!^1\.(\d+)$!) {
	my $min = $1;
	if ($min gt 1) {
	    my $prev = sprintf("1.%d", $min-1);
	    open DIFF, "rcsdiff -r$prev -r$rev -u " . $_->{name} . " 2>/dev/null |";
	    my $nothing = <DIFF>;
	    $nothing = <DIFF>;
	    my @diff = <DIFF>;
	    close DIFF;
	    @diff = grep {$_ !~ m!^\\ No newline at end!} @diff;

	    $_->{name} =~ m!^(\w+)\:(.+)$!;
	    my ($key, $subkey) = ($1, $2);
	
	    my ($view, $edit);
	    if ($key eq 'printer') {
		$edit = sprintf("edit.cgi?id=%s", $_->{name});
		$view = sprintf("show_printer.cgi?recnum=%s", $subkey);
	    } elsif ($key eq 'prose') {
		$edit = sprintf("edit.cgi?id=%s", $_->{name});
		$view = sprintf("prose.cgi?id=%s", $subkey);
	    } else {
		$view = $edit = 'unknown key type';
	    }
	    
	    my $whohtml = $_->{who};
	    $whohtml =~ s!\&!\&amp;!g;
	    $whohtml =~ s!\<!\&lt;!g;

	    push(@html,
		 h2($_->{name}), 
		 p("Revision $rev, ", 
		   scalar(localtime($_->{datestamp})) .
		   ", ip=" . $_->{ip} .
		   ", who=" . $whohtml . ".",
		   br(),
		   a({href=>$edit},'Edit') . ',',
		   a({href=>$view}, 'View')),
		 '<pre>'
		 );

	    my $l;
	    for $l (@diff) {
		$l =~ s!\&!\&amp;!g; 
		$l =~ s!\<!\&lt;!g;
		$l =~ s!\r!!g;
		
		chomp $l;
		
		if ($l =~ m!^\-!) {
		    push(@html, "<font color='#a0a0a0'>$l</font>\n");
		} elsif ($l =~ m!^\+!) {
		    push(@html, "<font color='#00b000'>$l</font>\n");
		} else { 
		    push(@html, "$l\n");
		}
	    }
	    
	    push(@html, '</pre>');
	}
    }
}

my $idclause = '';
if (scalar(@files eq 1)) {
    $idclause = "&id=$id";
}

print
    header(-expires=>'-15m', -type=>'text/html; charset=UTF-8'),
    start_html(-title=>'OpenPrinting database - Wikish changes'),
    navbar2('wikish-history.cgi',
	    'Foomatic Wikish History',
	    p("Change from $gmstart to $gmstop."),
	    p(a({href=>sprintf("wikish-history.cgi?toff=%d&scale=%d$idclause", $starttime+1, $scale)}, '&lt;&lt;Earlier'),
	      (($starttime gt 0) ? a({href=>sprintf("wikish-history.cgi?toff=%d&scale=%d$idclause", $starttime-1, $scale)}, 
				     'Later>>') : '')),
	    @html),
    end_html();

exit 0;
