# Simple news system to have an easy (static) news page updated from small files
# To use, put <!-- _GP_ recipe("news.gp") --> into the template

#
#
# This genpage module creates a table (whose style you specify) and fills it with
# news items. Those news items are simply the contents of files. As I just found,
# this module can also be conveniently be used to create FAQs and other similar stuff.
#
# Usage is as follows :
# 1. In your news page (or wherever appropriate) create the news section by
#    placing a call to news() there :
# <BODY>
# <P>Recent news :<BR>
#   <-- _GP_ news("news/", 5, "news/news.conf" ) -->
# </BODY>
#
# If the second or last parameters are missing, some sensible defaults are taken
#
  if (!defined( $cmt_start )) { $cmt_start = "<!--"; };
  if (!defined( $cmt_end )) { $cmt_end = "-->"; };

  # This most propably messes up the global namespace, but I don't know exactly ...
  if (!defined &File::stat) { use File::stat; };

  # File::find is now already used by Genpage
  #if (!defined &File::find) { use File::find; };

	sub newsitem {
	  my $item = $_;
	  if ($item =~ /.news$/) {
	    push @NEWS_NewsFiles, $item;
	  };
 	  #$_ = $item;
	};

	sub news {
	  my $NewsDir = shift;
	  my $NewsCount = shift || 5;
	  my $NewsConf = shift;

  	  # Set up the defaults
	  my $TableColor = "#BFBFBF";
	  my $HeadlineColor = "#9F9F9F";
          my $NewsTemplate = "<CENTER><TABLE width=\"90%\" bgcolor=$TableColor>$cmt_start _NEWS_ITEMS_ $cmt_end</TABLE></CENTER>";
	  my $ItemTemplate = "<TR><TD><TABLE cellspacing=\"0\" width=\"100%\"><TR bgcolor=$HeadlineColor><TD>$cmt_start _NEWS_ TITLE $cmt_end</TD><TD align=\"right\">$cmt_start _NEWS_ DATE $cmt_end</TD></TR><TR><TD colspan=\"2\">$cmt_start _NEWS_ BODY $cmt_end</TD></TR></TABLE></TD></TR>";

	  # Now read in the defaults from the $NewsConf (if available)
	  if (defined($NewsConf)) {
	    if (-f $NewsConf) {
	      open( NEWS_NEWSCONF, "<$NewsConf" ) or die "Opening news configuration file $NewsConf: $!\n";
	      my @NewsConf = <NEWS_NEWSCONF>;
	      close( NEWS_NEWSCONF );
	      push @NewsConf, "%_NEWS_END_%\n";
	      ($NewsTemplate = join( "", @NewsConf ) )=~ s/.*%_NEWS_BODY_%((.(?!%_NEWS_))*.)(?=%_NEWS_).*/$1/s;
	      ($ItemTemplate = join( "", @NewsConf ) )=~ s/.*%_NEWS_ITEM_%((.(?!%_NEWS_))*.)(?=%_NEWS_).*/$1/s;
            } else { 
              print "Couldn't find news theme $NewsConf, using defaults."
            };
	  };

	  my $Items;
	  my $result;

	  #print "NEWS from :$NewsDir " unless $opt_q;
	  find( \&newsitem, $NewsDir );
	  #print "(" . ($#NEWS_NewsFiles+1) . " entries total)"  unless $opt_q;

	  # First, fix the names by constructing relative paths
	  my $ItemIndex = 0;
	  while ($ItemIndex <= $#NEWS_NewsFiles) {
	   $NEWS_NewsFiles[ $ItemIndex ] = build_path( $NewsDir, $NEWS_NewsFiles[ $ItemIndex ] );
	   $ItemIndex++;
	  };

	  # Now sort @NEWS_NewsFiles in descending order according to 
          # The DATE tag or the time last modified :
	  my %LastModified;
	  foreach $File (@NEWS_NewsFiles) {
	    if (open F, $File) {
	      while (<F>) {
		if (/^\s*DATE:(.*)$/) {
		  my $date = $1;
		  close F;
		  system("touch -d \"$date\" $File");
		  last;
		}
	      }
	    }
	    my $st = stat( $File ) or die "news.gp: $File $!";
	    $LastModified{ $File } = $st->mtime;
	  };
	  @NEWS_SortedFiles = sort { $LastModified{ $b } <=> $LastModified{ $a } } @NEWS_NewsFiles;

	  # Limit the amount of news :
	  if ($NewsCount > 0) {
	    $NewsCount--;
	    if ($#NEWS_SortedFiles > $NewsCount) {
  	      $#NEWS_SortedFiles = $NewsCount;
	    };
	  };
	  #print ", limited to " . ($#NEWS_SortedFiles+1) . " entries.\n" unless $opt_q; 

	  undef @NEWS_NewsFiles;

	  #my CONTENT;

	  my @Content = "";
	  my $CurrentItem = "";
	  my %Item;
	  $ItemIndex = 0;

	  while ($ItemIndex <= $#NEWS_SortedFiles) {
	    #print "NEWS: $CurrItemFile";
	    $CurrItemFile = $NEWS_SortedFiles[ $ItemIndex ];

	    open( NEWS_CONTENT, "<$CurrItemFile" ) || die "opening $CurrItemFile : $!\n";
	    @Content = <NEWS_CONTENT>;
	    close( NEWS_CONTENT );

	    my $Line = 0;
	    my $CurrLine = "";
	    my $Tag;

	    # Set up some defaults
	    $Item{ "BODY" } = "<TT><B>news.gp</B>:no news is good news (no news body given).</TT>";
	    $Item{ "DATE" } = localtime( $LastModified{ $CurrItemFile });
	    $Item{ "ID" } = $ItemIndex;

	    while (($Line <= $#Content) && (($Content[ $Line ] =~ /^(\w+):(.*)$/) || ($Content[ $Line ] =~ /^()\s*$/))) {
	      $Line++;
	      next if !$1;
	      $Tag = uc($1);
	      $CurrLine = $2;
	      $Item{ $Tag } = $CurrLine;
	    };

	    # This could be done better by using "join( "\n", @Content[ .. ])" but I'm too lazy
	    # to figure the indices out ...
	    while ($Line <= $#Content) {
	      $Item{ "BODY" } .= "\n" . $Content[ $Line++ ];
	    };

	    my $Buffer = $ItemTemplate;
            while ($Buffer =~ m/^(.*?)$cmt_start\s*_NEWS_\s*(.*?)\s*$cmt_end(.*)$/s) {
	      my ($left, $middle, $right) = ($1, $2, $3);
	      my $replacement = "<B>news.gp:</B>undefined :&quot;$middle&quot;";
	      if (defined( $Item{ $middle })) {
		$replacement = $Item{ $middle };
	      };
	      $Buffer = $left . $replacement . $right;
	    };
	    $Items .= $Buffer;

	    $ItemIndex++;
	  };

	  # Create the complete item
	  ($result = $NewsTemplate) =~ s!$cmt_start\s*_NEWS_ITEMS_\s*$cmt_end!$Items!is;

	  # And go back
	  return $result;	  
	};

return "$cmt_start Using news.gp $cmt_end";
