#
# This recipe generates a table of contents (within a file) complete
# with jumps to the relevant bits including substituting in anchors.
# It uses <H?> tags to find the relevant entries where '?' is a level
# of header. You specify the header level to use (e.g. start at H2's 
# and build the TOC from there) 
#
# This is a simple one level TOC... An obvious enhancement is to have it
# recurse within say a h1 block for h2s ,h3s etc
# Submitted by Joe Vaughan joev@freddyfrog.com  (11/06/1999)
# 
# initialisation: put <!-- _GP_ recipe($root/include/toc.gp) --> in one 
# of the parsed files e.g. your template or include it at startup using
# 'genpage -r toc.gp'
#
# use: Place a <!-- _GP_ content("TOC") --> wherever you want the TOC 
#      to appear and place <!-- _GP_ toc(start-header-level) --> 
#      somewhere that it'll get parsed late or last in the file.
#  

sub toc {
  my $startlevel = shift;
  my $stoplevel = shift;
  
  if ($startlevel < 1) { $startlevel = 1; }
  if ($startlevel > 7) { $startlevel = 7; }
  if (defined($stoplevel)) {
    if ($stoplevel < $startlevel) { $stoplevel = $startlevel; }
    if ($stoplevel > 7) { $stoplevel = 7; }
  }
  else {
    $stoplevel = 7;
  }
  my $anchornum = 0;
  my $tempbit = $left . "###TOC-SPLIT-HERE###" .$right;
  my $pattern = '^(.*?)<h([0-9])>(?!<A NAME)\s*(.*?)</h.>(.*)\$';
  my $toc = "<ul>\n";
  my $tablevel = $startlevel;
  my ($leftbit, $level,$middlebit, $rightbit );

  while ($tempbit =~ /$pattern/is ) {
    if ($tablevel < $startlevel ) { 
      $tablevel = $startlevel;
    }
    ($leftbit, $level,$middlebit, $rightbit ) = ($1, $2, $3, $4);
    if ($level >= $startlevel && $level <= $stoplevel) {
      if ($level < $tablevel) {
	for ($i = 0; $i<($tablevel-$level);$i++) {
	  $toc .="</ul>";
	}
      }
      else {
	if ($level > $tablevel) {
	  for ($i = 0; $i<($level-$tablevel);$i++) {
	    $toc .="<ul>";
	  }
	}
      }

      $tablevel = $level;
      $toc= $toc . "<li><A HREF=\"#$anchornum\">$middlebit</A></li>\n";
      $middlebit = "<H$level><A NAME= \"$anchornum\">" . $middlebit . "</A></H$level>";
      $anchornum++;
    }
    else {
      $middlebit = "<H$level><A NAME=\"\">" . $middlebit . "</A></H$level>";
    }
    $tempbit = $leftbit . $middlebit . $rightbit; 
  }
  $toc = $toc . "</ul>\n";
  # Second pass replaces the content("TOC") error tag (from earlier 
  # pass content() filtering 
  $pattern = "^(.*?)<!-- content: undefined tag: TOC -->(.*)\$";
  while ($tempbit =~ /$pattern/is ) {
    $tempbit = $1 . $toc . $2;
  }
  
  ($left,$right) = split ("###TOC-SPLIT-HERE###",$tempbit);
  
  return "<!-- Table of Contents has been autoinserted by recipe TOC.gp -->";
}

return ("<!-- Genpage recipe TOC.gp included -->");
